
# Routine to start the capture procedure.
#   Does error checking and then calls procedure to
#   create the capture GUI
proc trees_capture_run {} {
    global gui session capture labels
    
    if {$gui(hardware) == 0} {
        messagebox "Capture Error" "Cannot capture frames because this copy of TREES is not able to access system hardware."
        set mosaic(status) NONE
    } elseif {$session(status) == "NONE"} {
        messagebox "Capture Error" "Cannot capture frames because a session has not been started."
        set mosaic(status) NONE
    } elseif {$capture(status) == "COMPLETE"} {
        messagebox "Capture Info" "Frames have already been captured."
    } else {
        set capture(status) ACTIVE
        trees_update_status


        trees_capture_creategui
        
    }

    trees_update_status

    trees_autosave
}


# Build the capture GUI and initialize camera and stage variables
proc trees_capture_creategui {} {
    global trees_dir

    toplevel .capturegui -visual truecolor

    # Set up procedure to handle the case when the window manager
    #  unexpectedly deletes the capture GUI window
    wm protocol .capturegui WM_DELETE_WINDOW { 
	global capture camera

	set capture(status) NONE
	trees_update_status
	
        $camera(name) configure -loopcount 0

	StageMotion_Close
	FocusMotion_Close
        destroy .capturegui
    }

    # Initialize the stage drivers
    StageMotion_Open

    # Initialize the focus control drivers
    FocusMotion_Open


    # Define cursors for various parts of the GUI
    global cursor
    set cursor(video,normal) left_ptr
    set cursor(video,point) crosshair


    # Initialize stage motion variables
    global motion
    # By default move stage in steps of ~32 pixels
    set motion(increment) 32
    # Stage motion calibration values (should periodically be verified)
    set motion(pixperstep,x) 13.4
    set motion(pixperstep,y) 13.4


    # Initialize the stage position variables
    global position
    # Current stage position in pixels
    set position(stage-pixel,x) 0
    set position(stage-pixel,y) 0
    # Start flag stage position in pixels
    set position(start-pixel,x) 0
    set position(start-pixel,y) 0
    # Current stage position in steps
    set position(stage-physical,x) 0
    set position(stage-physical,y) 0
    # Stop flag stage position in steps
    set position(stop-physical,x) 0
    set position(stop-physical,y) 0
    # Stop flag stage position in pixels
    set position(stop-pixel,x) 0
    set position(stop-pixel,y) 0

    
    # Initialize the camera variables
    global camera
    set camera(width) 1024
    set camera(height) 768
    # Require a minimum overlap of 45% between adjacent frames
    #  in x and y directions
    set camera(minoverlap,x) [expr round($camera(width) * 0.45)]
    set camera(minoverlap,y) [expr round($camera(height) * 0.45)]


    # Initialze video frame variables
    global video_frame
    # Subsample video frame to display captured frames at 1/3 resolution
    set video_frame(deci) 3
    set video_frame(width) [expr round($camera(width) / $video_frame(deci))]
    set video_frame(height) [expr round($camera(height) / $video_frame(deci))]
    set video_frame(center,x) [expr round($video_frame(width) / 2)]
    set video_frame(center,y) [expr round($video_frame(height) / 2)]


    # Initialize symbolic tracker frame variables
    global tracker_frame
    # Set zoom value of tracker frame to 1/25
    set tracker_frame(deci) 25
    set tracker_frame(width) 1000
    set tracker_frame(height) 1000
    set tracker_frame(center,x) [expr round($tracker_frame(width) / 2)]
    set tracker_frame(center,y) [expr round($tracker_frame(height) / 2)]
    set tracker_frame(camera,width) [expr round($camera(width) / $tracker_frame(deci))]
    set tracker_frame(camera,height) [expr round($camera(height) / $tracker_frame(deci))]

    image create photo greenflag -file ${trees_dir}/flag_green.gif
    image create photo redflag -file ${trees_dir}/flag_red.gif
    image create photo greenflag_small -file ${trees_dir}/flag_green_small.gif
    image create photo redflag_small -file ${trees_dir}/flag_red_small.gif
    image create photo greenflag_button -file ${trees_dir}/flag_green_button.gif
    image create photo redflag_button -file ${trees_dir}/flag_red_button.gif




    # Create the live video window
    frame .capturegui.live -relief raised -bd 4
    canvas .capturegui.live.canvas -bg lightblue -width $video_frame(width) -height $video_frame(height)

    if {![info exists camera(state)]} {
	set camera(state) live
    }
    set camera(name) .capturegui.live.mycamera
    # Create the tkCamera image.  This is effectively a Tk photo image that is
    #  automatically updated by the camera interface routines.  See tkCamera.C
    #  for more information.
    if [catch { set camera(photo) [camera $camera(name) -deci $video_frame(deci)] } result] {
	puts stderr $result
    }
    # Place the tkCamera image in its canvas
    .capturegui.live.canvas create image $video_frame(center,x) $video_frame(center,y) -image $camera(photo) -anchor c

    # Draw the crosshair on the live video canvas
    .capturegui.live.canvas create oval [expr $video_frame(center,x) - 10] \
	    [expr $video_frame(center,y) - 10] [expr $video_frame(center,x) + 10] \
	    [expr $video_frame(center,y) + 10] -width 1 -tag crosshair
    .capturegui.live.canvas create line $video_frame(center,x) [expr $video_frame(center,y) - 15] \
	    $video_frame(center,x) [expr $video_frame(center,y) + 15] -width 1 -tag crosshair
    .capturegui.live.canvas create line [expr $video_frame(center,x) - 15] $video_frame(center,y) \
	    [expr $video_frame(center,x) + 15] $video_frame(center,y) -width 1 -tag crosshair
    pack .capturegui.live.canvas



    # Create the graphical location tracker
    Scrolled_Canvas .capturegui.tracker -relief raised -bd 4 -width 600 -height 400 \
	    -scrollregion [subst {0 0 $tracker_frame(width) $tracker_frame(height)}]
    .capturegui.tracker.canvas xview moveto 0.2
    .capturegui.tracker.canvas yview moveto 0.3
    .capturegui.tracker.canvas create line $tracker_frame(center,x) 1 \
	    $tracker_frame(center,x) $tracker_frame(height) -width 2 -tag coord_line
    .capturegui.tracker.canvas create line 1 $tracker_frame(center,y) \
	    $tracker_frame(width) $tracker_frame(center,y) -width 2 -tag coord_line
    for {set i 0} {$i <= 20} {incr i} {
        if {[expr fmod($i,2)] == 0.0} {
            if {$i != 10} {
                .capturegui.tracker.canvas create line [expr $i * ($tracker_frame(width) / 20)] 1 \
			[expr $i * ($tracker_frame(width) / 20)] $tracker_frame(height) -stipple gray50 \
			-width 1 -tag coord_line
                .capturegui.tracker.canvas create line 1 [expr $i * ($tracker_frame(height) / 20)] \
			$tracker_frame(width) [expr $i * ($tracker_frame(height) / 20)] -stipple gray50 \
			-width 1 -tag coord_line
            }
        } else {
            .capturegui.tracker.canvas create line [expr $i * ($tracker_frame(width) / 20)] 1 \
		    [expr $i * ($tracker_frame(width) / 20)] $tracker_frame(height) \
		    -width 1 -tag coord_line
            .capturegui.tracker.canvas create line 1 [expr $i * ($tracker_frame(height) / 20)] \
		    $tracker_frame(width) [expr $i * ($tracker_frame(height) / 20)] \
		    -width 1 -tag coord_line
        }
    }
    # Create a symbolic rectangle representing the current location of the stage,
    #  i.e., the current location of the video window
    .capturegui.tracker.canvas create rectangle \
	    [expr $tracker_frame(center,x) - round($tracker_frame(camera,width) / 2)] \
	    [expr $tracker_frame(center,y) - round($tracker_frame(camera,height) / 2)] \
	    [expr $tracker_frame(center,x) + round($tracker_frame(camera,width) / 2)] \
	    [expr $tracker_frame(center,y) + round($tracker_frame(camera,height) / 2)] \
	    -outline black -fill lightblue -width 2 -tag camera_frame



    # Create the control panel
    frame .capturegui.control -relief raised -bd 4

    frame .capturegui.control.camera -relief groove -bd 3
    label .capturegui.control.camera.title -text "Camera Control"
    button .capturegui.control.camera.button -text "Turn Off" \
	    -command {
	global camera
	if {$camera(state) == "live"} {
	    set camera(state) "paused"
	    .capturegui.control.camera.button configure -text "Turn On"
	    $camera(name) configure -pause 1
	} else {
	    set camera(state) "live"
	    .capturegui.control.camera.button configure -text "Turn Off"
	    $camera(name) configure -pause 0
	}
    }
    if {$camera(state) == "paused"} {
	set camera(state) "live"
	.capturegui.control.camera.button configure -text "Turn Off"
	$camera(name) configure -pause 0
    }


    grid .capturegui.control.camera.title -sticky news
    grid .capturegui.control.camera.button -sticky news



    frame .capturegui.control.motion -relief groove -bd 3
    label .capturegui.control.motion.label -text "X-Y Stage Control"
    image create bitmap uparrow -file ${trees_dir}/up_arrow.xbm
    image create bitmap downarrow -file ${trees_dir}/down_arrow.xbm
    image create bitmap rightarrow -file ${trees_dir}/right_arrow.xbm
    image create bitmap leftarrow -file ${trees_dir}/left_arrow.xbm
    button .capturegui.control.motion.right -image rightarrow \
	    -command { move_stage "right" }
    button .capturegui.control.motion.left  -image leftarrow \
	    -command { move_stage "left" }
    button .capturegui.control.motion.up    -image uparrow \
	    -command { move_stage "up" }
    button .capturegui.control.motion.down  -image downarrow \
	    -command { move_stage "down" }
    label .capturegui.control.motion.incrlbl -text "Motion Increment:"

    # Allow the stage to be moved in increments of either
    #  32 pixels (~ min. resolution of stepper motors),
    #  512 pixels (~ half a frame),
    #  1024 pixels (~ full frame),
    #  or 4096 pixels (~ four frames)
    set motion(increment) 1024
    tk_optionMenu .capturegui.control.motion.incrmenu motion(increment) 32 512 1024 4096
    label .capturegui.control.motion.incrunits -text "pixels"

    grid .capturegui.control.motion.label - - - - - -sticky news
    grid x .capturegui.control.motion.up x x .capturegui.control.motion.incrlbl - -sticky news
    grid .capturegui.control.motion.left x .capturegui.control.motion.right x .capturegui.control.motion.incrmenu .capturegui.control.motion.incrunits -sticky news
    grid x .capturegui.control.motion.down x x x x
    grid columnconfigure .capturegui.control.motion 3 -minsize 10
    grid columnconfigure .capturegui.control.motion 4 -minsize 80


    frame .capturegui.control.focus -relief groove -bd 3
    label .capturegui.control.focus.title -text "Focus Control"
    frame .capturegui.control.focus.motion
    button .capturegui.control.focus.motion.up -image uparrow \
	    -command {FocusMotion_MoveUp 16}
    button .capturegui.control.focus.motion.down -image downarrow \
	    -command {FocusMotion_MoveDown 16}

    grid .capturegui.control.focus.title -sticky news
    grid .capturegui.control.focus.motion -sticky news
    grid .capturegui.control.focus.motion.up -sticky news
    grid .capturegui.control.focus.motion.down -sticky news

    frame .capturegui.control.markers -relief groove -bd 3
    label .capturegui.control.markers.startlbl -text "Place Start Flag"
    button .capturegui.control.markers.startbtn -image greenflag_button -command place_start_flag
    label .capturegui.control.markers.stoplbl -text "Place Stop Flag"
    button .capturegui.control.markers.stopbtn -image redflag_button -command place_stop_flag
    grid .capturegui.control.markers.startlbl x .capturegui.control.markers.stoplbl -sticky news
    grid .capturegui.control.markers.startbtn x .capturegui.control.markers.stopbtn -sticky ns
    grid columnconfigure .capturegui.control.markers 1 -minsize 10


    frame .capturegui.control.coord -relief groove -bd 3
    label .capturegui.control.coord.pixlbl -text "Stage Position (pixel): "
    label .capturegui.control.coord.physlbl -text "Stage Position (steps):  "
    label .capturegui.control.coord.pixcoord -text "(0,0)"
    label .capturegui.control.coord.physcoord -text "(0,0)"
    grid .capturegui.control.coord.pixlbl .capturegui.control.coord.pixcoord -sticky news
    grid .capturegui.control.coord.physlbl .capturegui.control.coord.physcoord -sticky news
    grid columnconfigure .capturegui.control.coord 1 -minsize 100


    grid .capturegui.control.camera -row 0 -column 0 -sticky news
    grid .capturegui.control.motion -row 0 -column 1 -sticky news
    grid .capturegui.control.focus -row 0 -column 2 -sticky news
    grid .capturegui.control.markers -row 2 -column 0 -columnspan 2 -sticky nw
    grid .capturegui.control.coord -row 4 -column 0 -columnspan 2 -sticky nw
    grid rowconfigure .capturegui.control 1 -minsize 10
    grid rowconfigure .capturegui.control 3 -minsize 10




    # Lay out the overall gui configuration
    grid .capturegui.live -row 0 -column 0 -sticky news
    grid .capturegui.control -row 0 -column 1 -sticky news
    grid .capturegui.tracker -row 1 -column 0 -columnspan 2 -sticky news
    grid rowconfigure .capturegui 0 -weight 0
    grid rowconfigure .capturegui 1 -weight 1
    grid columnconfigure .capturegui 0 -weight 0
    grid columnconfigure .capturegui 1 -weight 1


}





# Procedure to create a scrolled canvas, taken from Practical
#  Programming in Tcl/Tk, by Brent Welch
proc Scrolled_Canvas { c args } {
    frame $c
    eval {canvas $c.canvas \
        -xscrollcommand [list $c.xscroll set] \
        -yscrollcommand [list $c.yscroll set] \
        -highlightthickness 0 \
        -borderwidth 0} $args
    scrollbar $c.xscroll -orient horizontal \
        -command [list $c.canvas xview]
    scrollbar $c.yscroll -orient vertical \
        -command [list $c.canvas yview]
    grid $c.canvas $c.yscroll -sticky news
    grid $c.xscroll -sticky ew
    grid rowconfigure $c 0 -weight 1
    grid columnconfigure $c 0 -weight 1
    return $c.canvas
}


# Procedure to drop a start flag on the video window
proc place_start_flag {} {
    global cursor tracker_frame video_frame

    .capturegui.live config -cursor $cursor(video,point)
    
    bind .capturegui.live.canvas <Button-1> {
        global position
    
	# If a start flag already exists, delete it before placing
	#  the new start flag
	if {[.capturegui.live.canvas find withtag greenflag] != ""} {
            .capturegui.live.canvas delete greenflag
            .capturegui.tracker.canvas delete greenflag
        }

	# Place the start flag in the video canvas
        .capturegui.live.canvas create image %x %y -image greenflag -anchor center -tag {flag greenflag}
        .capturegui.live.canvas raise crosshair

	# Compute the position of the start flag relative to the center of the full resolution
	#  frame that is displayed in the live video window (in pixels)
	set flag_offset_from_center(x) [expr round(([lindex [.capturegui.live.canvas coords greenflag] 0] - $video_frame(center,x)) * $video_frame(deci))]
        set flag_offset_from_center(y) [expr round(([lindex [.capturegui.live.canvas coords greenflag] 1] - $video_frame(center,y)) * $video_frame(deci))]

	# Set the start flag pixel position relative to the global position coordinate
	#  system
        set position(start-pixel,x) [request_x_pos [expr $position(stage-pixel,x) + $flag_offset_from_center(x)]]
        set position(start-pixel,y) [request_y_pos [expr $position(stage-pixel,y) - $flag_offset_from_center(y)]]

	# Draw a mini-start flag on the tracker canvas
	.capturegui.tracker.canvas create image [expr  round($position(start-pixel,x)/$tracker_frame(deci))+$tracker_frame(center,x)] [expr  -(round($position(start-pixel,y)/$tracker_frame(deci)))+$tracker_frame(center,y)] -image greenflag_small -anchor center -tag {flag greenflag}

	
        .capturegui.control.coord.physcoord configure -text "($position(stage-physical,x),$position(stage-physical,y))"
        
        .capturegui.live config -cursor $cursor(video,normal)

	# Turn off the binding that was created when the user clicked on the "Create Start Flag" button
	bind .capturegui.live.canvas <Button-1> ""

    }
}



# Procedure to drop a stop flag on the video canvas
proc place_stop_flag {} {
    global cursor tracker_frame video_frame

    .capturegui.live config -cursor $cursor(video,point)
    
    bind .capturegui.live.canvas <Button-1> {
        global position
    
	# If a stop flag already exists, delete it before creating the
	#  new stop flag
        if {[.capturegui.live.canvas find withtag redflag] != ""} {
            .capturegui.live.canvas delete redflag
            .capturegui.tracker.canvas delete redflag
        }

	# Draw the stop flag on the live video canvas
        .capturegui.live.canvas create image %x %y -image redflag -anchor center -tag {flag redflag}
        .capturegui.live.canvas raise crosshair

	# Compute the position of the stop flag relative to the center of the full resolution
	#  frame that is displayed in the live video window (in pixels)
        set flag_offset_from_center(x) [expr round(([lindex [.capturegui.live.canvas coords redflag] 0] - $video_frame(center,x)) * $video_frame(deci))]
        set flag_offset_from_center(y) [expr round(([lindex [.capturegui.live.canvas coords redflag] 1] - $video_frame(center,y)) * $video_frame(deci))]

	# Set the stop flag pixel position relative to the global position coordinate
	#  system
        set position(stop-pixel,x) [request_x_pos [expr $position(stage-pixel,x) + $flag_offset_from_center(x)]]
        set position(stop-pixel,y) [request_y_pos [expr $position(stage-pixel,y) - $flag_offset_from_center(y)]]

	# Draw a mini-stop flag on the tracker canvas
        .capturegui.tracker.canvas create image [expr  round($position(stop-pixel,x)/$tracker_frame(deci))+$tracker_frame(center,x)] [expr  -(round($position(stop-pixel,y)/$tracker_frame(deci)))+$tracker_frame(center,y)] -image redflag_small -anchor center -tag {flag redflag}

        .capturegui.live config -cursor $cursor(video,normal)

	# Delete the binding that was created when the user clicked on the
	#  "Drop Stop Flag" button
        bind .capturegui.live.canvas <Button-1> ""

	# Compute the number and positions of overlapping frames that must be captured
	#  in order to create a mosaic of the sample between the start and stop
	#  flag locations
        compute_capture_frames
    }
}


# Compute the number and positions of overlapping frames that must be captured
#  in order to create a mosaic of the sample between the start and stop
#  flag locations
proc compute_capture_frames {} {
    global position tracker_frame camera frame motion
    global session
    global capture
    global gaugevar
    global stagemotion
    
    # If symbolic capture frames have already been drawn to the tracker
    #  canvas, delete them
    if {[.capturegui.tracker.canvas find withtag capture_frame] != ""} {
        .capturegui.tracker.canvas delete capture_frame
    }

    ####
    # Compute the frame count and offsets of captured frames, requiring a 30% overlap in each axis
    ####
    set width [expr $position(stop-pixel,x) - $position(start-pixel,x)]
    set height [expr $position(stop-pixel,y) - $position(start-pixel,y)]
    
    if {[expr abs($width)] > [expr abs($height)]} {
        set frame(count) [expr int(abs($width)/($camera(width) - $camera(minoverlap,x)) + 1) + 1]
        set y_overlap [expr $camera(height) - (abs($height)/($frame(count) - 1))]
	
        if {$y_overlap < $camera(minoverlap,y)} {
            set frame(count) [expr int(abs($height)/($camera(height) - $camera(minoverlap,y)) + 1) + 1]
            set y_overlap [expr $camera(height) - (abs($height)/($frame(count) - 1))]
        }
    } else {
        set frame(count) [expr int(abs($height)/($camera(height) - $camera(minoverlap,y)) + 1) + 1]
        set x_overlap [expr $camera(width) - (abs($width)/($frame(count) - 1))]

        if {$x_overlap < $camera(minoverlap,x)} {
            set frame(count) [expr int(abs($width)/($camera(width) - $camera(minoverlap,x)) + 1) + 1]
            set x_overlap [expr $camera(width) - (abs($width)/($frame(count) - 1))]
        }
    }
    
    set x_offset [request_x_pos [expr round(($position(stop-pixel,x) - $position(start-pixel,x))/($frame(count) - 1))]]
    set y_offset [request_y_pos [expr round(($position(stop-pixel,y) - $position(start-pixel,y))/($frame(count) - 1))]]


    # Draw symbolic rectangles in the tracker canvas to show the locations of each of the
    #  frames that must be captured in order to create the mosaic
    for {set i 0} {$i < $frame(count)} {incr i} {
        set center_x [expr (($position(start-pixel,x) + ($i * $x_offset))/$tracker_frame(deci)) + $tracker_frame(center,x)]
        set center_y [expr ((-$position(start-pixel,y) - ($i * $y_offset))/$tracker_frame(deci)) + $tracker_frame(center,y)]

        .capturegui.tracker.canvas create rectangle [expr $center_x - round($tracker_frame(camera,width)/2)] [expr $center_y - round($tracker_frame(camera,height)/2)] [expr $center_x + round($tracker_frame(camera,width)/2)] [expr $center_y + round($tracker_frame(camera,height)/2)] -outline black -fill wheat -width 1 -tag capture_frame
    }
        
    .capturegui.tracker.canvas raise camera_frame
    .capturegui.tracker.canvas raise flag


    set savedir $session(directory)/images/frames
    set basename "image"
    set infoname "image_info.txt"
    set illumname "illum0.SAD"

    # Ask the analyst if the frames should be captured
    set choice [tk_messageBox -type yesno -default yes \
	    -message "A total of $frame(count) frames are required to capture this sequence of images.  Do you wish to capture these frames?" \
	    -icon question]

    # If yes, go ahead and capture the frames
    if {$choice == "yes"} {
	
        set gaugenum [gauge_gui_add "Capture Frames"]

	#Pause the camera
    	$camera(name) configure -pause 1
    	update
                
	# Capture the frames, one at a time
        for {set i 0} {$i < $frame(count)} {incr i} {
	    if {$gaugenum != 0} {gauge_gui_update $gaugenum [expr (double($i) / double($frame(count))) * 100.0]}
	    if {$gaugevar(cancel) != 0} {
		if {$gaugenum != 0} {gauge_gui_delete $gaugenum}
		set capture(status) NONE
		trees_update_status
		return;
	    }

	    puts "Working on capturing frame $i"

	    # Compute the center coordinates of the current frame, in pixels
	    set center_x [expr round($position(start-pixel,x) + ($i * $x_offset))]
	    set center_y [expr round($position(start-pixel,y) + ($i * $y_offset))]

	    # Move the stage so that the current frame is centered in the video window
	    move_stage_to $center_x $center_y
	    # Wait for the stage to stop moving
            set motionstatus "vague"
	    while {$motionstatus != "stopped"} {
                puts $motionstatus
		update
                set motionstatus [StageMotion_GetState]
	    }
	    
	    
	    set filename "${basename}${i}.SAD"
	    #$camera(name) configure -filename $filename
            
            set capture(frame$i,file) $filename
            if {$i > 0} {
                set capture(frame$i,estx) $x_offset
                set capture(frame$i,esty) [expr -$y_offset]
            } else {
                set capture(frame$i,estx) 0
                set capture(frame$i,esty) 0
            }



            # Perform Focus Test
	    global FocusTestVar
	    set FocusTestVar(nimg) 5
            global FocusAutoStepIncr
            set FocusAutoStepIncr 16
            FocusMotion_MoveDown [expr 3 * $FocusAutoStepIncr]
	    # Capture five frames at 4 focus step intervals, with the third frame using
	    #  the current focus setting
            while {[FocusMotion_GetState] != "stopped"} {
                update
            }
	    for {set j 0} {$j < $FocusTestVar(nimg)} {incr j} {
                FocusMotion_MoveUp $FocusAutoStepIncr
                while {[FocusMotion_GetState] != "stopped"} {
                    update
                }
                set FocusTestVar(inimg,addr[expr $j + 1]) [Sadie_Camera_Acquire 384 639 1 512 767 1]
            }
	    FocusMotion_MoveDown [expr 2 * $FocusAutoStepIncr]
            while {[FocusMotion_GetState] != "stopped"} {
                update
            }

	    # Find the best focused image out of the set of five
            set bestfocus [Sadie_Proto_FindBestFocus FocusTestVar]

	    # Delete the five focus test frames
	    for {set j 0} {$j < $FocusTestVar(nimg)} {incr j} {
            	#Sadie_FileIO_Save $focusimg($j) "/home/conner/test2/focus${j}.SAD" 32
                Sadie_General_Close $FocusTestVar(inimg,addr[expr $j + 1])
            }
	    unset FocusTestVar
            puts "bestfocus = $bestfocus (tcl)"

	    # Adjust the focus to the appropriate position
            if {$bestfocus < 2} {
                puts "Frame #$i -- Moving Focus Down by [expr (2 - $bestfocus)*$FocusAutoStepIncr] steps."
                FocusMotion_MoveDown [expr (2 - $bestfocus)*$FocusAutoStepIncr]
            } elseif {$bestfocus > 2} {
                puts "Frame #$i -- Moving Focus Up by [expr ($bestfocus - 2)*$FocusAutoStepIncr] steps."
                FocusMotion_MoveUp [expr ($bestfocus - 2)*$FocusAutoStepIncr]
            }
            unset FocusAutoStepIncr
            while {[FocusMotion_GetState] != "stopped"} {
                update
            }

	    # Capture the current frame
	    #  (Note:  this should perform frame averaging)
            after 2000
            set capturedframe [Sadie_Camera_Acquire 0 767 1 0 1023 1]
            
            puts "Saving image $i to disk as ${savedir}/${filename}"
            Sadie_FileIO_Save $capturedframe ${savedir}/${filename} 8
            puts "Closing image $i"
            Sadie_General_Close $capturedframe
            
            #update
    	}

	puts "Done with loop..."
        
        set capture(type) CAMERA
        set capture(magnification) 1.1
        set capture(markcount) NA
        set capture(framewidth) 1024
        set capture(frameheight) 768
        set capture(framecount) $frame(count)
	set capture(pixperstep,x) $motion(pixperstep,x)
	set capture(pixperstep,y) $motion(pixperstep,y)

	# Store the stage coordinates at the center of the first captured frame
	set capture(frame0,stagex) $position(start-pixel,x)
	set capture(frame0,stagey) $position(start-pixel,y)

	set choice [tk_messageBox -type yesno -default yes \
	    -message "Do you wish to adjust frames for illumination distortion?" \
	    -icon question]

	if {$choice == "yes"} {

	   set choice [tk_messageBox -type okcancel -default ok \
	       -message "Please place a grayscale or uniform white paper under the lens and refocus.  Press \"OK\" when ready." \
               -icon info]

           if {$choice == "ok"} {
               global IllumBack
               puts "Capturing background illumination image"
               set illum_background [Sadie_Camera_Blank  0 767 0 1023 256]
               Sadie_FileIO_Save $illum_background ${savedir}/${illumname} 8
               Sadie_General_Close $illum_background
               set illum_adjust 1
	       set choice [tk_messageBox -type okcancel -default ok \
		   -message "Please remove the uniform object. Press \"OK\" when you are ready to continue." \
		   -icon info]
           } else {
               set illum_adjust 0
           }

        } else {
            set illum_adjust 0
        }

        if {$illum_adjust} {
            set capture(illumination) $illumname
        } else {
            set capture(illumination) "NONE"
        }
            
        set capture(status) COMPLETE
	set stagemotion(capture,session) "OTHER"
        
        trees_update_status
	$camera(name) configure -loopcount 0
    	StageMotion_Close
    	FocusMotion_Close

        destroy .capturegui

        if {$gaugenum != 0} {gauge_gui_update $gaugenum 100}
        if {$gaugenum != 0} {gauge_gui_delete $gaugenum}

	#Resume capturing live images
        #$camera(name) configure -pause 0

    }
}


# Move the stage either up, down, right, or left by the current increment
proc move_stage {dir} {

    global motion position

    if {$dir == "left"} {
        move_stage_to [expr $position(stage-pixel,x) - $motion(increment)] $position(stage-pixel,y)
    } elseif {$dir == "right"} {
        move_stage_to [expr $position(stage-pixel,x) + $motion(increment)] $position(stage-pixel,y)
    } elseif {$dir == "up"} {
        move_stage_to $position(stage-pixel,x) [expr $position(stage-pixel,y) + $motion(increment)]
    } else {
        move_stage_to $position(stage-pixel,x) [expr $position(stage-pixel,y) - $motion(increment)]
    }
}


####
#
# move_stage_to  --  Procedure to move the stage to a particular location
#			and update the maps on the gui.  The input parameters
#			are specified in pixels.
#
####
proc move_stage_to {xpos_request ypos_request} {

    global motion position tracker_frame video_frame
    
    
    # Move is limited by the step size of the positioning table motors
    set position(stage-pixel,x) [request_x_pos $xpos_request]
    set position(stage-pixel,y) [request_y_pos $ypos_request]
    set position(stage-physical,x) [get_physical_x $position(stage-pixel,x)]
    set position(stage-physical,y) [get_physical_y $position(stage-pixel,y)]

    puts "moving stage to:            (after request ($xpos_request,$ypos_request))"
    puts "\tpixel coordinate: ($position(stage-pixel,x),$position(stage-pixel,y))   step coordinate: ($position(stage-physical,x),$position(stage-physical,y))"
    
    # Command the stage to move to a particular step location
    StageMotion_Move $position(stage-physical,x) $position(stage-physical,y)

    # Update the position of the start flag in the video canvas
    if {[.capturegui.live.canvas find withtag greenflag] != ""} {
    	.capturegui.live.canvas coords greenflag [expr round(($position(start-pixel,x) - $position(stage-pixel,x)) / $video_frame(deci)) + $video_frame(center,x)] [expr round((-$position(start-pixel,y) + $position(stage-pixel,y)) / $video_frame(deci)) + $video_frame(center,y)]
	}

    # Update the position of the stop flag in the video canvas
    if {[.capturegui.live.canvas find withtag redflag] != ""} {
        .capturegui.live.canvas coords redflag [expr round(($position(stop-pixel,x) - $position(stage-pixel,x)) / $video_frame(deci)) + $video_frame(center,x)] [expr round((-$position(stop-pixel,y) + $position(stage-pixel,y)) / $video_frame(deci)) + $video_frame(center,y)]
    }


    # Update the position of the symbolic live video window rectangle in the tracker canvas
    .capturegui.tracker.canvas coords camera_frame \
	    [expr round($position(stage-pixel,x) / $tracker_frame(deci)) + $tracker_frame(center,x) - round($tracker_frame(camera,width) / 2)] \
	    [expr round(-$position(stage-pixel,y) / $tracker_frame(deci)) + $tracker_frame(center,y) - round($tracker_frame(camera,height) / 2)] \
	    [expr round($position(stage-pixel,x) / $tracker_frame(deci)) + $tracker_frame(center,x) + round($tracker_frame(camera,width) / 2)] \
	    [expr round(-$position(stage-pixel,y) / $tracker_frame(deci)) + $tracker_frame(center,y) + round($tracker_frame(camera,height) / 2)]
    

    .capturegui.control.coord.physcoord configure -text "($position(stage-physical,x),$position(stage-physical,y))"
    .capturegui.control.coord.pixcoord configure -text "($position(stage-pixel,x),$position(stage-pixel,y))"
}


# Compute the closest valid pixel position from a requested pixel position
#  that is within the resolution of the stepper motors
proc request_x_pos {request} {
    global motion
    set actual [expr round(round($request / $motion(pixperstep,x)) * $motion(pixperstep,x))]
    return $actual
}
proc request_y_pos {request} {
    global motion
    set actual [expr round(round($request / $motion(pixperstep,y)) * $motion(pixperstep,y))]
    return $actual
}

# Convert a stage position specified in pixels to stepper motor units
proc get_physical_x {pixelpos} {
    global motion
    set physpos [expr round($pixelpos / $motion(pixperstep,x))]
    return $physpos
}
proc get_physical_y {pixelpos} {
    global motion
    set physpos [expr round($pixelpos / $motion(pixperstep,y))]
    return $physpos
}
