# This procedure is called when it is required to select
# a subsection of the mosaic for analysis.
# Creates binding on mouse events over canvas
proc trees_roi { var } {
    global roi zoom
    upvar #0 $var flag
    # Make note of cursor type to revert back when ROI selection is done
    set zoom(cursor) [.trees_main.imagepanel.canvas cget -cursor]

    # Create mouse binding on Button-1 click to initiate ROI
    bind .trees_main.imagepanel.canvas <Button-1> { trees_roi_line_begin %W %x %y }
    
    .trees_main.imagepanel.canvas config -cursor tcross

}

# This procedure is called when the user initiates drawing a line for ROI.
proc trees_roi_line_begin { w windowx windowy } {
    global roi zoom

    # Remove Button-1 mouse binding
    bind .trees_main.imagepanel.canvas <Button-1> {}

    # Create mouse bindings on B1-Motion & ButtonRelease-1 to track line
    bind .trees_main.imagepanel.canvas <B1-Motion> { trees_roi_line_update %W %x %y } 
    bind .trees_main.imagepanel.canvas <ButtonRelease-1> { trees_roi_line_end %W %x %y }

    set canv_x [$w canvasx $windowx]
    set canv_y [$w canvasy $windowy]

    # Check whether co-ordinates are within limits
    if { $canv_x < 0 } {
	set canv_x 0
    } elseif { $canv_x >= $zoom(canvasWidth) } {
	set canv_x [expr $zoom(canvasWidth) -1]
    }

    if { $canv_y < 0 } {
	set canv_y 0
    } elseif { $canv_y >= $zoom(canvasHeight) } {
	set canv_y [expr $zoom(canvasHeight) -1]
    }
    
    # Store the starting coordinates of the line.
    set roi(region,P1) [list [expr int($canv_x)] [expr int($canv_y)]]

    # Remove polygon drawn earlier
    catch {unset roi(region,last)}
    set roi(line,coords) $roi(region,P1)

}

# This procedure is called when the user drags the mouse over canvas.
# Updates the line on the canvas.
proc trees_roi_line_update { w windowx windowy } {
    global roi zoom

    set allowscroll 1
    set outofrange 0
    set canv_x [$w canvasx $windowx]
    set canv_y [$w canvasy $windowy]

    # Check whether co-ordinates are within limits
    if { $canv_x < 0  || $canv_x >= $zoom(canvasWidth)} {
	set allowscroll 0
	set outofrange 1
    }

    if { $canv_y < 0  || $canv_y >= $zoom(canvasHeight)} {
	set allowscroll 0
	set outofrange 1
    }

    if { $outofrange != 1 } {
	set roi(lastinrange,P2,x) $canv_x
	set roi(lastinrange,P2,y) $canv_y
    } else {
	set canv_x $roi(lastinrange,P2,x)
	set canv_y $roi(lastinrange,P2,y)
    }

    # Allow scrolling if co-ordinates are within range
    if { $allowscroll != 0 } {
	set fracX [.trees_main.imagepanel.xsbar get]
	set fracWidth [expr [lindex $fracX 1] - [lindex $fracX 0]]
	if { [expr $canv_x / $zoom(canvasWidth)] <= [lindex $fracX 0] } {
	    .trees_main.imagepanel.canvas xview moveto [expr $canv_x / $zoom(canvasWidth)]
	}
	if { [expr $canv_x / $zoom(canvasWidth)] >= [lindex $fracX 1] } {
	    .trees_main.imagepanel.canvas xview moveto [expr $canv_x / $zoom(canvasWidth) - $fracWidth]
	}
	
	set fracY [.trees_main.imagepanel.ysbar get]
	set fracHeight [expr [lindex $fracY 1] - [lindex $fracY 0]]
	if { [expr $canv_y / $zoom(canvasHeight)] <= [lindex $fracY 0] } {
	    .trees_main.imagepanel.canvas yview moveto [expr $canv_y / $zoom(canvasHeight)]
	}
	if { [expr $canv_y / $zoom(canvasHeight)] >= [lindex $fracY 1] } {
	    .trees_main.imagepanel.canvas yview moveto [expr $canv_y / $zoom(canvasHeight) - $fracHeight]
	}
    }
    
    # Delete previously drawn line
    catch {$w delete $roi(region,last)}
    set roi(region,last) [eval {$w create line} $roi(region,P1) {$canv_x $canv_y -fill yellow -width 1}]
}

# This procedure is called when the user releases mouse button
# to indicate end of line.
proc trees_roi_line_end { w windowx windowy } {
    global roi zoom

    # Remove mouse bindings on canvas
    bind .trees_main.imagepanel.canvas <B1-Motion> {} 
    bind .trees_main.imagepanel.canvas <ButtonRelease-1> {}
    
    # Create mouse bindings on canvas to track rectangular ROI.
    bind .trees_main.imagepanel.canvas <Motion> { trees_roi_region_track %W %x %y } 
    bind .trees_main.imagepanel.canvas <ButtonRelease-1> { trees_roi_region_end %W %x %y }
    
    set outofrange 0
    set canv_x [$w canvasx $windowx]
    set canv_y [$w canvasy $windowy]

    # Check whether co-ordinates are within limits
    if { $canv_x < 0 || $canv_x >= $zoom(canvasWidth)} {
	set outofrange 1
    }

    if { $canv_y < 0 || $canv_y >= $zoom(canvasHeight)} {
	set outofrange 1
    }

    if { $outofrange != 1 } {
	set roi(lastinrange,P2,x) $canv_x
	set roi(lastinrange,P2,y) $canv_y
    } else {
	set canv_x $roi(lastinrange,P2,x)
	set canv_y $roi(lastinrange,P2,y)
    }

    set roi(region,P2) [list [expr int($canv_x)] [expr int($canv_y)]]
    lappend roi(line,coords) [lindex $roi(region,P2) 0] [lindex $roi(region,P2) 1]


}

# This procedure tracks the region. Updates the rectangle on the canvas.
proc trees_roi_region_track { w windowx windowy } {
    global roi zoom

    set allowexpand 1

    set allowscroll 1
    set outofrange 0
    set canv_x [$w canvasx $windowx]
    set canv_y [$w canvasy $windowy]

    # Check whether co-ordinates are within limits
    if { $canv_x < 0 || $canv_x >= $zoom(canvasWidth)} {
	set allowscroll 0
	set outofrange 1
    }

    if { $canv_y < 0 || $canv_y >= $zoom(canvasHeight)} {
	set allowscroll 0
	set outofrange 1
    }

    if { $outofrange != 1 } {
	set roi(lastinrange,P3,x) $canv_x
	set roi(lastinrange,P3,y) $canv_y
    } else {
	set canv_x $roi(lastinrange,P3,x)
	set canv_y $roi(lastinrange,P3,y)
    }

    # Allow scrolling if co-ordinates are within range
    if { $allowscroll != 0 } {
	set fracX [.trees_main.imagepanel.xsbar get]
	set fracWidth [expr [lindex $fracX 1] - [lindex $fracX 0]]
	if { [expr $canv_x / $zoom(canvasWidth)] <= [lindex $fracX 0] } {
	    .trees_main.imagepanel.canvas xview moveto [expr $canv_x / $zoom(canvasWidth)]
	}
	if { [expr $canv_x / $zoom(canvasWidth)] >= [lindex $fracX 1] } {
	    .trees_main.imagepanel.canvas xview moveto [expr $canv_x / $zoom(canvasWidth) - $fracWidth]
	}
	
	set fracY [.trees_main.imagepanel.ysbar get]
	set fracHeight [expr [lindex $fracY 1] - [lindex $fracY 0]]
	if { [expr $canv_y / $zoom(canvasHeight)] <= [lindex $fracY 0] } {
	    .trees_main.imagepanel.canvas yview moveto [expr $canv_y / $zoom(canvasHeight)]
	}
	if { [expr $canv_y / $zoom(canvasHeight)] >= [lindex $fracY 1] } {
	    .trees_main.imagepanel.canvas yview moveto [expr $canv_y / $zoom(canvasHeight) - $fracHeight]
	}
    }

    set roi(region,coords) $roi(line,coords)
    set roi(region,P3) [list [expr int($canv_x)] [expr int($canv_y)]]

    set X1 [lindex $roi(region,P1) 0]
    set Y1 [lindex $roi(region,P1) 1]
    set X2 [lindex $roi(region,P2) 0]
    set Y2 [lindex $roi(region,P2) 1]
    set X3 [lindex $roi(region,P3) 0]
    set Y3 [lindex $roi(region,P3) 1]

    set m [expr double($Y2 - $Y1) / double($X2 - $X1)]
    set theta [expr atan($m)]
    set c1 [expr $Y1 - ($m * $X1)]
    set c2 [expr $Y3 - ($m * $X3)]
    set c [expr $c2 - $c1]
    set b [expr $c * cos($theta)]
    set dy [expr $b * cos($theta)]
    set dx [expr 0 - $b * sin($theta)]

    set X4 [expr int($X2 + $dx)]
    set Y4 [expr int($Y2 + $dy)]
    set X5 [expr int($X1 + $dx)]
    set Y5 [expr int($Y1 + $dy)]

    set outofrange 0
    
    # Check whether co-ordinates are within limits
    if { $X4 < 0 || $X4 >= $zoom(canvasWidth)} {
	set allowexpand 0
	set outofrange 1
    }
    
    if { $Y4 < 0 || $Y4 >= $zoom(canvasHeight)} {
	set allowexpand 0
	set outofrange 1
    }
    
    if { $X5 < 0 || $X5 >= $zoom(canvasWidth)} {
	set allowexpand 0
	set outofrange 1
    }
    
    if { $Y5 < 0 || $Y5 >= $zoom(canvasHeight)} {
	set allowexpand 0
	set outofrange 1
    }

    if { $outofrange != 1 } {
	set roi(lastinrange,P4,x) $X4
	set roi(lastinrange,P4,y) $Y4
	set roi(lastinrange,P5,x) $X5
	set roi(lastinrange,P5,y) $Y5
    } else {
	set X4 $roi(lastinrange,P4,x)
	set Y4 $roi(lastinrange,P4,y)
	set X5 $roi(lastinrange,P5,x)
	set Y5 $roi(lastinrange,P5,y)
    }
   
    # Draw polygon only if co-ordinates stay within limits
    if { $allowexpand != 0 } {
	set roi(region,P4) [list [expr int($X4)] [expr int($Y4)]]
	set roi(region,P5) [list [expr int($X5)] [expr int($Y5)]]
	
	lappend roi(region,coords) [lindex $roi(region,P4) 0] [lindex $roi(region,P4) 1]
	lappend roi(region,coords) [lindex $roi(region,P5) 0] [lindex $roi(region,P5) 1]
	lappend roi(region,coords) [lindex $roi(region,P1) 0] [lindex $roi(region,P1) 1]
    
	catch {$w delete $roi(region,last)}
	set roi(region,last) [eval {$w create line} $roi(region,coords) {-fill yellow -width 1}]
    }
   
}

# This procedure is called when the analyst releases the mouse button
# to indicate end of selection
proc trees_roi_region_end { w windowx windowy } {
    global roi labels mosaic zoom session findrings
    global flag
    # Remove mouse bindings
    bind .trees_main.imagepanel.canvas <Motion> {} 
    bind .trees_main.imagepanel.canvas <ButtonRelease-1> {}

    set roi(region,P3) [list $roi(lastinrange,P3,x) $roi(lastinrange,P3,y)]

    set roi(region,P4) [list $roi(lastinrange,P4,x) $roi(lastinrange,P4,y)]
    set roi(region,P5) [list $roi(lastinrange,P5,x) $roi(lastinrange,P5,y)]

    # Prompt user to confirm ROI
    set answer [tk_dialog .dlg "Confirm ROI" \
	    "Do you wish to continue with the selected region?" \
	    question \
	    2 "Skip" "Reselect" $labels(button,yes) ]

    if { $answer == 0 } {
	catch {$w delete $roi(region,last)}
	set roi(selected) 1
    } elseif { $answer == 1 } {
	catch {$w delete $roi(region,last)}
	trees_roi $flag
    } else {
	catch {$w delete $roi(region,last)}
	
	if {![info exists mosaic(loaded,fullresol,index)]} {
	    set mosaic(loaded,fullresol,index) [Sadie_TREES_MosaicIndexOpen\
		    $session(directory)/images/mosaic/$mosaic(file,fullresol,index)]
	}

	set zoombuffer [expr $zoom(default)/$findrings(ring,resolution)]
	
	if {$flag==1} {
        Sadie_TREES_UpdateROI\
		$mosaic(loaded,fullresol,index) $zoombuffer\
		[lindex $roi(region,P1) 0] [lindex $roi(region,P1) 1]\
		[lindex $roi(region,P2) 0] [lindex $roi(region,P2) 1]\
		[lindex $roi(region,P4) 0] [lindex $roi(region,P4) 1]\
		[lindex $roi(region,P5) 0] [lindex $roi(region,P5) 1] $flag

	Sadie_TREES_MosaicIndexSave\
		$mosaic(loaded,fullresol,index)\
		$session(directory)/images/mosaic/$mosaic(file,fullresol,index)

	after 5
	set roi(selected) 1
     } elseif {$flag==0} {
    # segment ROI selection   
     Sadie_TREES_UpdateROI\
		$mosaic(loaded,fullresol,index) $zoombuffer\
		[lindex $roi(region,P1) 0] [lindex $roi(region,P1) 1]\
		[lindex $roi(region,P2) 0] [lindex $roi(region,P2) 1]\
		[lindex $roi(region,P4) 0] [lindex $roi(region,P4) 1]\
		[lindex $roi(region,P5) 0] [lindex $roi(region,P5) 1] $flag 
        after 5
	set roi(selected) 1      
       }      
 
    }

    # Revert back to previous cursor type
    .trees_main.imagepanel.canvas config -cursor $zoom(cursor)

}


