/** @file bytedisk2img.c
 * Reading single-byte images from external SADIE files.
 * The DISK2IMG  function in libsadie can read external data in SADIE
 * format with the value of each pixel encoded in a single byte, but
 * it cannot handle the similar internal representation, the IMAGE_BYTE
 * structure, so something more specialized is needed.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: bytedisk2img.c,v 1.4 2004/08/06 22:47:11 mmunro Exp $
 */

/* This file is part of TREES.

   TREES is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   TREES is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with TREES; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <sadie.h>
#include "sadie_byte.h"

#ifdef WORDS_BIGENDIAN
#define NEEDSWAP 0
#else
#define NEEDSWAP 1
#endif

/**
 * Read a byte image header structure from a disk file (or equivalent).
 * We play safe by reading each member of the structure independently,
 * in case there are any alignment issues when trying to read the structure
 * as a whole.
 * @param *buf  points at the buffer into which to read the header.
 * @param *fp   file pointer to something already open for reading.
 * @return An error code on failure, 0 on success.
 */
static int
sad_byte_header_read (IMAGE_BYTE *buf, FILE *fp)
{
  size_t n_read;
  
  if ((buf == NULL) || (fp == NULL))
    return -1;
  n_read = fread (&buf->fileversion, sizeof (buf->fileversion), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->application, sizeof (buf->application), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->appversion, sizeof (buf->appversion), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->dataformat, sizeof (buf->dataformat), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->datapacking, sizeof (buf->datapacking), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->datarotation, sizeof (buf->datarotation), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->npix, sizeof (buf->npix), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->nlin, sizeof (buf->nlin), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->nbnd, sizeof (buf->nbnd), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->ntvl, sizeof (buf->ntvl), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->nbit, sizeof (buf->nbit), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->gmin, sizeof (buf->gmin), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->gmax, sizeof (buf->gmax), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->text, sizeof (buf->text), 1, fp);
  if (n_read != 1)
    return -1;
  if (NEEDSWAP)
    {
      SWAP2 ((BYTE *)&buf->fileversion);
      SWAP2 ((BYTE *)&buf->application);
      SWAP2 ((BYTE *)&buf->appversion);
      SWAP2 ((BYTE *)&buf->dataformat);
      SWAP2 ((BYTE *)&buf->datapacking);
      SWAP4 ((BYTE *)&buf->npix);
      SWAP4 ((BYTE *)&buf->nlin);
      SWAP4 ((BYTE *)&buf->nbnd);
      SWAP4 ((BYTE *)&buf->ntvl);
      SWAP4 ((BYTE *)&buf->nbit);
      SWAP4 ((BYTE *)&buf->gmin);
      SWAP4 ((BYTE *)&buf->gmax);
    }
  if ((buf->dataformat != EXTDATA_UINT8)
      || (buf->datapacking != BE_BYTEPACKED)
      || (buf->nbit != SBPB))
    return -1;
  else
    return 0;
}

/**
 * Read a byte image header structure from the old disk representation.
 * Any version of TREES before 2.0.1 stored byte images in a representation
 * that differed from the current SADIE image format: this reads the
 * old format and converts it to the current internal representation.
 * We play safe by reading each member of the structure independently,
 * in case there are any alignment issues when trying to read the structure
 * as a whole.
 * @param *buf  points at the buffer into which to read the header.
 * @param *fp   file pointer to something already open for reading.
 * @return An error code on failure, 0 on success.
 */
static int
old_byte_header_read (IMAGE_BYTE *buf, FILE *fp)
{
  size_t n_read;
  uint16_t old_lines, old_pixperline, old_bits;
  char old_text[496];
  PIXEL gmin, gmax;

  if ((buf == NULL) || (fp == NULL))
    return -1;
  buf->nbnd = buf->fileversion;
  buf->fileversion = FILE_VERSION;
  buf->application = Application;
  buf->appversion = AppVersion;
  buf->dataformat = EXTDATA_UINT8;
  buf->datapacking = BE_BYTEPACKED;
  buf->datarotation[0] = PIXEL_DIMENSION    | VALUES_NORMAL;
  buf->datarotation[1] = LINE_DIMENSION     | VALUES_NORMAL;
  buf->datarotation[2] = BAND_DIMENSION     | VALUES_NORMAL;
  buf->datarotation[3] = INTERVAL_DIMENSION | VALUES_NORMAL;
  buf->ntvl = 1;
  buf->nbit = SBPB;
  n_read = fread (&old_lines, sizeof (old_lines), 1, fp);
  if (n_read != 1)
    return -1;
  buf->nlin = old_lines;
  n_read = fread (&old_pixperline, sizeof (old_pixperline), 1, fp);
  if (n_read != 1)
    return -1;
  buf->npix = old_pixperline;
  n_read = fread (&old_bits, sizeof (old_bits), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->gmin, sizeof (buf->gmin), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&buf->gmax, sizeof (buf->gmax), 1, fp);
  if (n_read != 1)
    return -1;
  n_read = fread (&old_text, sizeof (old_text), 1, fp);
  if (n_read != 1)
    return -1;
  if (NEEDSWAP)
    {
      SWAP4 ((BYTE *)&buf->npix);
      SWAP4 ((BYTE *)&buf->nlin);
      SWAP4 ((BYTE *)&buf->nbnd);
      SWAP4 ((BYTE *)&buf->gmin);
      SWAP4 ((BYTE *)&buf->gmax);
    }
  return 0;
}

/**
 * Read the image data themselves as an undifferentiated slab of bytes.
 * The image header defining the size and other image attributes should
 * already be in a buffer; we create a new image with these attributes,
 * copy the old header into it without overwriting the pointer to the
 * image data storage area, then read the image data themselves.
 * @param *buf  points at the buffer holding the image header.
 * @param **newimage  the complete newly read image, or NULL on failure.
 * @param *fp   file pointer to something already open for reading.
 */
static int
byte_data_read (IMAGE_BYTE *buf, IMAGE_BYTE **newimg, FILE *fp)
{
  BYTE ***new_data;
  size_t n_read;

  if ((buf == NULL) || (fp == NULL))
    return -1;
  GETMEM_BYTE (buf->nbnd, buf->nlin, buf->npix, newimg);
  if (!CHECKIMG_BYTE (*newimg))
    {
      *newimg = NULL;
      return -1;
    }
  new_data = (*newimg)->data;
  **newimg = *buf;
  (*newimg)->data = new_data;
  n_read = fread ((*newimg)->data[0][0],
                  (buf->nbnd * buf->nlin * buf->npix), 1, fp);
  if (n_read == 1)
    return 0;
  else
    {
      RELMEM_BYTE (*newimg);
      *newimg = NULL;
      return -1;
    }
}

/**
 * Read an image in the TREES-specific format that has only one byte per pixel.
 * The name should be something like the path to a disk file holding
 * the byte image.  The handle defining where to put the newly read image
 * should not point to a pointer to an image that has already been created
 * but not freed: it will always be overwritten.
 * @param *name  the path name from which to read the image.
 * @param **out  the complete newly read image, or NULL on failure.
 */
void
BYTEDISK2IMG (const char *name, IMAGE_BYTE **out)
{
  IMAGE_BYTE inbuf;
  size_t n_read;
  int err;
  char msg[SLEN];
  FILE *fp;

  memset (msg, 0, SLEN);
  if (out == NULL)
    {
      snprintf (msg, SLEN - 1, " Invalid location to read byte image.");
      MESSAGE ('E', msg);
      return;
    }
  *out = NULL;
  if (name == NULL)
    {
      snprintf (msg, SLEN - 1, " No byte image file path specified.");
      MESSAGE ('E', msg);
      return;
    }
  fp = fopen (name, "r");
  if (fp == NULL)
    {
      snprintf (msg, SLEN - 1, " Can't open byte image %s.", name);
      MESSAGE ('E', msg);
      return;
    }
  n_read = fread (&inbuf.fileformat, sizeof (inbuf.fileformat), 1, fp);
  err = (n_read != 1);
  if (!err)
    {
      if (NEEDSWAP)
        SWAP2 ((BYTE *)&inbuf.fileformat);
      if (inbuf.fileformat == FILE_FORMAT)
        err = sad_byte_header_read (&inbuf, fp);
      else
        err = old_byte_header_read (&inbuf, fp);
    }
  if (err)
    {
      snprintf (msg, SLEN - 1, " Can't read byte image header from %s.", name);
      MESSAGE ('E', msg);
    }
  else if (err = byte_data_read (&inbuf, out, fp))
    {
      snprintf (msg, SLEN - 1, " Can't read byte image data from %s.", name);
      MESSAGE ('E', msg);
    }
  else
    {
      (*out)->text[TLEN - 1] = 0;
      strncpy ((*out)->text, name, TLEN - 1);
    }
  err = fclose(fp);
  if (err)
    {
      snprintf (msg, SLEN - 1, " Can't close byte image %s.", name);
      MESSAGE ('W', msg);
    }
}
