/** @file byteimg2disk.c
 * Writing single-byte images to external SADIE files.
 * The IMG2DSK  function in libsadie can write external data in SADIE
 * format with the value of each pixel encoded in a single byte, but
 * it cannot handle the similar internal representation, the IMAGE_BYTE
 * structure, so something more specialized is needed.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: byteimg2disk.c,v 1.4 2004/08/06 22:47:11 mmunro Exp $
 */

/* This file is part of TREES.

   TREES is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   TREES is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with TREES; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <sadie.h>
#include "sadie_byte.h"

#ifdef WORDS_BIGENDIAN
#define NEEDSWAP 0
#else
#define NEEDSWAP 1
#endif

/**
 * Write a byte image header structure to a disk file (or equivalent).
 * We play safe by writing each member of the structure independently,
 * in case there are any alignment issues when trying to write the structure
 * as a whole.
 * @param *buf  points at the buffer containing the image header.
 * @param *fp   file pointer to something already open for writing.
 * @return An error code on failure, 0 on success.
 */
int
sad_byte_header_write (IMAGE_BYTE *buf, FILE *fp)
{
  size_t n_written;
  
  if ((buf == NULL) || (fp == NULL))
    return -1;
  if (NEEDSWAP)
    {
      SWAP2 ((BYTE *)&buf->fileformat);
      SWAP2 ((BYTE *)&buf->fileversion);
      SWAP2 ((BYTE *)&buf->application);
      SWAP2 ((BYTE *)&buf->appversion);
      SWAP2 ((BYTE *)&buf->dataformat);
      SWAP2 ((BYTE *)&buf->datapacking);
      SWAP4 ((BYTE *)&buf->npix);
      SWAP4 ((BYTE *)&buf->nlin);
      SWAP4 ((BYTE *)&buf->nbnd);
      SWAP4 ((BYTE *)&buf->ntvl);
      SWAP4 ((BYTE *)&buf->nbit);
      SWAP4 ((BYTE *)&buf->gmin);
      SWAP4 ((BYTE *)&buf->gmax);
    }
  n_written = fwrite (&buf->fileformat, sizeof (buf->fileformat), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->fileversion, sizeof (buf->fileversion), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->application, sizeof (buf->application), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->appversion, sizeof (buf->appversion), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->dataformat, sizeof (buf->dataformat), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->datapacking, sizeof (buf->datapacking), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->datarotation, sizeof (buf->datarotation), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->npix, sizeof (buf->npix), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->nlin, sizeof (buf->nlin), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->nbnd, sizeof (buf->nbnd), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->ntvl, sizeof (buf->ntvl), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->nbit, sizeof (buf->nbit), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->gmin, sizeof (buf->gmin), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->gmax, sizeof (buf->gmax), 1, fp);
  if (n_written != 1)
    return -1;
  n_written = fwrite (&buf->text, sizeof (buf->text), 1, fp);
  if (n_written != 1)
    return -1;
  return 0;
}

/**
 * Write an image in the TREES-specific format that has only one byte per pixel.
 * The name should be something like the path to a disk file to receive
 * the byte image.  Currently it cannot handle regular SADIE images.
 * @param *in    an existing image (already in byte image format).
 * @param *name  the path name from which to read the image.
 */
void
BYTEIMG2DISK (IMAGE_BYTE *in, const char *name)
{
  char msg[SLEN];
  FILE *fp;
  int err;
  size_t n_written;
  IMAGE_BYTE outbuf;

  memset (msg, 0, SLEN);
  if (name == NULL)
    {
      snprintf (msg, SLEN - 1, " No byte image file path specified.");
      MESSAGE ('E', msg);
      return;
    }
  if (!CHECKIMG_BYTE(in)
      || (in->fileversion != FILE_VERSION) || (in->dataformat != EXTDATA_UINT8)
      || (in->datapacking != BE_BYTEPACKED) || (in->nbit != SBPB))
    {
      snprintf (msg, SLEN - 1, " Trying to write invalid byte image to %s.", name);
      MESSAGE ('E', msg);
      return;
    }
  fp = fopen (name, "w");
  if (fp == NULL)
    {
      snprintf (msg, SLEN - 1, " Can't open byte image %s.", name);
      MESSAGE ('E', msg);
      return;
    }
  outbuf = *in;
  /* Change image name to match disk name */
  memset (outbuf.text, 0, TLEN);
  strncpy (outbuf.text, name, TLEN-1);
  err = sad_byte_header_write (&outbuf, fp);
  if (err)
    {
      snprintf (msg, SLEN - 1, " Can't write byte image header to %s.", name);
      MESSAGE ('E', msg);
    }
  else
    {
      n_written = fwrite (in->data[0][0],
                          (in->nbnd * in->nlin * in->npix), 1, fp);
      if (n_written != 1)
        {
          snprintf (msg, SLEN - 1,
                    " Can't write byte image data to %s.", name);
          MESSAGE ('E', msg);
        }
    }
  err = fclose(fp);
  if (err)
    {
      snprintf (msg, SLEN - 1, " Can't close byte image %s.", name);
      MESSAGE ('W', msg);
    }
}
