// Wrapper for the IEEE1394 digital camera library.  In addition to
// calls to the library, the wrapper maintains its own decimated
// version of the current camera image, which may be any size less
// than or equal to that of the camera image, and may be resized
// (although the aspect ratio is held constant).

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include "cameradevice.h"
#include <sadie.h>

namespace std
{

  static const int PORTNUM = 0;
  static const int DESCRIBE_CAMERAS = 0;
  static const int ISO_CHANNEL = 0;
  static const int DROP_FRAMES = 1;
  static const int MAX_W = 1024;
  static const int MAX_H = 768;
  static const char* DEVICE_NAME = 0;

  Cameradevice::Cameradevice(int unit, int picture_timeout, int numbufs,
                             unsigned int h, unsigned int v,
                             unsigned int dn, unsigned int dd)
  {
    this->camdevprogres_type = 0;
    this->viewbuf = 0;
    this->fullbuf = 0;
    this->incr    = 0;
    this->hoff = h;
    this->voff = v;
    this->nbufs = numbufs;
    if (this->hoff >= MAX_W || this->voff > MAX_H)
      {
        cerr << "image offset too far: exiting" << endl;
        exit(1);
      }
    // Create a raw1394_handle for just one port.
    this->rawhandle = dc1394_create_handle(PORTNUM);
    if (!this->rawhandle)
      {
        cerr << "Cannot acquire raw1394 handle (check ieee1394 kernel modules and devices?): exiting" << endl;
        exit(1);
      }
    /* Count the total number of nodes (of all types) on the IEEE 1394 bus. */
    int nnodes = raw1394_get_nodecount(this->rawhandle);
    int ncameras;
    /* Traverse the bus node list, selecting just the cameras. */
    nodeid_t* camera_nodes = dc1394_get_camera_nodes(this->rawhandle, &ncameras, DESCRIBE_CAMERAS);
    if (ncameras < 1)
      {
        cerr << "Cannot find any cameras: exiting" << endl;
        exit(1);
      }
    if (unit < 0 || unit >= ncameras)
      unit = 0;
    if (camera_nodes[unit] == nnodes-1)
      {
        cerr << "Cameras is root node! iso transfers broken: exiting" << endl;
        exit(1);
      }
    this->dc = new(dc1394_cameracapture);
    this->multibuffer = (numbufs > 1);
    if ((multibuffer ?
         dc1394_dma_setup_capture(this->rawhandle,
                                  camera_nodes[unit],
                                  ISO_CHANNEL, 
                                  FORMAT_SVGA_NONCOMPRESSED_1,
                                  MODE_1024x768_MONO,
                                  SPEED_400,
                                  FRAMERATE_7_5,
                                  numbufs,
                                  0,
                                  DROP_FRAMES,
                                  DEVICE_NAME,
                                  this->dc) :
         dc1394_setup_capture(this->rawhandle,
                              camera_nodes[unit],
                              ISO_CHANNEL, 
                              FORMAT_SVGA_NONCOMPRESSED_1,
                              MODE_1024x768_MONO,
                              SPEED_400,
                              FRAMERATE_7_5,
                              this->dc))       
        != DC1394_SUCCESS)
      {
        cerr << "Unable to setup camera with requested format, mode, and frame rate: exiting" << endl;
        if (this->multibuffer)
          dc1394_dma_release_camera(this->rawhandle, this->dc);
        else
          dc1394_release_camera(this->rawhandle, this->dc);
        raw1394_destroy_handle(this->rawhandle);
        exit(1);
      }
    if (dc1394_set_trigger_mode(this->rawhandle, this->dc->node, TRIGGER_MODE_0)
        != DC1394_SUCCESS)
      {
        cerr << "Unable to set camera trigger mode: exiting" << endl;
        if (this->multibuffer)
          dc1394_dma_release_camera(this->rawhandle, this->dc);
        else
          dc1394_release_camera(this->rawhandle, this->dc);
        raw1394_destroy_handle(this->rawhandle);
        exit(1);
      }
    dc1394_feature_set* features = new(dc1394_feature_set);
    if (dc1394_get_camera_feature_set(this->rawhandle, this->dc->node, features) != DC1394_SUCCESS)
      {
        cerr << "cannot find camera features: exiting" << endl;
        exit(1);
      }     
    else
      dc1394_print_feature_set(features);
    delete features;
    if (dc1394_start_iso_transmission(this->rawhandle, this->dc->node)
        != DC1394_SUCCESS) 
      {
        cerr << "Unable to start camera iso transmission: exiting" << endl;
        if (this->multibuffer)
          dc1394_dma_release_camera(this->rawhandle, this->dc);
        else
          dc1394_release_camera(this->rawhandle, this->dc);
        raw1394_destroy_handle(this->rawhandle);
        exit(1);
      }
    // set bounds for the decimated image
    horiz_size = dn*(this->dc->frame_width - this->hoff) / dd;
    vert_size  = dn*(this->dc->frame_height - this->voff) / dd;
    if (!resize(this->horiz_size, this->vert_size))
      {
        cerr << "can't acquire the image sized "
             << this->dc->frame_width << " X " << this->dc->frame_height << ": exiting" << endl;
        exit(1);
      }
    if (this->multibuffer)
      {
        if (dc1394_dma_single_capture(this->dc) != DC1394_SUCCESS)
          {
            cerr << "can't start acquiring the first dma image: exiting" << endl;
            exit(1);
          }
      }
    else
      {
        if (dc1394_single_capture(this->rawhandle, this->dc) != DC1394_SUCCESS)
          {
            cerr << "can't start acquiring the first image: exiting" << endl;
            exit(1);
          }
      }      
  }

  Cameradevice::~Cameradevice()
  {
    if (dc1394_stop_iso_transmission(this->rawhandle, this->dc->node) != DC1394_SUCCESS)
      cerr << "warning: cannot stop camera" << endl;;
    if (multibuffer)
      {
        dc1394_dma_release_camera(this->rawhandle, this->dc);
        dc1394_dma_unlisten(this->rawhandle, this->dc);
      }
    else
      dc1394_release_camera(this->rawhandle, this->dc);
    raw1394_destroy_handle(this->rawhandle);
    delete this->dc;
    delete [] this->viewbuf;
    delete [] this->fullbuf;
    delete [] this->incr;
  }

  unsigned char*
  Cameradevice::snapshot()
    // if we have more than one buffer then get the next image
    // buffer, start a new acquire, then process the data we
    // just got. We do this so that we can process images in parallel
    // with the acquire of the next image.
  {
    if (!this->fullbuf)
      {
        if (this->dc->frame_width > 0 && this->dc->frame_height > 0)
          this->fullbuf = new unsigned char[this->snap_width()*this->snap_height()];
        else
          return this->fullbuf;
      }
    unsigned char* livebuf;
    if (!multibuffer && (dc1394_single_capture(this->rawhandle, this->dc) != DC1394_SUCCESS))
      {
        cerr << "can't acquire an image" << endl;
        exit(1);
      }     
    livebuf = (unsigned char*)this->dc->capture_buffer;
    unsigned char* origin = livebuf + this->hoff + this->dc->frame_width * this->voff;
    unsigned char* dst = this->fullbuf;
    unsigned int linecount = this->snap_height();
    while (linecount--)
      {
        const unsigned char * maxsrc = origin + (this->snap_width() - 1);
        for (unsigned char * src = origin; src <= maxsrc; src++)
          *dst++ = *src;
        origin += this->dc->frame_width;
      }
    if (multibuffer)
      {
        dc1394_dma_done_with_buffer(this->dc);
        if (dc1394_dma_single_capture(this->dc) != DC1394_SUCCESS)
          {
            cerr << "can't start acquiring the next image" << endl;
            exit(1);
          }     
      }
    return fullbuf;
  }

  unsigned char*
  Cameradevice::acquire()
    // if we have more than one buffer then get the next image
    // buffer, start a new acquire, then process the data we
    // just got. We do this so that we can process images in parallel
  {
    if (!this->viewbuf)
      return this->viewbuf;
    unsigned char* livebuf;
    if (!multibuffer && (dc1394_single_capture(this->rawhandle, this->dc) != DC1394_SUCCESS))
      {
        cerr << "can't acquire an image" << endl;
        exit(1);
      }     
    livebuf = (unsigned char*)this->dc->capture_buffer;
    unsigned char* origin = livebuf + this->hoff + this->dc->frame_width * this->voff;
    unsigned char* next = origin;
    unsigned char* dst = this->viewbuf;;
    unsigned int linecount = this->vert_size;
    int j = 0;
    while (linecount--)
      {
        if (j)
          next += (this->dc->frame_width * this->incr[j--]);
        else
          {
            next = origin;
            origin += (this->dc->frame_width * this->skipsize);
            j = this->dsize;
          }
        emitline(next, dst);
      }
    if (this->multibuffer)
      {
        dc1394_dma_done_with_buffer(this->dc);
        if (dc1394_dma_single_capture(this->dc) != DC1394_SUCCESS)
          {
            cerr << "can't start acquiring the next image" << endl;
            exit(1);
          }     
      }
    return viewbuf;
  }

  void
  Cameradevice::emitline(unsigned char* startaddr, unsigned char*& dst)
  {
    unsigned char* baddr = startaddr;
    unsigned char* src = baddr;
    int i = 0;
    unsigned int colcount = this->horiz_size;
    while (colcount--)
      {
        if (i) 
          src += this->incr[i--];
        else
          {
            src = baddr;
            baddr += this->skipsize;
            i = this->dsize;
          }
        *dst++ = *src;
      }
  }

  bool
  Cameradevice::resize(unsigned int h_req, unsigned int v_req)
  {
    if (!h_req || !v_req || this->dc->frame_width <= 0 || this->dc->frame_height <= 0)
      return false;
    unsigned int dd = this->dc->frame_width - this->hoff;      // decimation denominator
    unsigned int v_sc = int((this->dc->frame_height - this->voff) * double(h_req) / double(dd));
    if (v_sc > v_req)
      {
        horiz_size = int(dd * double(v_req) / double(this->dc->frame_height - voff));
        vert_size  = v_req;
      }
    else
      {
        horiz_size = h_req;
        vert_size  = v_sc;
      }
    unsigned int m = dd;            // DEK ``The Art...'' Book 1, p. 2,
    unsigned int n = this->horiz_size;
    unsigned int r = m % n;
    while (r)
      {
        m = n;
        n = r;
        r = m % n;
      }
    dd /= n;
    unsigned int dn = this->horiz_size / n;  // decimation keeps dn/dd pixels
    delete [] this->incr;
    incr = 0;
    this->skipsize = dd;
    this->dsize = dn - 1;
    this->incr = new int[dn];
    *this->incr = 0;
    int last = *incr;
    for (int i = dsize; i != 0; i--) 
      {
        int current = (int)rint(double(dd*(dn - i)) / double(dn));
        this->incr[i] = current - last;
        last = current;
      }
    delete [] this->viewbuf;
    this->viewbuf = new unsigned char[horiz_size * vert_size];
    return true;
  }

  IMAGE*
  Cameradevice::acquire_sadie(int xbgn, int ybgn, int xend, int yend)
  {

    IMAGE* liveimg = 0;
    unsigned char * livebuf = 0;
    int j, k;
    int nlin, npix;

    if ((xbgn < 0) || (xbgn > xend) ||
        (ybgn < 0) || (ybgn > yend) ||
        (xend >= (this->dc->frame_width  - this->hoff)) ||
        (yend >= (this->dc->frame_height - this->voff)))
      {
        cerr << "Error: coordinates out of range!" << endl;
        return liveimg;
      }
    nlin = 1 + yend - ybgn;
    npix = 1 + xend - xbgn;
    if (this->multibuffer)
      {
        dc1394_dma_done_with_buffer(this->dc);
        if (dc1394_dma_single_capture(this->dc) != DC1394_SUCCESS)
          {
            cerr << "can't start acquiring the next image" << endl;
            exit(1);
          }     
      }
    else
      dc1394_single_capture(this->rawhandle, this->dc);
    livebuf = (unsigned char *)this->dc->capture_buffer;
    if (!CHECKIMG(liveimg))
      GETMEM(1, nlin, npix, &liveimg);
    if (!liveimg)
      {
        cerr << "Error: creating SADIE image!" << endl;
        return liveimg;
      }
    for (j=0; j<nlin; j++)
      {
        for (k=0; k<npix; k++)
          liveimg->data[0][j][k] =
            (PIXEL) livebuf[(k+this->hoff+xbgn) +
                            ((this->dc->frame_width)*(j+this->voff+ybgn))];
      }
    return liveimg;
  }

  unsigned int
  Cameradevice::size()
  {
    unsigned int s = 0;
    if (this->rawhandle && this->dc)
      dc1394_get_capture_size(this->rawhandle, this->dc->node, &s);
    return s;
  }

  int
  Cameradevice::enable_live(int flag)
  {
    return 0;
  }


  int
  Cameradevice::set_gain(int gain)
  {
    dc1394_set_gain(this->rawhandle, this->dc->node, gain);
    return 0;
  }

} // namespace std
