/** @file chain_provisional.c
 * Interpolated measurements for insertion and removal of ring boundaries.
 * This provides several operations to support the operations that add
 * or remove ring boundaries.  They generate new boundaries by either
 * interpolating between a pair of previously detected boundaries
 * close to a specified position (e.g., a mouse click on an image), or
 * by taking the long axis of the region of interest within which the
 * other boundaries lie and placing a copy of one of these previously
 * detected boundaries at a position along this axis close to the
 * secified position.  We use the Sadie_Index hash table
 * implementation to store the measurements generated or modified by
 * the boundary insertions and deletions, and provide a complete range
 * of functions to retrieve these data while hiding the messy details
 * of the implementation.
 * 
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: chain_provisional.c,v 1.4 2004/11/19 19:11:11 mmunro Exp $
 */

/* This file is part of TREES.

   TREES is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   TREES is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with TREES; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <sadie.h>
#include <math.h>
#include "junkmath.h"
#include "Sadie_Index.h"
#include "logerr.h"
#include "chain_provisional.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** Pointer to a container of messy easurement and boundary data. */
typedef struct Provisionalrings *provisionalringsp_t;

/** Pointer to the pairing of a point and a boundary. */
typedef struct Boundbracket *boundbracketp_t;

/** Pointer to the summary of the ROI coordinate frame. */
typedef struct Roiframe *roiframep_t;

/**
 * Measurements for rings affected by an added or removed boundary (2 at most).
 */
struct Provisionalrings
{
  int prov_id;  /**< provisional boundary sequence number. */
  int insert_id; /**< id number of the item just before the insert point. */
  int n;  /**< number of rings affected. */
  LIST_NODE *fullchain;  /**< full-resolution chaincoded boundary. */
  LIST_NODE *smallchain;  /**< low-resolution chaincoded boundary. */
  BOUNDBOX_NODE box;  /**< bounding box for the boundary (full resolution). */
  PROFILE_LIST prof[2];  /**< brightness profiles for the modified rings. */
  PROFILE_LIST diffprof[2];  /**< differenced profiles for the rings. */
  double totalwidth[2];  /**< total ring widths. */
  double earlywidth[2];  /**< earlywood widths. */
  double latewidth[2];   /**< latewood widths. */
};

/**
 * Used for interpolating new points between old points and boundaries.
 * The @a curr_point is the current old point from which to drop perpendiculars
 * onto the boundary, but may also be the head of a list of subsequent points;
 * the perpendiculars will determine the locations of the new interpolated
 * points.
 */
struct Boundbracket
{
  BOUNDPOINT *curr_point;   /**< start of lines along which to interpolate. */
  BOUNDER *ref_bound;   /**< the lines end on this boundary. */
  double relpos;  /**< relative position along the lines for new points. */
  double d_min; /**< minimum absolute point-to-boundary distance. */
  int nperp; /**< number of point-to-boundary perpendiculars. */
};

/**
 * Holds a summary of the region of interest coordinate frame.
 * This derives from the ROI in the mosaic index, but uses doubles for the
 * coordinates, and has the most commonly used transformation pre-applied.
 */
struct Roiframe
{
  double orig_x;  /**< coordinate of the origin of the ROI coordinate frame. */
  double orig_y;  /**< coordinate of the origin of the ROI coordinate frame. */
  double x_len;  /**< length of the frame along its x axis. */
  double y_len;  /**< length of the frame along its y axis. */
  double phi;  /**< x axis angle, relative to the mosaic coordinates. */
};

/**
 * Hash table storing arbitrary data (including Provisionalrings structures).
 */
extern sad_doclistp_t global_doclist;

/**
 * Create a new structure to hold modified measurements and boundaries.
 * @n_rings  the number of modified sets of measurements.
 * @provbnum  the provisional boundary number.
 * @insert_after  the insertion point in the chain coded boundary list.
 * @return  a pointer to the new structure, with its fields all initially zero.
 */
static provisionalringsp_t
new_provisionalrings (int n_rings, int provbnum, LIST_NODE * insert_after)
{
  provisionalringsp_t newrings;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n_rings <= 2));
  newrings = malloc (sizeof (struct Provisionalrings));
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newrings != NULL));
  newrings->prov_id = provbnum;
  newrings->insert_id = (insert_after) ? insert_after->ring_id : INVALID_ID;
  newrings->n = n_rings;
  newrings->fullchain = NULL;
  newrings->smallchain = NULL;
  newrings->prof[0] = NULL;
  newrings->prof[1] = NULL;
  newrings->diffprof[0] = NULL;
  newrings->diffprof[1] = NULL;
  newrings->box.next = NULL;
  return newrings;
}

/**
 * Destroy a provisional measurement collection and all its sub-structures.
 * @param data  points to a Provisionalrings structure to destroy.
 */
void
provisionalrings_destruct (void *data)
{
  provisionalringsp_t victim;

  victim = (provisionalringsp_t) data;
  if (victim == NULL)
    return;
  if (victim->fullchain)
    demolish_chainlist (&(victim->fullchain));
  if (victim->smallchain)
    demolish_chainlist (&(victim->smallchain));
  if (victim->prof[0])
    demolish_profiles (victim->prof[0]);
  if (victim->prof[1])
    demolish_profiles (victim->prof[1]);
  if (victim->diffprof[0])
    demolish_profiles (victim->diffprof[0]);
  if (victim->diffprof[1])
    demolish_profiles (victim->diffprof[1]);
  free (victim);
}

/**
 * Scale the coordinate representation of a boundary by a scalar multiplier.
 * Multiply all the x and y coordinates in the boundary representation,
 * and remove the (now inavalid) index to the points.
 * @param bound  points to the BOUNDER encoded representation of the boundary.
 * @param scale  scale factor.
 */
void
scale_boundary (BOUNDER * bound, double scale)
{
  BOUNDPOINT *pt;

  if ((bound == NULL) || (scale == 1))
    return;
  bound->stale = (bound->point_index != NULL);
  if (bound->point_index)
    {
      demolish_index (bound->point_index);
      bound->point_index = NULL;
    }
  bound->enclosure.minx /= scale;
  bound->enclosure.miny /= scale;
  bound->enclosure.maxx /= scale;
  bound->enclosure.maxy /= scale;
  bound->centroid.xpos /= scale;
  bound->centroid.ypos /= scale;
  pt = bound->points;
  while (pt)
    {
      pt->xpos /= scale;
      pt->ypos /= scale;
      pt = pt->next;
    }
}

/** A null bounding box. */
static const BOUNDBOX_NODE nullbox = { NULL, 0.0, 0.0, 0.0, 0.0 };

/**
 * Use the same boundary point index used for measurement to re-order points.
 * The boundary index is a balanced binary tree, and re-ordering will not
 * be a frequent operation, so a simplistic recursive tree traversal should
 * suffice.
 * @param subtree  the current sub-tree within the boundary index.
 * @param newhead  points to the head of the list of reordered points.
 * @param newtail  the handle to the location to add the next point.
 * @return  a linear list all the points in the subtree, in the new order.
 */
static BOUNDPOINT *
reorder_points (RBNODE * subtree, BOUNDPOINT *newhead, BOUNDPOINT **newtail)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (newtail != NULL));
  if (subtree == NULL)
    return newhead;
  if (subtree->left != NULL)
    newhead = reorder_points (subtree->left, newhead, newtail);
  else if (newhead == NULL)
    newhead = subtree->point;
  if (*newtail != NULL)
    (*newtail)->next = subtree->point;
  *newtail = subtree->point;
  subtree->point->next = NULL;
  return reorder_points (subtree->right, newhead, newtail);
}

/**
 * Re-join the points making up an irregular boundary in a more sensible order.
 * Interpolation operations can generate points along a boundary in an
 * order that does not respect the continuity of the boundary between
 * adjacent points: it may zigzag backwards and forwards in a counter-
 * intuitive fashion.  However they will correctly identify the positions
 * of the inerpolated points, not only in terms of the original coordinate
 * frame, but also by an index that reflects the projection of each point
 * onto the principal axis of the entire cloud of boundary points.
 * We use this index to re-impose an ordering that is more likely to
 * link spatially adjacent points.
 * @param crooked  the boundary requiring re-ordering.
 * @return  a boundary with its points closer to a sensible monotonic order.
 */
static BOUNDER *
straighten_boundary (BOUNDER * crooked)
{
  if ((crooked != NULL) && (crooked->points != NULL)
      && (crooked->points->next != NULL))
    {
      if (crooked->stale)
        {
          build_index (crooked);
          crooked->stale = 0;
        }
      crooked->endpoint = NULL;
      crooked->points = reorder_points (crooked->point_index, NULL,
                                        &(crooked->endpoint));
    }
  return crooked;
}

/**
 * Generate the next point on a copied translated boundary.
 * Advance to the next point on an existing boundary, generating
 * the coordinates of a point on a new boundary by applying
 * unvarying displacements in the x and y directions.
 * @param thiscontext  points to the BOUNDARY_CONTEXT for the previous point.
 * @return  a pointer to the updated BOUNDARY_CONTEXT for the current point.
 */
static BOUNDARY_CONTEXT *
extrapolation_iterator (BOUNDARY_CONTEXT * thiscontext)
{
  BOUNDPOINT *pt;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (thiscontext != NULL));
  if (!thiscontext->hasnext)
    {
      logerr_debug (LOGERR_FUNC, "extrapolating off the end of a boundary");
      return NULL;
    }
  thiscontext->hasnext = (thiscontext->points != NULL);
  if (thiscontext->hasnext)
    {
      pt = (BOUNDPOINT *) thiscontext->points;
      thiscontext->xpos = pt->xpos + thiscontext->xc[0];
      thiscontext->ypos = pt->ypos + thiscontext->yc[0];
      thiscontext->valid = pt->valid;
      thiscontext->points = (void *) pt->next;
    }
  return thiscontext;
}

/**
 * Initialize the context for generating a translated copy of a boundary.
 * Given an existing boundary and relative horizontal and vertical
 * displacements, this initializes the context for making a copy
 * of the boundary at those displacements.
 * @param existing_boundary  the BOUNDPOINT list representation of a boundary.
 * @param dx  a pointer (not a double!) to the horizontal displacement.
 * @param dy  a pointer (not a double!) to the vertical displacement.
 * @param newcontext  points to the BOUNDARY_CONTEXT to initialize.
 */
static void
initialize_extrapolation_context (BOUNDPOINT * existing_boundary,
                                  double *dx, double *dy,
                                  BOUNDARY_CONTEXT * newcontext)
{
  newcontext->hasnext = (existing_boundary != NULL);
  if (newcontext->hasnext)
    {
      newcontext->points = (void *) existing_boundary->next;
      newcontext->prevx = 0;
      newcontext->prevy = 0;
      newcontext->xpos = existing_boundary->xpos + *dx;
      newcontext->ypos = existing_boundary->ypos + *dy;
      newcontext->valid = existing_boundary->valid;
      newcontext->nterms = 1;
      newcontext->xc = dx;
      newcontext->yc = dy;
    }
  newcontext->iterator = extrapolation_iterator;
}

/**
 * Generate a horizontally and vertically translated copy of a boundary.
 * Given an existing boundary and relative horizontal and vertical
 * displacements, this iterates down the list of existing points and
 * generates a new boundary, possibly translated in both the horizontal
 * and vertical directions.
 * @param boundary  points to the BOUNDER representation of a boundary.
 * @param dx  the horizontal displacement (as a double).
 * @param dy  the vertical displacement (as a double).
 * @return  a pointer to the BOUNDER for the new translated boundary.
 */
static BOUNDER *
boundary_extrapolate (BOUNDER * existing, double dx, double dy)
{
  BOUNDARY_CONTEXT added_context;
  BOUNDER *added = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (existing != NULL));
  if (existing->points != NULL)
    {
      initialize_extrapolation_context (existing->points, &dx, &dy,
                                        &added_context);
      added = make_boundary (&added_context);
    }
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (added != NULL));
  if (added->points == NULL)
    {
      added->centroid.theta = existing->centroid.theta;
      added->centroid.xpos = existing->centroid.xpos + dx;
      added->centroid.ypos = existing->centroid.ypos + dy;
      added->enclosure.minx = existing->enclosure.minx + dx;
      added->enclosure.miny = existing->enclosure.miny + dy;
      added->enclosure.maxx = existing->enclosure.maxx + dx;
      added->enclosure.maxy = existing->enclosure.maxy + dy;
    }
  return added;
}

/**
 * Interpolate a new point between an existing point and boundary.
 * We recursively traverse the tree-structured index to the points making up
 * the existing boundary, looking for those that are close to perpendiculars
 * dropped onto the boundary from an existing point.  On each such line
 * we identify the coordinates of a point at an unvarying relative position.
 * @param pt  points to the BOUNDPOINT representation of the existing point.
 * @param lob  low bound of the search interval on the boundary principal axis.
 * @param hib high bound of the search interval on the boundary principal axis.
 * @param sub  subtree of the boundary point index (a red-black tree).
 * @param thecontext  points to the BOUNDARY_CONTEXT for accumulating results.
 */
static void
point_interpolate (BOUNDPOINT * pt,
                   double lob, double hib, RBNODE * sub,
                   BOUNDARY_CONTEXT * thecontext)
{
  double deltax, deltay, locallength;
  double thetamean, psi;
  boundbracketp_t bracket;

  if ((sub == NULL) || (lob > hib))
    return;
  if ((lob <= sub->key) && (sub->key <= hib)
      && (fabs (sub->point->theta - pt->theta) < MATCH_TOLERANCE))
    {
      deltax = sub->point->xpos - pt->xpos;
      deltay = sub->point->ypos - pt->ypos;
      thetamean = (sub->point->theta + pt->theta) / 2;
      psi = atan2 (-deltay, deltax);
      if (fabs (thetamean - psi) < MEASURE_TOLERANCE)
        {
          locallength = hypot (deltax, deltay);
          bracket = thecontext->points;
          if (locallength < bracket->d_min)
            {
              bracket->d_min = locallength;
              thecontext->xpos = (pt->xpos + (deltax * bracket->relpos));
              thecontext->ypos = (pt->ypos + (deltay * bracket->relpos));
              (bracket->nperp)++;
            }
        }
    }
  if (sub->key >= lob)
    point_interpolate (pt, lob, fmin (hib, sub->key), sub->left,
                       thecontext);
  if (sub->key <= hib)
    point_interpolate (pt, fmax (lob, sub->key), hib, sub->right,
                       thecontext);
}

/**
 * Generate the next point on a new interpolated boundary.
 * Advance to the next point on one existing boundary, generating
 * the coordinates of a point on a new boundary by interpolating
 * a point between it and another existing boundary.  Confusingly,
 * the list of points along which this iterates is nested within
 * a Boundbracket structure pointed to by the @a points member of the
 * BOUNDARY_CONTEXT structure, not by the @a points member itself.
 * We generate only one interpolated point for each point on one
 * of the old boundaries, but base this on the mean position of possibly
 * several points along the perpendiculars that can be dropped from
 * the old point onto the other old boundary. 
 * @param thiscontext  points to the BOUNDARY_CONTEXT for the previous point.
 * @return  a pointer to the updated BOUNDARY_CONTEXT for the current point.
 */
static BOUNDARY_CONTEXT *
interpolation_iterator (BOUNDARY_CONTEXT * thiscontext)
{
  boundbracketp_t bracket;
  BOUNDPOINT *pt;
  BOUNDER *bound;
  double xp, tol;
  double dx, yp, dy;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (thiscontext && thiscontext->points));
  if (!thiscontext->hasnext)
    {
      logerr_debug (LOGERR_FUNC,
                    "interpolating off the end of an inner boundary");
      return NULL;
    }
  bracket = thiscontext->points;
  pt = bracket->curr_point;
  bound = bracket->ref_bound;
  bracket->nperp = 0;
  thiscontext->xpos = 0;
  thiscontext->ypos = 0;
  while (pt && (bracket->nperp == 0))
    {
      project_on_principal_axis (pt, bound, dx, dy, yp);
      xp = dx * cos (bound->centroid.theta) - dy * sin (bound->centroid.theta);
      tol = fabs (xp * tan (SEARCH_TOLERANCE));
      point_interpolate (pt, yp - tol, yp + tol, bound->point_index,
                         thiscontext);
      thiscontext->valid = pt->valid;
      pt = pt->next;
    }
  bracket->curr_point = pt;
  bracket->nperp = 0;
  bracket->d_min = HUGE_VAL;
  thiscontext->hasnext = (pt != NULL);
  return thiscontext;
}

/**
 * Ascribe the weighted mean of angles from two existing points to a new point.
 * The weights applied to the angles from two existing BOUNDPOINT structures
 * are the euclidian distances from their positions to the position already
 * represented within the new BOUNDPOINT structure (so it must already
 * have valid coordinates).
 * @param newpoint  points to the BONDPOINT structure for the new point.
 * @param inner  points to the BONDPOINT structure for the first old point.
 * @param outer  points to the BONDPOINT structure for the second old point.
 */
static void
weight_newpoint_angle (BOUNDPOINT * newpoint,
                       BOUNDPOINT * inner, BOUNDPOINT * outer)
{
  double d_inner, d_outer;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newpoint && inner && outer));
  d_inner =
    hypot (newpoint->xpos - inner->xpos, newpoint->ypos - inner->ypos);
  d_outer =
    hypot (newpoint->xpos - outer->xpos, newpoint->ypos - outer->ypos);
  newpoint->theta =
    ((inner->theta * d_inner + outer->theta * d_outer) / (d_inner + d_outer));
}

/**
 * Determine the perpendicular distance from a point to an axis.
 * A BOUNDPOINT structure represents the axis (by the coordinates of
 * its origin and an angle); the point is just a coordinate pair in
 * the same reference frame used for the coordinates and angle
 * within the BOUNDPOINT.
 * @param x  horizontal coordinate of the point.
 * @param y  vertical coordinate of the point.
 * @param axis  points at the first BOUNDPOINT structure.
 */
static double
point_to_axis_distance (double x, double y, BOUNDPOINT * axis)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (axis != NULL));
  return (axis->theta == NOT_AN_ANGLE)
    ? (hypot ((x - axis->xpos), (y - axis->ypos)))
    : (((y - axis->ypos) * sin (axis->theta))
       - ((x - axis->xpos) * cos (axis->theta)));
}

/**
 * Initialize the context for generating a new interpolated boundary.
 * Given a pair of existing boundaries and the absolute coordinates of a 
 * point, this initializes the context for making a new boundary interpolated
 * between the existing boundaries at a position close to the point
 * coordinates.
 * @param in   points to a BOUNDER representation of the first boundary.
 * @param out  points to a BOUNDER representation of the second boundary.
 * @param x  the horizontal coordinate of the point.
 * @param y  the vertical coordinate of the point.
 * @param newcontext  points to the BOUNDARY_CONTEXT to initialize.
 */
static void
initialize_interpolation_context (BOUNDER * in, BOUNDER * out,
                                  double x, double y, boundbracketp_t bracket,
                                  BOUNDARY_CONTEXT * newcontext)
{
  double interbound;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (in && out && bracket && newcontext));
  bracket->curr_point = in->points;
  bracket->ref_bound = out;
  bracket->nperp = 0;
  bracket->d_min = HUGE_VAL;
  interbound = point_to_axis_distance (out->centroid.xpos, out->centroid.ypos,
                                       &(in->centroid));
  bracket->relpos = (fabs (interbound) < TINY)
    ? 1 : (point_to_axis_distance (x, y, &(in->centroid)) / interbound);
  newcontext->points = bracket;
  newcontext->prevx = 0;
  newcontext->prevy = 0;
  newcontext->xpos = in->centroid.xpos;
  newcontext->ypos = in->centroid.ypos;
  newcontext->valid = (in->centroid.valid && out->centroid.valid);
  newcontext->nterms = 0;
  newcontext->xc = NULL;
  newcontext->yc = NULL;
  newcontext->iterator = interpolation_iterator;
  newcontext->hasnext = (in->points != NULL) && (out->points != NULL);
  if (newcontext->hasnext)
    {
      if (out->stale)
        {
          build_index (out);
          out->stale = 0;
        }
      iterate (newcontext);
    }
}

/**
 * Interpolate a new boundary from the positions of two existing ones.
 * Given a pair of existing boundaries and the absolute coordinates of a 
 * point, this iterates down the list of points on one boundary, drops
 * perpendiculars from these onto the second boundary, and interpolates
 * points for a new boundary at positions along the perpendiculars
 * similar to the relative position of the point whose coordinates
 * are given and the existing boundaries.
 * @param in   points to a BOUNDER representation of the first boundary.
 * @param out  points to a BOUNDER representation of the second boundary.
 * @param x  the horizontal coordinate of the point.
 * @param y  the vertical coordinate of the point.
 * @return  a pointer to the BOUNDER for the new translated boundary.
 */
static BOUNDER *
boundary_interpolate (BOUNDER * in, BOUNDER * out, double x, double y)
{
  BOUNDARY_CONTEXT added_context;
  BOUNDER *added = NULL;
  boundbracketp_t place = NULL;
  double deltax, deltay;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (in && out));
  place = malloc (sizeof (struct Boundbracket));
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (place != NULL));
  deltax = out->centroid.xpos - in->centroid.xpos;
  deltay = out->centroid.ypos - in->centroid.ypos;
  if ((out->centroid.theta != NOT_AN_ANGLE) && (in->points != NULL)
      && (out->points != NULL))
    {
      initialize_interpolation_context (in, out, x, y, place, &added_context);
      added = make_boundary (&added_context);
    }
  if (added == NULL)
    {
      added = malloc (sizeof (BOUNDER));
      logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (added != NULL));
      added->points = NULL;
      added->endpoint = NULL;
      added->point_index = NULL;
      added->stale = 0;
    }
  if (added->points == NULL)
    {
      added->centroid.xpos = x;
      added->centroid.ypos = y;
      added->centroid.theta = NOT_AN_ANGLE;
      if (out->centroid.theta == NOT_AN_ANGLE)
        added->centroid.theta = in->centroid.theta;
      else if (in->centroid.theta == NOT_AN_ANGLE)
        added->centroid.theta = out->centroid.theta;
      else
        weight_newpoint_angle (&(added->centroid), &(in->centroid),
                               &(out->centroid));
      added->enclosure.minx = added->centroid.xpos;
      added->enclosure.miny = added->centroid.ypos;
      added->enclosure.maxx = added->centroid.xpos;
      added->enclosure.maxy = added->centroid.ypos;
    }
  free (place);
  return straighten_boundary(added);
}

/**
 * Generate a chain-coded boundary from a coordinate boundary representation.
 * A BOUNDER structure represents boundaries by lists of points with
 * explicit coordinates in continuous space, whereas chain-coded boundaries
 * use lists of discrete relative steps instead.  Conversion may reduce the
 * total number of points, as the coordinates of several points may fall
 * within the same discrete interval.
 * @param bound  points to the old BOUNDER boundary representation to convert.
 * @return  a newly-created chain-code boundary representation.
 */
LIST_NODE *
make_chain (BOUNDER * bound)
{
  int i, j, k, jp, kp, nsteps;
  double deltax, deltay, scale;
  BOUNDPOINT *pt = NULL;
  BOUNDPOINT *prevpt = NULL;
  LIST_NODE *newchain = NULL;
  CHAIN_NODE **prevlink = NULL;
  CHAIN_NODE *newnode = NULL;


  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (bound != NULL));
  newchain = malloc (sizeof (LIST_NODE));
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newchain != NULL));
  newchain->ring_type = NORMAL;
  newchain->chain = NULL;
  newchain->next = NULL;
  pt = bound->points;
  if (pt == NULL)
    {
      newchain->startj = round (bound->centroid.xpos);
      newchain->startk = round (bound->centroid.ypos);
      return newchain;
    }
  j = round (pt->xpos);
  k = round (pt->ypos);
  newchain->startk = j; /* Yes, ALL coordinates are transposed. */
  newchain->startj = k;
  prevpt = pt;
  prevlink = &(newchain->chain);
  pt = pt->next;
  while (pt)
    {
      deltax = pt->xpos - prevpt->xpos;
      deltay = pt->ypos - prevpt->ypos;
      nsteps = (int) ceil (hypot (deltax, deltay));
      scale = 0;
      for (i = 1; i <= nsteps; i++)
        {
          scale = (double) i / (double) nsteps;
          jp = (int) round (prevpt->xpos + deltax * scale);
          kp = (int) round (prevpt->ypos + deltay * scale);
          logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                         ((abs (jp - j) <= 1) && (abs (kp - k) <= 1)));
          if ((jp != j) || (kp != k))
            {
              newnode = malloc (sizeof (CHAIN_NODE));
              logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newnode != NULL));
              newnode->valid = pt->valid;
              CHAIN_COMPUTE_CODE (k, j, kp, jp, &(newnode->code));
              newnode->next = NULL;
              *prevlink = newnode;
              prevlink = &(newnode->next);
            }
          j = jp;
          k = kp;
        }
      prevpt = pt;
      pt = pt->next;
    }
  return newchain;
}

/** Angular limit beyond which lines are not considered nearly parallel. */
static const double PARALLEL_TOLERANCE = 0.16;

/**
 * Find the offsets needed to place a copy of a boundary near a point.
 * The new copy must not only be a short distance from the point coordinates,
 * but must also be displaced along the long axis of the ROI.
 * @param x  coordinate of the point near which to copy the boundary.
 * @param y  coordinate of the point near which to copy the boundary.
 * @param orig_x  the x coordinate of the region of interest origin.
 * @param orig_y  the y coordinate of the region of interest origin.
 * @param orig_phi  the angle of the axis of the ROI.
 * @param guide_bound  points to the boundery needing copied.
 * @param deltax  where to deposit the x displacement to use for the copy.
 * @param deltay  where to deposit the y displacement to use for the copy.
 */
static void
find_boundary_offsets (double x, double y, double orig_x, double orig_y,
                       double orig_phi, BOUNDER * guide_bound,
                       double * deltax, double * deltay)
{
  double d, dx_0, dy_0, dx_e, dy_e, x_0p, y_0p, x_ep, y_ep, psi;

  dx_e = x - orig_x;
  dy_e = y - orig_y;
  dx_0 = guide_bound->centroid.xpos - orig_x;
  dy_0 = guide_bound->centroid.ypos - orig_y;
  x_ep = dx_e * cos (orig_phi) - dy_e * sin (orig_phi);
  y_ep = dy_e * cos (orig_phi) + dx_e * sin (orig_phi);
  x_0p = dx_0 * cos (orig_phi) - dy_0 * sin (orig_phi);
  y_0p = dy_0 * cos (orig_phi) + dx_0 * sin (orig_phi);
  psi = guide_bound->centroid.theta - orig_phi;
  d = ((fabs (fabs (psi) - M_PI_2)) < PARALLEL_TOLERANCE)
    ? (x_ep - x_0p)
    : ((x_ep - y_ep * tan (psi)) - (x_0p - y_0p * tan (psi)));
  *deltax = d * cos (-orig_phi);
  *deltay = d * sin (-orig_phi);
}

/**
 * Copy an existing chain-coded boundary, and measure the ring thus defined.
 * Taking the region of interest within the image mosaic, the coordinates
 * of the new boundary position, and an existing chain-coded boundary,
 * we generate a new boundary as copy of the existing one, translated along
 * the region of interest so that it falls near the specified coordinates.
 * We transcribe the old chain-coded boundary to a representation that
 * records the absolute coordinates of each point in continuous space
 * (rather than the discrete relative steps of the chain code): a BOUNDER
 * structure.  The new boundary is initially also a BOUNDER structure,
 * and all the ring measurements are derived from this pair of BOUNDER
 * representations.  We convert the new boundary to a chain-code representation
 * (at two different resolutions), and simply discard the BOUNDER structures,
 * which isn't very elegant.
 * @param mimage  points to the mosaic image.
 * @param mindex  points to the index to the mosaic image.
 * @param scale   the downsampling factor for the low-resolution chaincode.
 * @param x the horizontal coordinate of the new boundary location.
 * @param y the vertical coordinate of the new boundary location.
 * @param orig_x  the x coordinate of the region of interest origin.
 * @param orig_y  the y coordinate of the region of interest origin.
 * @param orig_phi  the angle of the long axis of the ROI.
 * @param guide  points to the chaincoded existing boundary.
 * @param box  will point to the bounding box for the new boundary.
 * @param width  will point to the total ring width for the new ring.
 * @param prof  will point to the brightness profile for the new ring.
 * @param extrapolated  will point to the new chain-coded boundary.
 * @param small_extrapolated  will point to the new boundary, downsampled.
 */
void
chain_extrapolate (IMAGE_BYTE * mimage, MOSAIC_INDEX * mindex,
                   double scale, double x, double y,
                   double orig_x, double orig_y, double orig_phi,
                   LIST_NODE * guide, BOUNDBOX_LIST box,
                   double *width, PROFILE_LIST * prof,
                   LIST_NODE ** extrapolated, LIST_NODE ** small_extrapolated)
{
  IMAGE_CONTEXT mosaic;
  PROFILE_LIST global_head;
  PROFILE_LIST last_ringprofile;
  BOUNDARY_CONTEXT guide_context;
  BOUNDER *guide_bound = NULL;
  BOUNDER *added_bound = NULL;
  double deltax, deltay, z;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (mimage && mindex && guide && box && width && prof));
  global_head = new_profile (0);
  last_ringprofile = global_head;
  initialize_mosaic_context (mimage, mindex, &mosaic);
  initialize_chaincode_context (guide, &guide_context);
  guide_bound = make_boundary (&guide_context);
  if (guide_bound->centroid.theta != NOT_AN_ANGLE)
    {
      find_boundary_offsets (x, y, orig_x, orig_y, orig_phi, guide_bound,
                             &deltax, &deltay);
      added_bound = boundary_extrapolate (guide_bound, deltax, deltay);
      if (added_bound)
        {
          box->minx = (added_bound->enclosure).minx;
          box->miny = (added_bound->enclosure).miny;
          box->maxx = (added_bound->enclosure).maxx;
          box->maxy = (added_bound->enclosure).maxy;
          z = (((added_bound->centroid.xpos - guide_bound->centroid.xpos)
                * cos (guide_bound->centroid.theta))
               - ((added_bound->centroid.ypos - guide_bound->centroid.ypos)
                  * sin (guide_bound->centroid.theta)));
          *width =
            (z < 0)
            ? find_width (added_bound, guide_bound,
                          &mosaic, &last_ringprofile)
            : find_width (guide_bound, added_bound,
                          &mosaic, &last_ringprofile);
          *extrapolated = make_chain (added_bound);
          scale_boundary (added_bound, scale);
          *small_extrapolated = make_chain (added_bound);
          discard_boundary (added_bound);
        }
      else
        {
          *box = nullbox;
          *width = 0;
          *extrapolated = NULL;
          *small_extrapolated = NULL;
        }
      discard_boundary (guide_bound);
      *prof = last_ringprofile;
    }
}

/**
 * Measure the new ring formed by deleting the common boundary of two rings.
 * Although we will generally use this to derive a new set of measurements
 * for the single ring formed from two rings by the deletion of their
 * common boundary, in principle it is just a general single-ring measurement
 * function, analogous to a single iteration of the main loop in the
 * CHAIN_RINGWIDTHS function.
 * @param mimage  points to the mosaic image.
 * @param mindex  points to the index to the mosaic image.
 * @param inner  points to the first pre-existing chaincoded boundary.
 * @param outer  points to the second pre-existing chaincoded boundary.
 * @param width  will point to the total ring width for the new ring.
 * @param prof  will point to the brightness profile for the new ring.
 */
void
chain_collapse (IMAGE_BYTE * mimage, MOSAIC_INDEX * mindex,
                LIST_NODE * inner, LIST_NODE * outer, double *width,
                PROFILE_LIST * prof)
{
  IMAGE_CONTEXT mosaic;
  PROFILE_LIST global_head;
  PROFILE_LIST last_ringprofile;
  BOUNDARY_CONTEXT inner_context;
  BOUNDARY_CONTEXT outer_context;
  BOUNDER *inner_bound = NULL;
  BOUNDER *outer_bound = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (mimage && mindex && inner && outer && width && prof));
  global_head = new_profile (0);
  last_ringprofile = global_head;
  initialize_mosaic_context (mimage, mindex, &mosaic);
  initialize_chaincode_context (inner, &inner_context);
  initialize_chaincode_context (outer, &outer_context);
  inner_bound = make_boundary (&inner_context);
  outer_bound = make_boundary (&outer_context);
  *width = find_width (inner_bound, outer_bound, &mosaic, &last_ringprofile);
  discard_boundary (inner_bound);
  discard_boundary (outer_bound);
  *prof = last_ringprofile;
}

/**
 * Retrieve the region of interest coordinate frame in a usable form.
 * The mosaic index uses a structure for recording the current region
 * of interest, but it's slightly unwieldy to use this directly, so we
 * extract the coordinates and apply the most commonly needed
 * transformations.
 * @param roi  points to the current region of interest structure.
 * @param theframe  initialize the coordinate frame structure this points to.
 */
static void
frame_from_roi (const REGION_OF_INTEREST *roi, roiframep_t theframe)
{
  double right_x, right_y, left_x, left_y, top_x, top_y;
  double ldx, ldy, rdx, rdy, l_len, r_len;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (roi && theframe));
  right_x = roi->right[1];
  right_y = roi->right[0];
  left_x = roi->left[1];
  left_y = roi->left[0];
  top_x = roi->top[1];
  top_y = roi->top[0];
  rdx = right_x - top_x;
  rdy = right_y - top_y;
  ldx = top_x - left_x;
  ldy = left_y - top_y;
  l_len = hypot (ldx, ldy);
  r_len = hypot (rdx, rdy);
  if (l_len > r_len)
    {
      theframe->phi = atan2 (ldy, ldx);
      theframe->orig_x = left_x;
      theframe->orig_y = left_y;
      theframe->x_len = l_len;
      theframe->y_len = r_len;
    }
  else
    {
      theframe->phi = - (atan2 (rdy, rdx));
      theframe->orig_x = top_x;
      theframe->orig_y = top_y;
      theframe->x_len = r_len;
      theframe->y_len = l_len;
    }
}

/**
 * Add a single ring boundary before or after the others.
 * The chain_extrapolate function does most of the real work of inserting
 * the boundary and measuring the ring, but it needs the position and
 * orientation of the region of interest within the image mosaic, which
 * we derive here from the mosaic index.  
 * @param mimage  points to the mosaic image.
 * @param mindex  points to the index to the mosaic image.
 * @param scale   the downsampling factor for the low-resolution chaincode.
 * @param x  the horizontal coordinate of the new boundary location.
 * @param y  the vertical coordinate of the new boundary location.
 * @param guide  points to the existing chaincoded boundary to copy.
 * @param newrings  points to a Provisionalrings structure to hold the results.
 */
static void
provisional_append (IMAGE_BYTE * mimage, MOSAIC_INDEX * mindex,
                    double scale, double x, double y,
                    LIST_NODE * guide, provisionalringsp_t newrings)
{
  struct Roiframe roiframe;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (mimage && mindex && guide && newrings));
  frame_from_roi (mindex->roi, &roiframe);
  chain_extrapolate (mimage, mindex, scale, x, y,
                     roiframe.orig_x, roiframe.orig_y, roiframe.phi, guide,
                     &(newrings->box), newrings->totalwidth, newrings->prof,
                     &(newrings->fullchain), &(newrings->smallchain));
}

/** Tolerance for matching the edges of the ROI (in pixels). */
const static double roi_match_tol = 15;

/**
 * Test if coordinates fall within the region of interest in the image mosaic.
 * Transform the coordinates to a frame with its origin an upper corner of the
 * ROI and its x axis along the longest edge, then compare them to the lengths
 * of the edges of the ROI.
 * @param mindex  points to the index to the mosaic image.
 * @param x  the horizontal coordinate of the new boundary location.
 * @param y  the vertical coordinate of the new boundary location.
 * @return true if the coordinates are inside (the int is really a boolean).
 */
int
provisional_point_in_roi (MOSAIC_INDEX * mindex, double x, double y)
{
  double x_ep, y_ep, dx_e, dy_e;
  struct Roiframe roiframe;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (mindex != NULL));
  frame_from_roi (mindex->roi, &roiframe);
  dx_e = x - roiframe.orig_x;
  dy_e = y - roiframe.orig_y;
  x_ep = dx_e * cos (roiframe.phi) - dy_e * sin (roiframe.phi);
  y_ep = dy_e * cos (roiframe.phi) + dx_e * sin (roiframe.phi);
  return ((0 <= (x_ep + roi_match_tol))
          && (x_ep <= (roiframe.x_len + roi_match_tol)) 
          && (0 <= (y_ep + roi_match_tol))
          && (y_ep <= (roiframe.y_len + roi_match_tol)));
}

/** Distance limit below which points get considered coincident. */
const static double point_line_match_tol = 2e-10;

/**
 * Find the distance between a point and a bounded straight line segment.
 * If the point lies on a perpendicular from the line, use that length
 * as the distance, otherwise the distance to the closest end point.
 * @param x  coordinate of the point.
 * @param y  coordinate of the point.
 * @param x_start  coordinate of the line segment start.
 * @param y_start  coordinate of the line segment start.
 * @param x_end  coordinate of the line segment end.
 * @param y_end  coordinate of the line segment end.
 * @return  the distance from the line segment.
 */
static double
point_to_line_dist (double x, double y, double x_start, double y_start,
                      double x_end, double y_end)
{
  double t, len2, delta_x, delta_y, x_p, y_p;

  delta_x = x_end - x_start;
  delta_y = y_end - y_start;
  x_p = x - x_start;
  y_p = y - y_start;
  len2 = delta_x * delta_x + delta_y * delta_y;
  if (len2 < point_line_match_tol)
    return hypot (x_p, y_p); /* line starts and ends at the same coordinates */
  t = ((delta_x * x_p + delta_y * y_p) / len2);
  if (t <= 0)                /* point is closest to the line start... */
    return hypot (x_p, y_p);
  else if (t >= 1)          /* point is closest to the line end... */
    return hypot (x - x_end, y - y_end);
  else  /* point is closest to a perpendicular dropped on the line... */
    return (sqrt (len2) * fabs ((delta_y * x_p - delta_x * y_p) / len2));
}

/**
 * Find the minimum distance between a point and a chain-coded boundary.
 * Although the individual line segments making up the chain-coded boundary
 * can only have two lengths, and are joined at nodes that lie on a regular
 * grid, the coordinates of the point are not necessarily constrained in any
 * way, so we treat it as the more general case of finding the distance
 * between a point and an arbitrary polyline in continuous coordinates.
 * @param x  coordinate of the point whose distance we are determining.
 * @param y  coordinate of the point whose distance we are determining.
 * @param x  points to the chain-coded representation of a boundary.
 * @return  the minimum point-boundary distance.
 */
double
provisional_point_distance (double x, double y, LIST_NODE * boundary)
{
  double d, d_min, x_start, x_end, y_start, y_end;
  int prevx, prevy, currx, curry;
  CHAIN_NODE *node = NULL;

  x_end = currx = boundary->startk;
  y_end = curry = boundary->startj;
  d_min = hypot ((x - x_end), (y - y_end));
  node = boundary->chain;
  while (node != NULL)
    {
      x_start = x_end;
      y_start = y_end;
      prevx = currx;
      prevy = curry;
      CHAIN_COMPUTE_COORDINATES (prevx, prevy, node->code, &currx, &curry);
      x_end = currx;
      y_end = curry;
      d = point_to_line_dist (x, y, x_start, y_start, x_end, y_end);
      if (d < d_min)
        d_min = d;
      node = node->next;
    }
  return d_min;
}

/**
 * Interpolate a new chain-coded ring boundary between two existing ones.
 * Given two existing chain-coded boundaries and the coordinates of
 * a point, we construct a new boundary by interpolation between the
 * boundaries, taking the point coordinates as an approximate guide to
 * its position.  As normally used, the new boundary will split a previously
 * measured ring in two, so there will be not just one set of modified
 * ring measurements, but two.   We transcribe the existing chain-coded
 * boundaries to a representation that use the absolute coordinates of each
 * point (BOUNDER structures); when first constructed, the new boundary
 * uses the same representation, however the new boundary gets converted to a
 * chain-coded representation (at two different resolutions), which along with
 * the new measurements are the only data output from this function.  The
 * original BOUNDER representation of the interpolated boundary lacks the
 * accurate information about local perpendicular directions needed for
 * the width measurement, so we wastefully discard it after creating the
 * chain-coded boundary, re-generate a new BOUNDER from the chain code,
 * and use this for the measurements, before discarding it as well.
 * @param mimage  points to the mosaic image.
 * @param mindex  points to the index to the mosaic image.
 * @param scale   the downsampling factor for the low-resolution chaincode.
 * @param x the horizontal coordinate of the new boundary location.
 * @param y the vertical coordinate of the new boundary location.
 * @param inner  points to the first pre-existing chaincoded boundary.
 * @param outer  points to the second pre-existing chaincoded boundary.
 * @param box  will point to the bounding box for the new boundary.
 * @param width  will point to the total ring width for the new rings.
 * @param prof  will point to the brightness profile for the new rings.
 * @param interpolated  will point to the new chain-coded boundary.
 * @param small_interpolated  will point to the new boundary, downsampled.
 */
void
chain_interpolate (IMAGE_BYTE * mimage, MOSAIC_INDEX * mindex,
                   double scale, double x, double y,
                   LIST_NODE * inner, LIST_NODE * outer,
                   BOUNDBOX_LIST box, double *w_before,
                   PROFILE_LIST * prof_before, double *w_after,
                   PROFILE_LIST * prof_after, LIST_NODE ** interpolated,
                   LIST_NODE ** small_interpolated)
{
  IMAGE_CONTEXT mosaic;
  PROFILE_LIST inner_head;
  PROFILE_LIST inner_last;
  PROFILE_LIST outer_head;
  PROFILE_LIST outer_last;
  BOUNDARY_CONTEXT inner_context;
  BOUNDARY_CONTEXT outer_context;
  BOUNDARY_CONTEXT added_context;
  BOUNDER *inner_bound = NULL;
  BOUNDER *outer_bound = NULL;
  BOUNDER *added_bound = NULL;
  BOUNDER *nearest_bound = NULL;
  struct Roiframe roiframe;
  double xdisp, ydisp;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (mimage && mindex && inner && outer && box
                  && w_before && prof_before && w_after && prof_after
                  && interpolated && small_interpolated));
  *box = nullbox;
  *w_before = 0;
  *w_after = 0;
  *interpolated = NULL;
  *small_interpolated = NULL;
  inner_head = new_profile (0);
  inner_last = inner_head;
  outer_head = new_profile (0);
  outer_last = outer_head;
  initialize_mosaic_context (mimage, mindex, &mosaic);
  initialize_chaincode_context (inner, &inner_context);
  initialize_chaincode_context (outer, &outer_context);
  inner_bound = make_boundary (&inner_context);
  outer_bound = make_boundary (&outer_context);
  added_bound = boundary_interpolate (inner_bound, outer_bound, x, y);
  if (added_bound && (added_bound->points == NULL))
    {
      discard_boundary (added_bound);
      frame_from_roi (mindex->roi, &roiframe);
      nearest_bound = (provisional_point_distance (x, y, inner)
                       < provisional_point_distance (x, y, outer))
        ? inner_bound : outer_bound;
      find_boundary_offsets (x, y, roiframe.orig_x, roiframe.orig_y,
                             roiframe.phi, nearest_bound,
                             &xdisp, &ydisp);
      added_bound = boundary_extrapolate (nearest_bound, xdisp, ydisp);
    }
  if (added_bound && (added_bound->points != NULL))
    {
      *interpolated = make_chain (added_bound);
      scale_boundary (added_bound, scale);
      *small_interpolated = make_chain (added_bound);
      discard_boundary (added_bound);
      initialize_chaincode_context (*interpolated, &added_context);
      added_bound = make_boundary (&added_context);
    }
  if (added_bound && (added_bound->points != NULL))
    {
      *w_before = find_width (inner_bound, added_bound, &mosaic, &inner_last);
      *w_after = find_width (added_bound, outer_bound, &mosaic, &outer_last);
    }
  if (added_bound)
    {
      box->minx = (added_bound->enclosure).minx;
      box->miny = (added_bound->enclosure).miny;
      box->maxx = (added_bound->enclosure).maxx;
      box->maxy = (added_bound->enclosure).maxy;
      discard_boundary (added_bound);
    }
  discard_boundary (inner_bound);
  discard_boundary (outer_bound);
  *prof_before = inner_last;
  *prof_after = outer_last;
}

/**
 * Store new ring data in the hash table, computing any derived measurements.
 * Assuming there are already a set of essential measurements (ring widths
 * and brightness profiles) in a Provisionalrings structure, we compute
 * the derived measurements (e.g., earlywood width and latewood width),
 * then place the structure in the hash table, with a key of the form
 * "Provisional-<provbnum>".
 * @param doc  the string identifying the document in the hash table.
 * @param newrings  points to the structure holding all the measurements.
 * @param provbnum  the number used to derive the hash key.
 * @return an error code on failure, zero otherwise.
 */
static int
provisional_store (char *doc, provisionalringsp_t newrings, int provbnum)
{
  struct Widthvec total, late, early;
  PROFILE_LIST *profiles;
  PROFILE_LIST *diffprofiles;
  char pname[SLEN];
  int i;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((doc != NULL) && (newrings != NULL)));
  total.len = 1;
  total.val = newrings->totalwidth;
  early.val = newrings->earlywidth;
  late.val = newrings->latewidth;
  profiles = newrings->prof;
  diffprofiles = newrings->diffprof;
  for (i = newrings->n; i > 0; i--)
    {
      if (*profiles)
        First_diff_profiles (profiles, diffprofiles, &total, &early, &late);
      else
        {
          *diffprofiles = NULL;
          *(early.val) = 0;
          *(late.val) = 0;
        }
      profiles++;
      diffprofiles++;
      total.val++;
      early.val++;
      late.val++;
    }
  snprintf (pname, SLEN - 1, "Provisional-%d", provbnum);
  err = sad_put_doc_property (global_doclist, doc, pname,
                              provisionalrings_destruct, newrings);
  return err;
}

/**
 * Insert a new boundary in the chain map between two existing boundaries.
 * Given a pair of chain-coded boundaries (which are already in the
 * full-resolution chain map) and the coordinates of a point between them,
 * we construct a new boundary close to this point, derive revised measurements
 * from the re-defined rings, and store them with the new boundaries in the
 * hash table.
 * @param doc  a string identifying the hash table document.
 * @param provbnum  used to form the hash key to store the measurements.
 * @param mimage  points to the mosaic image.
 * @param mindex  points to the index to the mosaic image.
 * @param scale   the downsampling factor for the low-resolution chaincode.
 * @param x  the horizontal coordinate of the new boundary location.
 * @param y  the vertical coordinate of the new boundary location.
 * @param inner  points to the inner existing chain-coded boundary.
 * @param outer  points to the outer existing chain-coded boundary.
 * @return an error code on failure, zero otherwise.
 */
int
provisional_insert (char *doc, int provbnum, IMAGE_BYTE * mimage,
                    MOSAIC_INDEX * mindex, double scale, double x, double y,
                    LIST_NODE * inner, LIST_NODE *outer)
{
  provisionalringsp_t newrings = NULL;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (doc && mimage && mindex && inner && outer));
  newrings = new_provisionalrings (2, provbnum, inner);
  chain_interpolate (mimage, mindex, scale, x, y, inner, outer,
                     &(newrings->box), newrings->totalwidth, newrings->prof,
                     (newrings->totalwidth) + 1, (newrings->prof) + 1,
                     &(newrings->fullchain), &(newrings->smallchain));
  if (newrings->fullchain == NULL)
    err = -1;
  else
    {
      newrings->fullchain->ring_id = provbnum;
      newrings->smallchain->ring_id = provbnum;
      err = provisional_store (doc, newrings, provbnum);
    }
  if (err)
    provisionalrings_destruct (newrings);
  return err;
}

/**
 * Delete an existing chain-coded boundary.
 * Given a chain-coded boundary that is already in the chain map lists, and
 * two enclosing boundaries (which are also already in the chain map, or
 * are NULL), remove the boundary and re-compute the measurements
 * to be associated with the enclosing boundaries (if they exist).
 * @param doc  a string identifying the hash table document.
 * @param provbnum  used to form the hash key to store the measurements.
 * @param mimage  points to the mosaic image.
 * @param mindex  points to the index to the mosaic image.
 * @param inner  points to the inner enclosing chain-coded boundary.
 * @param delete  points to the deleted existing chain-coded boundary.
 * @param outer  points to the outer enclosing chain-coded boundary.
 * @param small  points to the deleted low-resolution chain-coded boundary.
 * @return an error code on failure, zero otherwise.
 */
int
provisional_delete (char *doc, int provbnum, IMAGE_BYTE * mimage,
                    MOSAIC_INDEX * mindex, LIST_NODE * inner,
                    LIST_NODE * deleted, LIST_NODE * outer, LIST_NODE * small)
{
  provisionalringsp_t newrings = NULL;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (doc && mimage && mindex && deleted && small));
  if (inner == NULL)
    {
      newrings = new_provisionalrings (1, provbnum, NULL);
    }
  else
    {
      newrings = new_provisionalrings (1, provbnum, inner);
      if (outer != NULL)
        chain_collapse (mimage, mindex, inner, outer, newrings->totalwidth,
                        newrings->prof);
    }
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (((inner != NULL) && (outer != NULL))
                  == (*(newrings->prof) != NULL)));
  copy_chainlist_head (deleted, &(newrings->fullchain));
  copy_chainlist_head (small, &(newrings->smallchain));
  err = provisional_store (doc, newrings, provbnum);
  if (err)
    provisionalrings_destruct (newrings);
  return err;
}

/** A constant larger than any measurable distance. */
#define MAXIMUM_DISTANCE HUGE_VAL

/**
 * Extend the list of chain-coded boundaries by a new boundary at the end.
 * Given the first and last chain-coded boundaries (which are already in the
 * full-resolution chain map) and the coordinates of a point outside them,
 * we construct a new boundary close to this point, derive revised measurements
 * from the re-defined rings, and store them with the new boundaries in the
 * hash table.
 * @param doc  a string identifying the hash table document.
 * @param provbnum  used to form the hash key to store the measurements.
 * @param mimage  points to the mosaic image.
 * @param mindex  points to the index to the mosaic image.
 * @param scale   the downsampling factor for the low-resolution chaincode.
 * @param x  the horizontal coordinate of the new boundary location.
 * @param y  the vertical coordinate of the new boundary location.
 * @param initial  points to the first existing chain-coded boundary.
 * @param final  points to the last existing chain-coded boundary.
 * @return an error code on failure, zero otherwise.
 */
int
provisional_extend (char *doc, int provbnum, IMAGE_BYTE * mimage,
                    MOSAIC_INDEX * mindex, double scale, double x, double y,
                    LIST_NODE * initial, LIST_NODE * final)
{
  provisionalringsp_t newrings = NULL;
  double d_initial = MAXIMUM_DISTANCE;
  double d_final = MAXIMUM_DISTANCE;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (doc && mimage && mindex && (initial || final)));
  if (initial != NULL)
    d_initial = provisional_point_distance (x, y, initial);
  if (final != NULL)
    d_final = provisional_point_distance (x, y, final);
  if (d_initial < d_final)
    {
      newrings = new_provisionalrings (1, provbnum, NULL);
      provisional_append (mimage, mindex, scale, x, y, initial, newrings);
    }
  else if (d_initial > d_final)
    {
      newrings = new_provisionalrings (1, provbnum, final);
      provisional_append (mimage, mindex, scale, x, y, final, newrings);
    }
  if (newrings->fullchain == NULL)
    err = -1;
  else
    {
      newrings->fullchain->ring_id = provbnum;
      newrings->smallchain->ring_id = provbnum;
      err = provisional_store (doc, newrings, provbnum);
    }
  if (err)
    provisionalrings_destruct (newrings);
  return err;
}

/**
 * If a test line intersects a polygon edge, flip the crossing flag.
 * Wm Randolph Franklin wrote the original version of this test:
 * see http://www.ecse.rpi.edu/Homepages/wrf/research/geom/pnpoly.html
 * and the FAQ for comp.graphics.algorithms (Subject 2.03):
 * http://www.faqs.org/faqs/graphics/algorithms-faq/
 * for more details.
 * @param inside  the edge crossing flag (effectively boolean).
 * @param x  the test line is along this (positive) x direction.
 * @param y  starting at a point with this y coordinate.
 * @param x_start  x coordinate of the start of a polygon edge.
 * @param y_start  y coordinate of the start of a polygon edge.
 * @param x_end  x coordinate of the start of a polygon edge.
 * @param y_end  y coordinate of the start of a polygon edge.
 * @return the crossing flag updated for the current polygon edge.
 */
static int
wrf_crossing_test (int inside, double x, double y,
                   double x_start, double y_start,
                   double x_end, double y_end)
{
  if ((((y_end <= y) && (y < y_start))
       || ((y_start <= y) && (y < y_end)))
      && (x < (x_start - x_end) * (y - y_end) / (y_start - y_end)
          + x_end))
    inside = !inside;
  return inside;
}

/**
 * Check if a point lies between two chain-coded boundaries.
 * We treat the boundaries as edges of a polygon, which is closed by defining
 * an extra edge between the first points of the boundary pair, and another
 * edge between the last points of the pair.  For full background see
 * Haines, Eric, "Point in Polygon Strategies", Graphics Gems IV,
 *   ed. Paul Heckbert : Academic Press : 1994.  p. 24--46. and equivalently:
 * http://www.acm.org/tog/editors/erich/ptinpoly/
 * @param x  coordinate of the point to test;
 * @param y  coordinate of the point to test;
 * @param inner  points to the chaincoded representation of one boundary.
 * @param outer  points to the chaincoded representation of the other boundary.
 */
int
provisional_check (double x, double y, LIST_NODE * inner, LIST_NODE * outer)
{
  double inner_lastx, inner_lasty, x_start, x_end, y_start, y_end;
  int prevx, prevy, currx, curry;
  CHAIN_NODE *forward = NULL;
  CHAIN_NODE *reverse = NULL;
  int inside = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (inner && outer));
  forward = inner->chain;
  x_end = currx = inner->startk;
  y_end = curry = inner->startj;
  while (forward)
    {
      x_start = x_end;
      y_start = y_end;
      prevx = currx;
      prevy = curry;
      CHAIN_COMPUTE_COORDINATES (prevx, prevy, forward->code, &currx, &curry);
      x_end = currx;
      y_end = curry;
      inside = wrf_crossing_test (inside, x, y, x_start, y_start,
                                  x_end, y_end);
      forward = forward->next;
    }
  inner_lastx = x_end;
  inner_lasty = y_end;
  x_end = inner->startk;
  y_end = inner->startj;
  x_start = currx = outer->startk;
  y_start = curry = outer->startj;
  inside = wrf_crossing_test (inside, x, y, x_start, y_start,
                              x_end, y_end);
  reverse = outer->chain;
  while (reverse)
    {
      x_end = x_start;
      y_end = y_start;
      prevx = currx;
      prevy = curry;
      CHAIN_COMPUTE_COORDINATES (prevx, prevy, reverse->code, &currx, &curry);
      x_start = currx;
      y_start = curry;
      inside = wrf_crossing_test (inside, x, y, x_start, y_start,
                                  x_end, y_end);
      reverse = reverse->next;
    }
  inside = wrf_crossing_test (inside, x, y, inner_lastx, inner_lasty,
                              x_start, y_start);
  return inside;
}

/**
 * Test if there are valid modified data in the provisional measurements.
 * Even if there are no valid profiles, there will always be a (possibly
 * empty) profile structure if there are valid widths, so use its
 * presence as the test.
 * @param doc  the string identifying the document in the hash table.
 * @param provbnum  the number used to derive the hash key.
 * @return  a boolena result, true if there are any valid measurements.
 */
int
have_provisional_changes (char * doc, char * pname)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc != NULL) && (pname != NULL));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  return (prov != NULL) && (*(prov->prof) != NULL) ;
}

/**
 * Recover the full-resolution chain-coded boundary from the hash table.
 * @param doc  the string identifying the document in the hash table.
 * @param provbnum  the number used to derive the hash key.
 * @return  a pointer to the head of the boundary, or NULL on failure.
 */
LIST_NODE *
get_provisional_full_chain (char * doc, char * pname)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc != NULL) && (pname != NULL));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  return (prov == NULL) ? NULL : prov->fullchain;
}

/**
 * Recover the downsampled chain-coded boundary from the hash table.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @return  a pointer to the head of the boundary, or NULL on failure.
 */
LIST_NODE *
get_provisional_small_chain (char * doc, char * pname)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc != NULL) && (pname != NULL));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  return (prov == NULL) ? NULL : prov->smallchain;
}

/**
 * Recover the insertion point in the list of chain-coded boundaries.
 * The id number for the list item immediately before the insertion point
 * should have been saved in the hash table, and may be the invalid id code
 * if the insertion point is at the head of the list.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @return  the ring_id number of a list item, or the invalid id on failure.
 */
int
get_provisional_insert_point (char * doc, char * pname)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc != NULL) && (pname != NULL));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (prov != NULL));
  return (prov == NULL) ? INVALID_ID : prov->insert_id;
}

/**
 * Find the array of provisional total ring widths in the hash table.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @param widthhandle  will get set to a pointer to the first width.
 * @return  the number of valid widths (should be 1 or 2).
 */
int
find_provisional_widths (char * doc, char * pname, double * * widthhandle)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && widthhandle));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  if ((prov != NULL) && ((prov->n == 1) || (prov->n == 2)))
    {
      *widthhandle = prov->totalwidth;
      return prov->n;
    }
  else
    {
      logerr_debug (LOGERR_FUNC, 
                    "Cannot find ring widths for %s", pname);
      *widthhandle = NULL;
      return 0;
    }
}

/**
 * Find the array of provisional brightness profiles in the hash table.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @param profhandle  will get set to a pointer to the first brightness list.
 * @return  the number of valid profiles (should be 1 or 2).
 */
int
find_provisional_profiles (char * doc, char * pname,
                           PROFILE_LIST * * profhandle)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && profhandle));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  if ((prov != NULL) && ((prov->n == 1) || (prov->n == 2)))
    {
      *profhandle = prov->prof;
      return prov->n;
    }
  else
    {
      logerr_debug (LOGERR_FUNC, 
                    "Cannot find brightness profile for %s", pname);
      *profhandle = NULL;
      return 0;
    }
}

/**
 * Find the array of provisional differenced brightness profiles.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @param diffprofhandle  will get set to a pointer to the first profile list.
 * @return  the number of valid profiles (should be 1 or 2).
 */
int
find_provisional_diff_profiles (char * doc, char * pname,
                                PROFILE_LIST * * diffprofhandle)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && diffprofhandle));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  if ((prov != NULL) && ((prov->n == 1) || (prov->n == 2)))
    {
      *diffprofhandle = prov->diffprof;
      return prov->n;
    }
  else
    {
      logerr_debug (LOGERR_FUNC, 
                    "Cannot find differenced profile for %s", pname);
      *diffprofhandle = NULL;
      return 0;
    }
}

/**
 * Find the array of provisional earlywood widths in the hash table.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @param earlywidthhandle  will get set to a pointer to the first width.
 * @return  the number of valid widths (should be 1 or 2).
 */
int
find_provisional_earlywidths (char * doc, char * pname,
                              double * * earlywidthhandle)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && earlywidthhandle));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  if ((prov != NULL) && ((prov->n == 1) || (prov->n == 2)))
    {
      *earlywidthhandle = prov->earlywidth;
      return prov->n;
    }
  else
    {
      logerr_debug (LOGERR_FUNC, 
                    "Cannot find earlywood widths for %s", pname);
      *earlywidthhandle = NULL;
      return 0;
    }
}

/**
 * Find the array of provisional latewood widths in the hash table.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @param latewidthhandle  will get set to a pointer to the first width.
 * @return  the number of valid widths (should be 1 or 2).
 */
int
find_provisional_latewidths (char * doc, char * pname,
                             double * * latewidthhandle)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && latewidthhandle));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  if ((prov != NULL) && ((prov->n == 1) || (prov->n == 2)))
    {
      *latewidthhandle = prov->latewidth;
      return prov->n;
    }
  else
    {
      logerr_debug (LOGERR_FUNC, 
                    "Cannot find latewood widths for %s", pname);
      *latewidthhandle = NULL;
      return 0;
    }
}

/**
 * Find the bounding box for the provisional ring.
 * Although this could be the head of a list of bounding boxes in a completely
 * general case, here we're assuming it's the single box for one ring boundary.
 * @param doc  the string identifying the document in the hash table.
 * @param pname  a string identifying the set of provisional measrements.
 * @param boundboxhandle  will get set to a pointer to bounding box.
 * @return  the number of valid boxes (should be exactly 1, or 0 if an error).
 */
int
find_provisional_boundbox (char * doc, char * pname,
                           BOUNDBOX_LIST * boundboxhandle)
{
  provisionalringsp_t prov = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && boundboxhandle));
  prov = sad_get_doc_property (global_doclist, doc, pname);
  if (prov != NULL)
    {
      *boundboxhandle = &(prov->box);
      return 1;
    }
  else
    {
      logerr_debug (LOGERR_FUNC, 
                    "Cannot find bounding box for %s", pname);
      *boundboxhandle = NULL;
      return 0;
    }
}
