/** @file data_export.c
 * Exports data such as ring widths from TREES in formats that other software
 * can undestand, including the traditional Tucson raw ring width format
 * (.RWL files) and a comma-separated value text format designed for import
 * into most spreadsheets.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: data_export.c,v 1.4 2004/09/07 23:35:08 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <errno.h>
#include <string.h>
#include "logerr.h"
#include <sadie.h>
#include "Sadie_Index.h"
#include "trees_widthvec.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

const static size_t initbuflen = 128; /**< initial line buffer length */
const static size_t cardbuflen = 82; /**< 80-column record + newline + null */
const static int shortendflag = 999; /**< end flag for ITRDB 0.01mm series */
/**
 * The maximum ITRDB id field length + 1.
 */
#define IDBUFLEN  9

struct Extenstr
{
  size_t max;
  size_t len;
  char *str;
};

typedef struct Extenstr * extenstrp_t;

static extenstrp_t
make_extenstr (size_t size)
{
  extenstrp_t es;

  es = malloc (sizeof (struct Extenstr));
  if (es == NULL)
    logerr_warning (LOGERR_FUNC, "could not allocate extensible string");
  else
    {
      es->max = size;
      es->len = 0;
      es->str = malloc (size);
      if (es->str != NULL)
        memset (es->str, 0, size);
      else
        {
          logerr_warning (LOGERR_FUNC,
                          "could not allocate extensible string data");
          free (es);
          es = NULL;
        }
    }
  return es;
}

static void
dispose_extenstr (extenstrp_t *eshandle)
{

  if ((eshandle != NULL) && (*eshandle != NULL))
    {
      free ((*eshandle)->str);
      free (*eshandle);
    }
}

static void
reset_extenstr (extenstrp_t es)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((es != NULL) && (es->str != NULL)));
  memset (es->str, 0, es->max);
  es->len = 0;
}

static void
print_extenstr (extenstrp_t es, const char *fmt, ...)
{
  va_list ap;
  int bytes;
  size_t avail;
  char *newstr;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((es != NULL) && (es->str != NULL)
                  && (es->max > es->len)));
  do
    {
      avail = es->max - es->len;
      va_start (ap, fmt);
      bytes = vsnprintf (es->str+es->len, avail, fmt, ap);
      va_end (ap);
      if ((bytes >= avail) || (bytes < 0))
        {
          newstr = realloc (es->str, es->max + initbuflen);
          if (newstr == NULL)
            logerrno (LOGERR_CRIT, errno, LOGERR_FUNC,
                      "cannot extend the string");
          else
            {
              es->str = newstr;
              es->max += initbuflen;
              memset (es->str+es->len, 0, avail + initbuflen);
            }
        }
    }
  while ((bytes >= avail) || (bytes < 0));
  es->len += bytes;
}
  

static void
format_itrdb_decade (const widthvecp_t measure, const char *seriesid,
                     int *decyr, int *declen, int *i, extenstrp_t buf)
{
  int mval;

  if ((*decyr <= -1000) || (seriesid[8] == '-'))
    print_extenstr (buf, "%-7s%5d", seriesid, *decyr);
  else
    print_extenstr (buf, "%-8s%4d", seriesid, *decyr);
  *decyr += *declen;
  while ((*i < measure->len) && *declen)
    {
      mval = (int) rint (measure->val[(*i)++]);
      if (mval == shortendflag)
        mval++;
      print_extenstr (buf, "%6d", mval);
      (*declen)--;
    }
}

int
trees_rwl_export (const sad_doclistp_t doclist, const char *dkey,
                  int (*buf_exporter) (void *, char *), void * ctx,
                  int yr, const char *seriesid, const char *prop)
{
  int err = 0;
  int i, kyr, dlen;
  extenstrp_t buf;
  widthvecp_t measure;
  char qid[IDBUFLEN]; /* RIP RLH */

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((doclist != NULL) && (dkey != NULL) && (buf_exporter != NULL)
                  && (prop != NULL)));
  measure = sad_get_doc_property (doclist, dkey, prop);
  if ((measure == NULL) || (measure->len == 0))
    {
      logerr_warning (LOGERR_FUNC,
                      "no valid measurements found for %s", prop);
      return -1;
    }
  buf = make_extenstr (cardbuflen);
  if (buf == NULL)
    logerr_crit (LOGERR_FUNC, "could not allocate the %s export buffer", prop);
  snprintf(qid, IDBUFLEN, "%-8s", seriesid);
  qid[IDBUFLEN - 1] = 0;
  i = 0;
  kyr = yr;
  if (yr >= 0)
    dlen = 10 - yr % 10;
  else
    {
      dlen = abs (yr) % 10;
      if (dlen == 0)
        dlen = 10;
    }
  while (!err && (i < measure->len))
    {
      format_itrdb_decade (measure, qid, &kyr, &dlen, &i, buf);
      if (dlen == 0)
        {
          print_extenstr (buf, "\n");
          err = (*buf_exporter) (ctx, buf->str);
          if (!err)
            {
              reset_extenstr (buf);
              dlen = 10;
            }
        }
    }
  if (!err)
    {
      if (dlen == 10)
        format_itrdb_decade (measure, qid, &kyr, &dlen, &i, buf);
      print_extenstr (buf, "%6d\n", shortendflag);
      err = (*buf_exporter) (ctx, buf->str);
    }
  if (err)
    logerrno (LOGERR_ERR, err, LOGERR_FUNC,
              "error exporting the %s measurements", prop);
  dispose_extenstr (&buf);
  return err;
}

int
trees_csv_export (const sad_doclistp_t doclist, const char *dkey,
                  int (*buf_exporter) (void *, char *), void * ctx,
                  int yr, int propc, const char *propv[])
{
  int err = 0;
  int i, n, iprop, jprop;
  extenstrp_t buf;
  widthvecp_t *measure;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((doclist != NULL) && (dkey != NULL) && (buf_exporter != NULL)
                  && (propc > 0) && (propv != NULL)));
  buf = make_extenstr (initbuflen);
  if (buf == NULL)
    logerr_crit (LOGERR_FUNC, "could not allocate the export buffer");
  measure = calloc (propc, sizeof (widthvecp_t));
  if (measure == NULL)
    logerr_crit (LOGERR_FUNC, "could not allocate measurement list");
  n = 0;
  print_extenstr (buf, "year");
  for (iprop = 0; iprop < propc; iprop++)
    {
      print_extenstr (buf, ", ");
      if (propv[iprop])
        {
          measure[iprop] = sad_get_doc_property (doclist, dkey, propv[iprop]);
          if (measure[iprop]->len > n)
            n = measure[iprop]->len;
          print_extenstr (buf, "%s", propv[iprop]);
        }
    }
  if (n == 0)
    {
      logerr_warning (LOGERR_FUNC, "no valid measurements found");
      err = -1;
    }
  else
    {
      print_extenstr (buf, "\n");
      err = (*buf_exporter) (ctx, buf->str);
      for (i = 0; !err && (i < n); i++, yr++)
        {
          reset_extenstr (buf);
          print_extenstr (buf, "%6d", yr);
          for (jprop = 0; jprop < propc; jprop++)
            {
              print_extenstr (buf, ",");
              if (measure[jprop] && (i < (measure[jprop]->len)))
                print_extenstr  (buf, "%12.5f", measure[jprop]->val[i]);
            }
          print_extenstr (buf, "\n");
          err = (*buf_exporter) (ctx, buf->str);
        }
      if (err)
        logerrno (LOGERR_ERR, err, LOGERR_FUNC,
                  "error exporting the measurements");
    }
  dispose_extenstr (&buf);
  free (measure);
  return err;
}
