/** @file mosaic_byte_export.c
 * Transpose rows and columns when storing some TREES mosaic images.
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: mosaic_byte_export.c,v 1.4 2004/08/06 22:47:11 mmunro Exp $
 */

/* This file is part of TREES.

   TREES is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   TREES is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with TREES; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <sadie.h>
#include "sadie_byte.h"

#ifdef WORDS_BIGENDIAN
#define NEEDSWAP 0
#else
#define NEEDSWAP 1
#endif

/**
 * Write an mosaic image in the TREES-specific format with one byte per pixel.
 * The name should be something like the path to a disk file to receive
 * the byte mosaic image.  Currently it cannot handle regular SADIE images.
 * @param *in    an existing image (already in byte image format).
 * @param *name  the path name from which to read the mosaic image.
 */
void
MOSAIC_BYTE_EXPORT (IMAGE_BYTE *in, const char *name)
{
  char msg[SLEN];
  FILE *fp;
  int err;
  size_t n_written, i, j, k;
  IMAGE_BYTE outbuf;
  BYTE *column;

  memset (msg, 0, SLEN);
  if (name == NULL)
    {
      snprintf (msg, SLEN - 1, " No byte mosaic image file path specified.");
      MESSAGE ('E', msg);
      return;
    }
  if (!CHECKIMG_BYTE(in)
      || (in->fileversion != FILE_VERSION) || (in->dataformat != EXTDATA_UINT8)
      || (in->datapacking != BE_BYTEPACKED) || (in->nbit != SBPB))
    {
      snprintf (msg, SLEN - 1, " Trying to write invalid byte mosaic image to %s.", name);
      MESSAGE ('E', msg);
      return;
    }
  fp = fopen (name, "w");
  if (fp == NULL)
    {
      snprintf (msg, SLEN - 1, " Can't open byte mosaic image %s.", name);
      MESSAGE ('E', msg);
      return;
    }
  outbuf = *in;
  /* Record the transposition of rows & columns. */
  outbuf.datarotation[0] = LINE_DIMENSION     | VALUES_NORMAL;
  outbuf.datarotation[1] = PIXEL_DIMENSION    | VALUES_NORMAL;
  outbuf.datarotation[2] = BAND_DIMENSION     | VALUES_NORMAL;
  outbuf.datarotation[3] = INTERVAL_DIMENSION | VALUES_NORMAL;

  /* Change the image name to match the file path name. */
  memset (outbuf.text, 0, TLEN);
  strncpy (outbuf.text, name, TLEN-1);
  err = sad_byte_header_write (&outbuf, fp);
  if (err)
    {
      snprintf (msg, SLEN - 1, " Can't write byte mosaic image header to %s.", name);
      MESSAGE ('E', msg);
    }
  else
    {
      column = malloc(in->nlin);
      err = (column == NULL);
      for (i = 0; !err && (i < in->nbnd); i++)
        {
          for (k = 0; !err && (k < in->npix); k++)
            {
              for (j=0; !err && (j < in->nlin); j++)
                {
                  column[j] = in->data[0][j][k];
                }
              n_written = fwrite (column, in->nlin, 1, fp);
              err = (n_written != 1);
            }
        }
      free (column);
      if (err)
        {
          snprintf (msg, SLEN - 1,
                    " Can't write byte mosaic image data to %s.", name);
          MESSAGE ('E', msg);
        }
    }
  err = fclose(fp);
  if (err)
    {
      snprintf (msg, SLEN - 1, " Can't close byte mosaic image %s.", name);
      MESSAGE ('W', msg);
    }
}
