#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        <sadie.h>
#include        "sadie_byte.h"
#include        "mosaic.h"
#include        "math.h"



/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function performs a moving window average on a byte mosaic image, and          */
/*   resamples the new image at specified intervals.                          */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void MOSAIC_BYTE_RESAMPL (
IMAGE_BYTE *in,      /*  I   Pointer to the input byte mosaic image.                           */
MOSAIC_INDEX *index, /* I Pointer to the input mosaic index.  */
uint32_t jsize,    /*  I   Number of lines in the window.                        */
uint32_t ksize,    /*  I   Number of pixels/line in the window.                  */
uint32_t jinc,     /*  I   Line resampling increment.                            */
uint32_t kinc,     /*  I   Pixels/line resampling increment.                     */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { uint32_t i, j, k, l, m, nlin, npix, pix, lin;
    char   msg[SLEN];
    double subsum, area=(double)(jsize*ksize);

    int progmeter;
    double progress;
    int count, total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Resample Byte Mosaic");
    progress = 0.0;


    /* if (TIMES) TIMING(T_RESAMPL); */
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"MOSAIC_BYTE_RESAMPL");
        MESSAGE('I',"");
        sprintf(msg," Input image:                  %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Window size: lines:           %d",jsize);
        MESSAGE('I',msg);
        sprintf(msg,"              pixels:          %d",ksize);
        MESSAGE('I',msg);
        sprintf(msg," Subsample increment: lines:   %d",jinc);
        MESSAGE('I',msg);
        sprintf(msg,"                      pixels:  %d",kinc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* check input */
    if (!CHECKIMG_BYTE(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (jsize == 0) {
        MESSAGE('E'," Number of lines in the averaging window must be greater than zero.");
        goto the_end;
    } else if (in->nlin < jsize) {
        MESSAGE('E'," Image size must be equal to or greater than window size in the line dimension.");
        goto the_end;
    } else if (ksize == 0) {
        MESSAGE('E'," Number of pixels/line in the averaging window must be greater than zero.");
        goto the_end;
    } else if (in->npix < ksize) {
        MESSAGE('E'," Image size must be equal to or greater than window sizein the pixel dimension.");
        goto the_end;
    } else if (jinc == 0) {
        MESSAGE('E'," Line subsample increment must be greater than zero.");
        goto the_end;
    } else if (kinc == 0) {
        MESSAGE('E'," Pixel subsample increment must be greater than zero.");
        goto the_end;
    } else if (!CHECKIMG_MOSAIC_INDEX(index)) {
        MESSAGE('E'," Can't identify mosaic index.");
        goto the_end;
    } else if (index->npix != in->npix) {
        MESSAGE('E'," Mosaic image and index are not the same size.");
        goto the_end;
    }

    /* create image of appropriate size */
    nlin = (index->nlin-jsize)/jinc + 1;
    npix = (index->npix-ksize)/kinc + 1;
    printf("index nlin = %d, index npix = %d\n",index->nlin,index->npix);
    printf("nlin = %d, npix = %d\n",nlin,npix);
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,nlin,npix,out);
    if (!*out) goto the_end;

    /* perform moving average and resample */
    if (progmeter) {
        count = 0;
        total = in->nbnd * (npix * nlin);
        increment = rint((double)total/20.0);
    }
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<nlin; j++) {
            for (k=0; k<npix; k++) {
                if (progmeter) {
                    count++;
                    if ((count%increment) == 0) {
                        progress = ((double)count/(double)total)*100.0;
                        cancel = ProgMeter_Update(progmeter,progress);
                        if (cancel != 0) goto the_end;
                    }
                }

                for (subsum=0.0,m=0; m<ksize; m++) {
                    pix = k*kinc+m;
                    for (l=0; l<jsize; l++) {
                        lin = (j*jinc+l)-index->voffset[pix];
                        if (0<lin && lin<in->nlin) {
                            subsum += (double)in->data[i][lin][k*kinc+m];
                        }
                    }
                }
                (*out)->data[i][j][k] = (PIXEL)(subsum/area);
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
    
    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }
    if (cancel != 0) {
        if (*out) RELMEM(*out);
        *out = NULL;
    }

}



/*-Copyright Information---------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab    */
/*-------------------------------------------------------------------------------*/
/*-General Information-----------------------------------------------------------*/
/*                                                                               */
/*   This function performs a moving window average on a byte mosaic image, and  */
/*   resamples the new image at specified intervals. Returns a byte mosaic.      */
/*                                                                               */
/*-------------------------------------------------------------------------------*/
/*-Interface Information---------------------------------------------------------*/
void MOSAIC2MOSAIC_RESAMPL (
IMAGE_BYTE *in,      /*  I   Pointer to the input byte mosaic image.             */
MOSAIC_INDEX *index, /*  I   Pointer to the input mosaic index.                  */
uint32_t jsize,         /*  I   Number of lines in the window.                      */
uint32_t ksize,         /*  I   Number of pixels/line in the window.                */
uint32_t jinc,          /*  I   Line resampling increment.                          */
uint32_t kinc,          /*  I   Pixels/line resampling increment.                   */
IMAGE_BYTE **out     /*  O   Address of a pointer to the output mosaic image.    */
/*-------------------------------------------------------------------------------*/
) { uint32_t i, j, k, l, m, nlin, npix;
    char   msg[SLEN];
    double subsum, area=(double)(jsize*ksize);

    /* if (TIMES) TIMING(T_RESAMPL); */
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"MOSAIC2MOSAIC_RESAMPL");
        MESSAGE('I',"");
        sprintf(msg," Input image:                  %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Window size: lines:           %d",jsize);
        MESSAGE('I',msg);
        sprintf(msg,"              pixels:          %d",ksize);
        MESSAGE('I',msg);
        sprintf(msg," Subsample increment: lines:   %d",jinc);
        MESSAGE('I',msg);
        sprintf(msg,"                      pixels:  %d",kinc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* check input */
    if (!CHECKIMG_BYTE(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (jsize == 0) {
        MESSAGE('E'," Number of lines in the averaging window must be greater than zero.");
        goto the_end;
    } else if (in->nlin < jsize) {
        MESSAGE('E'," Image size must be equal to or greater than window size in the line dimension.");
        goto the_end;
    } else if (ksize == 0) {
        MESSAGE('E'," Number of pixels/line in the averaging window must be greater than zero.");
        goto the_end;
    } else if (in->npix < ksize) {
        MESSAGE('E'," Image size must be equal to or greater than window sizein the pixel dimension.");
        goto the_end;
    } else if (jinc == 0) {
        MESSAGE('E'," Line subsample increment must be greater than zero.");
        goto the_end;
    } else if (kinc == 0) {
        MESSAGE('E'," Pixel subsample increment must be greater than zero.");
        goto the_end;
    } else if (!CHECKIMG_MOSAIC_INDEX(index)) {
        MESSAGE('E'," Can't identify mosaic index.");
        goto the_end;
    } else if (index->npix != in->npix) {
        MESSAGE('E'," Mosaic image and index are not the same size.");
        goto the_end;
    }

    /* create mosaic image of appropriate size */
    nlin = (index->height_max-jsize)/jinc + 1;
    npix = (index->npix-ksize)/kinc + 1;
    if (!CHECKIMG_BYTE(*out)) GETMEM_BYTE(in->nbnd,nlin,npix,out);
    if (!*out) goto the_end;

    /* perform moving average and resample */
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<nlin; j++) {
            for (k=0; k<npix; k++) {
                for (subsum=0.0,m=0; m<ksize; m++) {
                    for (l=0; l<jsize; l++) {
			subsum += (double)in->data[i][j*jinc+l][k*kinc+m];
                    }
                }
                (*out)->data[i][j][k] = (PIXEL)(subsum/area);
            }
        }
    }

    the_end:

}
