#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "trees.h"
#include "histogram.h"
#include <math.h>
#include <sys/time.h>

/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:  This procedure returns gradient magnitude and direction     */
/*                mosaics of prospective edge pixels.                         */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void MOSAIC_BYTE_SUPPRESS (
IMAGE_BYTE  *in,       /*  I   Pointer to the input mosaic image.             */
MOSAIC_INDEX  *index,  /*  I   Pointer to the input mosaic index.             */
IMAGE_BYTE  *iMag,     /*  I/O Pointer to the magnitude mosaic.               */
IMAGE_BYTE  *iDir,      /*  I/O Pointer to the direction mosaic.               */
HISTOGRAM  *ghist,     /*  O   Pointer to the global histogram.               */ 
HISTOGRAM  ***lhist    /*  O   Pointer to the local histogram.                */
/*----------------------------------------------------------------------------*/
) { register int i, j, k;
    int nrows, ncols;
    HISTOGRAM globalhist;
    HISTOGRAM **localhist;
    struct timeval start, end;
    /* check input */
    if (!CHECKIMG_BYTE(in)) {
        MESSAGE('E'," Can't identify mosaic image.");
        goto the_end;
    } else if (!CHECKIMG_MOSAIC_INDEX(index)) {
        MESSAGE('E'," Can't identify mosaic index.");
        goto the_end;
    } else if (!CHECKIMG_BYTE(iMag)) {
        MESSAGE('E'," Can't identify mosaic image: gradient.");
        goto the_end;
    } else if (!CHECKIMG_BYTE(iDir)) {
        MESSAGE('E'," Can't identify mosaic image: direction.");
        goto the_end;
    } else if(in->nlin != iMag->nlin || in->nlin != iDir->nlin ||
	      in->npix != iMag->npix || in->npix != iDir->npix) {
        MESSAGE('E'," Size of mosaics differ.");
        goto the_end;
    }

    
    /* Determine the number of blocks needed to cover the ROI */
    nrows = index->roi->bottom[0] - index->roi->top[0] + 1;
    ncols = index->hend - index->hstart + 1;
    nrows = ceil((double)nrows/HIST_BLOCK_DIM);
    ncols = ceil((double)ncols/HIST_BLOCK_DIM);

    /* Allocate memory */
    localhist = (HISTOGRAM **)MATRIX(nrows, ncols, 0, 0, sizeof(HISTOGRAM));

    /* Initialize histograms */
    HIST_INIT(&globalhist);
    for(i = 0; i < nrows; i++)
	for(j = 0; j < ncols; j++)
	    HIST_INIT(localhist[i]+j);

    /* Compute magnitude histogram */
    gettimeofday(&start, NULL);
    HIST_MAG(iMag, index, &globalhist, localhist, nrows, ncols);
    gettimeofday(&end, NULL);
    printf("    magnitude histogram = %ld ms.\n", delay(start, end));

    /* Compute the Rth percentile of magnitude histogram */
    gettimeofday(&start, NULL);
    HIST_RTH_PERCENTILE(&globalhist, localhist, MAG_PERCENTILE, nrows, ncols);
    gettimeofday(&end, NULL);
    printf("    compute Rth percentile = %ld ms.\n", delay(start, end));

    /* Compute direction histogram */
    gettimeofday(&start, NULL);
    HIST_DIR(iMag, iDir, index, &globalhist, localhist, nrows, ncols);
    gettimeofday(&end, NULL);
    printf("    direction histogram = %ld ms.\n", delay(start, end));
	
    /* Global & local orientation */
    gettimeofday(&start, NULL);
    HIST_ORIENTATION(&globalhist, localhist, nrows, ncols);
    gettimeofday(&end, NULL);
    printf("    compute ring orientation = %ld ms.\n", delay(start, end));

    /* Resolve gradient direction in the blocks */
    gettimeofday(&start, NULL);
    HIST_RESOLVE(&globalhist, localhist, nrows, ncols);
    gettimeofday(&end, NULL);
    printf("    resolve ring orientation = %ld ms.\n", delay(start, end));

    /* Suppress non-edge pixels */
    gettimeofday(&start, NULL);
    HIST_SUPPRESS(iMag, iDir, index, localhist);
    gettimeofday(&end, NULL);
    printf("    suppress non-edge pixels = %ld ms.\n", delay(start, end));

    for(k = index->hstart; k <= index->hend; k++) {
	iMag->data[0][index->vstart[k] - index->voffset[k]][k] = 0;
	iMag->data[0][index->vend[k] - index->voffset[k]][k] = 0;
    }

    ghist = &globalhist;
    lhist = &localhist;
    
    the_end:
}
