#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "trees.h"
#include <math.h>
#include <sys/time.h>


/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:  This procedure returns gradient magnitude and direction     */
/*                mosaics.                                                    */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void MOSAIC_BYTE_TRACKRINGS (
IMAGE_BYTE  *in,       /*  I   Pointer to the input mosaic image.             */
MOSAIC_INDEX  *index,  /*  I   Pointer to the input mosaic index.             */
IMAGE_BYTE  **iMag,    /*  O   Address of a pointer to the magnitude mosaic.  */
IMAGE_BYTE  **iDir     /*  O   Address of a pointer to the direction mosaic.  */
/*----------------------------------------------------------------------------*/
) { register int i, j, k;
    struct  timeval start, end, t1, t2;
    IMAGE *curframe=NULL;        
    IMAGE *gradmag=NULL;
    IMAGE *graddir=NULL;
    IMAGE *mag=NULL;         
    IMAGE *dir=NULL;         
    LOCAL_INDEX *localindex = NULL;

    int progmeter;
    double progress;
    double localpercent;
    double localprogress;
    int total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Compute Tree Ring Magnitudes");
    progress = 0.0;
    
    *iMag = NULL;
    *iDir = NULL;


    /* check input */
    if (!CHECKIMG_BYTE(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (!CHECKIMG_MOSAIC_INDEX(index)) {
        MESSAGE('E'," Can't identify mosaic index.");
        goto the_end;
    } else if (in->npix != index->npix) {
        MESSAGE('E'," Size of mosaic index does not match size of mosaic.");
        goto the_end;
    } else if (in->nbnd > 1) {
        MESSAGE('E'," Only first band of input image will be used.");
    }
    
    
    /* Allocate memory to hold the gradient magnitude and direction mosaics */
    if (!CHECKIMG_BYTE(*iMag)) GETMEM_BYTE(1,in->nlin,in->npix,iMag);
    if (!(*iMag)) goto the_end;
    
    if (!CHECKIMG_BYTE(*iDir)) GETMEM_BYTE(1,in->nlin,in->npix,iDir);
    if (!(*iDir)) goto the_end;


    /* Initialize mosaics to zero */
    for (j=0; j<in->nlin; j++) {
        for (k=0; k<in->npix; k++) {

            (*iMag)->data[0][j][k] = (PIXEL_BYTE)0;
            (*iDir)->data[0][j][k] = (PIXEL_BYTE)0;
        }
    }

    if (progmeter) {
        localpercent = 1.0;
        localprogress = 0.0;
	total = index->nframe;
        increment = 1;
    }
   
    /* Detect prospective edge pixels one frame at a time */
    for (i=0; i<index->nframe; i++) {

        if (progmeter) {
            if ((i%increment) == 0) {
                localprogress = ((double)i/(double)total)*100.0;
                cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
                if (cancel != 0) goto the_end;
            }
        }        

	/* No region of interest in current frame */
	if (index->frame[i].endx < index->hstart)
	    continue;
	if (index->frame[i].startx > index->hend)
	    continue;
	
        printf("Finding gradient in frame %d\n",i);
        gettimeofday(&start, NULL);

	/* Extract current frame from the mosaic */
        gettimeofday(&t1, NULL);
        MOSAIC_BYTE_EXTRACT(in, index,
			    index->frame[i].starty-6, index->frame[i].endy+6,
			    index->frame[i].startx-6, index->frame[i].endx+6,
			    (PIXEL)127.0, &curframe);
        gettimeofday(&t2, NULL);
        printf("    extract = %ld ms.\n", delay(t1, t2));

	/* Compute region of interest in the current frame */
	COMPUTE_LOCAL_INDEX(index, i,
			    index->frame[i].starty-6, index->frame[i].endy+6,
			    index->frame[i].startx-6, index->frame[i].endx+6,
			    &localindex);

        /* Return gradient magnitude and direction of prospective edge pixels */
        /* Use a gaussian of size 5 and std. deviation 3.0 */
        gettimeofday(&t1, NULL);
        TREES_TREERINGMAG(curframe, localindex, 3.0, 5, &gradmag, &graddir);
        gettimeofday(&t2, NULL);
        printf("    treeringmag = %ld ms.\n", delay(t1, t2));

        if (!CHECKIMG(gradmag) || !CHECKIMG(gradmag)) {
            MESSAGE('E'," Problem determining gradient magnitude and direction.");
            goto the_end;
        } else {
            sprintf(gradmag->text, "%s - Ring Magnitude", in->text);
            sprintf(graddir->text, "%s - Ring Direction", in->text);
        }



	/* Trim the images */
        gettimeofday(&t1, NULL);
        TRIM(gradmag,6,&mag);
        TRIM(graddir,6,&dir);
        gettimeofday(&t2, NULL);
        printf("    trim = %ld ms.\n", delay(t1, t2));

	
        if (!CHECKIMG(mag) || !CHECKIMG(dir)) {
            MESSAGE('E'," Problem trimming magnitude and direction images.");
            goto the_end;
        } else {
            sprintf(mag->text, "%s - Ring Magnitude", in->text);
	    sprintf(dir->text, "%s - Ring Direction", in->text);
        }
	
        if (CHECKIMG(gradmag))  RELIMG(&gradmag);
        gradmag = NULL;
        if (CHECKIMG(graddir))  RELIMG(&graddir);
        graddir = NULL;

	
	/* Insert the images into respective mosaics */
        gettimeofday(&t1, NULL);
        MOSAIC_BYTE_INSERT(mag, index, index->frame[i].starty, index->frame[i].startx, iMag);
        MOSAIC_BYTE_INSERT(dir, index, index->frame[i].starty, index->frame[i].startx, iDir);
        gettimeofday(&t2, NULL);
        printf("    insert frame = %ld ms.\n", delay(t1, t2));

        if (CHECKIMG(mag))  RELIMG(&mag);
        if (CHECKIMG(dir))  RELIMG(&dir);
        mag = NULL;
        dir = NULL;

        printf("Done finding gradient in frame %d\n",i);

	gettimeofday(&end, NULL);
        printf("Time required for frame %d= %ld ms.\n\n", i,delay(start, end));


	if (CHECKIMG(curframe))  RELIMG(&curframe);
        curframe = NULL; 
	if (localindex) {
	  free(localindex->vstart);
	  free(localindex->vend);
	  free(localindex);
	  localindex = NULL; 
	}
	  
    }    

    if (progmeter) {
	cancel = ProgMeter_Update(progmeter,100.0);
	if (cancel != 0) goto the_end;
    }        
    

    the_end:

    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }

    /* if (TIMES)   TIMING(T_EXIT); */
    if (CHECKIMG(curframe))  RELIMG(&curframe);
    if (CHECKIMG(mag))  RELIMG(&mag);
    if (CHECKIMG(dir))  RELIMG(&dir);
    if (CHECKIMG(gradmag))  RELIMG(&gradmag);
    if (CHECKIMG(graddir))  RELIMG(&graddir);

}
