#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        <stdio.h>
#include        <sadie.h>
#include        "mosaic.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */






/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This low-level function allocates memory for a new mosaic index and      */
/*   initializes its data structure.  It is based on the SADIE function       */
/*   GETMEM.                                                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void GETMEM_MOSAIC_INDEX(
int nlin,	  /*  I   Number of lines in mosaic.				*/
int npix,	  /*  I   Number of pixels/line.				*/
uint32_t	nframe,	  /*  I   Number of frames.				*/
MOSAIC_INDEX **out    /*  O   Address of a pointer to the mosaic index being allocated.    */
)
/*----------------------------------------------------------------------------*/
{   register int i;
    long total;
    BYTE *ptr=NULL;

    /* allocate space for image data structure, image data and list entry */
    total = sizeof(MOSAIC_INDEX) + 2*npix*sizeof(uint32_t) + 2*npix*sizeof(uint32_t) + (nframe*4)*sizeof(uint32_t) + sizeof(REGION_OF_INTEREST);
    if (total < 0L) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    if (!(ptr=(BYTE *)malloc(total*sizeof(BYTE)))) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* assign (DEFAULT) values */
    *out = (MOSAIC_INDEX *)ptr;
    (*out)->nlin = nlin;
    (*out)->npix = npix;
    (*out)->height_max = 0;
    (*out)->nframe = nframe;
    (*out)->voffset = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX));
    (*out)->vheight = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX) + npix*sizeof(uint32_t));
    (*out)->hstart = 0;
    (*out)->hend = npix - 1;
    (*out)->vstart = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX) + 2*npix*sizeof(uint32_t));
    (*out)->vend = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX) + 3*npix*sizeof(uint32_t));
    (*out)->roi = (REGION_OF_INTEREST *)(ptr + sizeof(MOSAIC_INDEX) + 4*npix*sizeof(uint32_t));
    (*out)->frame = (MOSAIC_FRAME *)(ptr + sizeof(MOSAIC_INDEX) + 4*npix*sizeof(uint32_t) + sizeof(REGION_OF_INTEREST));

    
    /* initialize arrays */
    for (i=0; i<npix; i++) {
        (*out)->voffset[i] = 0;
        (*out)->vheight[i] = 0;
        (*out)->vstart[i] = 0;
        (*out)->vend[i] = 0;
    }
    for (i=0; i<nframe; i++) {
        (*out)->frame[i].startx = 0;
        (*out)->frame[i].endx = 0;
        (*out)->frame[i].starty = 0;
        (*out)->frame[i].endy = 0;
    }


    the_end:
    return;
    /* if (TIMES) TIMING(T_EXIT); */
}

/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function releases a mosaic_index to free memory space.              */
/*   It is based on the SADIE function RELIMG.                                */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void RELIMG_MOSAIC_INDEX(
MOSAIC_INDEX **out    /* I/O  Address of the pointer to the mosaic index to be released.   */
)
/*----------------------------------------------------------------------------*/
{
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"RELIMG_MOSAIC_INDEX");
    }

    /* check input */
    if (!CHECKIMG_MOSAIC_INDEX(*out)) {
        MESSAGE('E'," Can't identify mosaic index.");
        goto the_end;
    }

    /* release image */
    RELMEM_MOSAIC_INDEX(*out);
    *out = NULL;

    the_end:
    return;
    /* if (TIMES) TIMING(T_EXIT); */
}
 
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This low-level function releases a mosaic index and frees its data        */
/*   structure.  It is based on the SADIE function RELMEM.                    */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void RELMEM_MOSAIC_INDEX(
MOSAIC_INDEX *in      /*  I   Pointer to the mosaic index to be released.     */
)
/*----------------------------------------------------------------------------*/
{   

    /* free allocated space */
    free(in);

    return;
    /* if (TIMES) TIMING(T_EXIT); */
}
 
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This low-level function determines whether or not a MOSAIC_INDEX pointer */
/*   points to an mosaic index that was properly created by SADIE.            */
/*   It is based on the SADIE routine CHECKIMG.                               */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int CHECKIMG_MOSAIC_INDEX(
MOSAIC_INDEX *in      /*  I   Pointer to the mosaic index to be checked.      */
)
/*----------------------------------------------------------------------------*/
{   

    return(in!=NULL);
}


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the scaled version of actual mosaic index.        */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void INDEX_RESAMPL(
MOSAIC_INDEX *index,
int scale,
MOSAIC_INDEX ** outindex
)
/*----------------------------------------------------------------------------*/
{
    register int i, npix, nlin, nframe;
    long total;
    BYTE *ptr=NULL;

    nlin = index->nlin/scale;
    npix = index->npix/scale;
    nframe = index->nframe;
    
    /* allocate space for image data structure, image data and list entry */
    total = sizeof(MOSAIC_INDEX) + 2*npix*sizeof(uint32_t) + 2*npix*sizeof(uint32_t) + (nframe*4)*sizeof(uint32_t) + sizeof(REGION_OF_INTEREST);
    if (total < 0L) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    if (!(ptr=(BYTE *)malloc(total*sizeof(BYTE)))) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* assign (DEFAULT) values */
    *outindex = (MOSAIC_INDEX *)ptr;
    (*outindex)->nlin = nlin;
    (*outindex)->npix = npix;
    (*outindex)->height_max = index->height_max/scale;
    (*outindex)->nframe = nframe;
    (*outindex)->voffset = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX));
    (*outindex)->vheight = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX) + npix*sizeof(uint32_t));
    (*outindex)->hstart = index->hstart/scale + 1;
    (*outindex)->hend = index->hend/scale - 1;
    (*outindex)->vstart = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX) + 2*npix*sizeof(uint32_t));
    (*outindex)->vend = (uint32_t *)(ptr + sizeof(MOSAIC_INDEX) + 3*npix*sizeof(uint32_t));
    (*outindex)->roi = (REGION_OF_INTEREST *)(ptr + sizeof(MOSAIC_INDEX) + 4*npix*sizeof(uint32_t));
    (*outindex)->frame = (MOSAIC_FRAME *)(ptr + sizeof(MOSAIC_INDEX) + 4*npix*sizeof(uint32_t) + sizeof(REGION_OF_INTEREST));

    
    /* initialize arrays */
    for (i=0; i<npix; i++) {
	if (i*scale >= index->npix)
	    break;
        (*outindex)->voffset[i] = index->voffset[i*scale]/scale;
        (*outindex)->vheight[i] = index->vheight[i*scale]/scale;
        (*outindex)->vstart[i] = index->vstart[i*scale]/scale + 1;
        (*outindex)->vend[i] = index->vend[i*scale]/scale - 1;
    }
    for (i=0; i<nframe; i++) {
        (*outindex)->frame[i].startx = index->frame[i].startx/scale;
        (*outindex)->frame[i].endx = index->frame[i].endx/scale;
        (*outindex)->frame[i].starty = index->frame[i].starty/scale;
        (*outindex)->frame[i].endy = index->frame[i].endy/scale;
    }
    for (i=0; i<2; i++) {
	(*outindex)->roi->top[i] = index->roi->top[i]/scale;
	(*outindex)->roi->bottom[i] = index->roi->bottom[i]/scale;
	(*outindex)->roi->left[i] = index->roi->left[i]/scale;
	(*outindex)->roi->right[i] = index->roi->right[i]/scale;
    }
    
    the_end:
    
}
