/** @file ppdev_motion.c
 * Motion control for stages driven by unintelligent stepper motors.
 * Some stages rely on the controlling computer to handle the
 * low-level details of the motors that drive them, in contrast to
 * those that dedicate an intelligent controller to these tasks, with
 * which the controlling computer can communicate using some
 * high-level protocol over something like a serial interface.
 * 
 * This particular implementation is for stages driven by stepper
 * motors connected directly to parallel ports.  Individual phases on
 * the motor are either active or inactive depending on whether a
 * particular bit on a parallel interface is asserted or not.  The
 * software can rotate the motor only by suppling a sequence of bit
 * patterns that will cause the motor phases to activate and
 * de-activate in the correct order.  Apart from possibly
 * microswitches that change input bits on the parallel interface when
 * the stage reaches a home position, only a count of the steps moved
 * by each motor gives any indication of the actual stage position.
 * Individual motors drive each axis of the stage, but more than one
 * motor may share a single port, provided we can assign each a
 * distinct group of bits within the port.
 *
 * We have based the low-level parallel port control on the @c ppdev
 * driver, which exposes the necessary details of the port to
 * user-level programs under Linux.  Unlike the Solaris version of the
 * motion control code, this means we do not have to provide a
 * custom-written kernel module; documentation for @c ppdev is
 * included as part of the Linux kernel source, and Tim Waugh (one of
 * the original authors) provides a manual at
 * <a href="http://people.redhat.com/twaugh/parport/html/parportguide.html">http://people.redhat.com/twaugh/parport/html/parportguide.html</a>
 * for the Linux 2.4 kernel version.  We rely on a Posix Threads
 * implementation for two important aspects of the stage motion
 * control: a distinct thread becomes responsible for handling each
 * axis of the stage, and variations in the timeouts as the threads
 * wait on condition variables determine the variations in the motor
 * speed.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2003
 * @version $Id: ppdev_motion.c,v 1.6 2004/11/18 19:14:06 mmunro Exp $
 */

/* This file is part of TREES.

   TREES is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   TREES is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with TREES; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <pthread.h>
#include <linux/ppdev.h>
#include <linux/parport.h>
#include "logerr.h"
#include "motion.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** All the stage axis labels are immutable strings. */
typedef const char *stagelabelp_t;
/** The X-axis label. */
static stagelabelp_t const xlabel = "X";
/** The Y-axis label. */
static stagelabelp_t const ylabel = "Y";
/** The Z-axis label (focusing uses Z-axis displacements). */
static stagelabelp_t const zlabel = "focus";

/** Representation of bit patterns controlling the stepper motor phases. */
typedef const unsigned char axistates_t;
/** Number of distinct states on the low bits of the port data register. */
static const unsigned int nlostates = 8;
/** Number of distinct states on the high bits of the port data register. */
static const unsigned int nhistates = 8;
/** Parallel port data register state for which no motors are active. */
static axistates_t offstate = '\xff';
/** Ordered sequence of states for the low bits of the port data register. */
static axistates_t *const lostate = "\x0e\x0c\x0d\x09\x0b\x03\x07\x06";
/** Ordered sequence of states for the high bits of the port data register. */
static axistates_t *const histate = "\xe0\xc0\xd0\x90\xb0\x30\x70\x60";

/** The data transfer mode to negotiate on the parallel port. */
static const int ppmode = IEEE1284_MODE_COMPAT;
/** How to frob the parallel port control lines for low-power activation. */
static const struct ppdev_frob_struct activate = {
  PARPORT_CONTROL_STROBE, /**< Line selection mask. */
  0xff /**< Control line data. */
};
/** How to frob the parallel port control lines for normal activation. */
static const struct ppdev_frob_struct heavy_activate = {
  PARPORT_CONTROL_STROBE | PARPORT_CONTROL_INIT, /**< Line selection mask. */
  0xff /**< Control line data. */
};
/** Control lines to frob for parallel port deactivation. */
static const struct ppdev_frob_struct passivate = {
  0xff, /**< Line selection mask. */
  0xff /**< Control line data. */
};

static const int deltamag = 1;  /**< Nominal motor step size magnitude. */
static const long int nnano = 1000000000; /**< Nanoseconds in a second. */
static const long int nmicro = 1000000; /**< Microseconds in a second. */
static const long int nmilli = 1000; /**< Milliseconds in a second. */

/** Stage status flags. */
typedef enum
  {
    STAGE_DEAD  = 0,
    STAGE_LIVE  = 1,
    STAGE_COMMANDABLE = 1<<1,
    STAGE_WATCHABLE = 1<<2
  } stagecommandflags;

/**
 * The bit pattern currently applied to the parallel port data register.
 */
struct Devicestate
{
  int pportfd;  /**< File descriptor for @c /dev/parportN. */
  portdata_t portdata; /**< Apply this bit pattern to the device register. */
};

/**
 * Logical to physical axis mapping.
 * @note At the moment we have no more than two axes per port, so this is more
 * elaborate than necessary.
 */
struct Axismapping
{
  unsigned int naxes;  /**< The number of axes on a port. */
  axismapentry_t *mapping; /**< A vector of physical axis numbers. */
};

/**
 * Coordinates representing a position along multiple axes.
 */
struct Stagelocation
{
  unsigned int naxes;  /**< The first @a naxes axes have positions defined. */
  int relative;  /**< Currently unused? */
  coord_t *pos; /**< Vector of positions, in logical axis order. */
};

/**
 * Grouping of definitions for one parallel port and all its dependent axes.
 */
struct Stagecontrol
{
  pthread_mutex_t modlock;  /**< Lock for this and everything it contains. */
  pthread_mutexattr_t *mattr; /**< Will specify non-default mutex attributes. */
  pthread_cond_t offtrack;  /**< Signalled when the motor needs to move. */
  pthread_cond_t ontrack;   /**< Signalled when reaching a resting state. */
  int ctrlstatus;  /**< Set of flags shewing the stage control status. */
  unsigned int naxes;  /**< Total number of axes using the port. */
  devicestatep_t devstate;  /**< Bit patterns and port file descriptor. */
  translationp_t axis;  /**< Vector of axis control structure pointers. */
};


/**
 * Everything needed to control a translational stage axis.
 */
struct Translation
{
  stagelabelp_t label;  /**< Intelligible axis label. */
  pthread_t tid;      /**< Identity of the thread dedicated to this axis. */
  pthread_attr_t *tattr; /**< Will specify non-default thread attributes. */
  int (*step) (translationp_t); /**< Function that moves the motor one step. */
  int responsive; /**< Non-zero iff the axis should respond to control. */
  int delta;  /**< Direction of motion (as +1 or -1). */
  int backlash;  /**< Backlash compensation state. */
  int maxlash;  /**< Maximum number of backlash compensation steps. */
  int drivestate; /**< Index into the bit-pattern vector (current pattern). */
  coord_t pos;  /**< Current axis position, in arbitrary step units. */
  coord_t goalpos;  /**< Desired axis position, in arbitrary step units. */
  wait_t delay;  /**< The wait between motor steps (determines speed). */
  wait_t maxdelay;  /**< The maximum allowed wait (determines minimum speed). */
  wait_t mindelay;  /**< The minimum allowed wait (determines maximum speed). */
  double accel;  /**< The rate at which to decrease @a delay. */ 
  double braking; /**< The rate at which to increase @a delay. */
  stagecontrolp_t controller; /**< The axis is associated with this stage. */
};

/* Resting condition for a translation axis, a. */
/* Even if the inline version works, respect the semantics of the macro:
   don't do things like resting_condition (axis++). */
#if HAVE_INLINE
static inline int
resting_condition (const translationp_t a)
{
  return ((a->pos == a->goalpos) &&
          (a->delay >= a->maxdelay) &&
          ((a->delta < 0) == (a->backlash == 0)) &&
          ((a->delta > 0) == (a->backlash == a->maxlash)));
}
#else /* !HAVE_INLINE */
#define resting_condition(a) ((a->pos == a->goalpos) && \
                      (a->delay >= a->maxdelay) && \
                      ((a->delta < 0) == (a->backlash == 0)) && \
                      ((a->delta > 0) == (a->backlash == a->maxlash)))
#endif /* !HAVE_INLINE */

/**
 * Make the non-default thread creation attributes we require.
 * The threads controlling each axis must be created detatched, so the default
 * thread creation attributes will not do: we create our own custom attributes.
 * @return the newly-allocated creation attributes.
 * @see disposetattr()
 */
static pthread_attr_t *
newdetachedtattr (void)
{
  pthread_attr_t *attr;
  int err = 0;

  attr = (pthread_attr_t *) malloc (sizeof (pthread_attr_t));
  if (attr != NULL)
    {
      err = pthread_attr_init (attr);
      if (!err)
        {
          err = pthread_attr_setdetachstate (attr, PTHREAD_CREATE_DETACHED);
          if (err)
            pthread_attr_destroy (attr);
        }
      if (err)
        {
          free (attr);
          attr = NULL;
        }
    }
  if (attr == NULL)
    logerr_crit (LOGERR_FUNC,
                 "could not create default thread attributes; error %d", err);
  return attr;
}

/**
 * Destroy explicitly specified thread creation attributes.
 * @param theattr  the attributes to destroy.
 * @see newdetachedtattr()
 */
static void
disposetattr (pthread_attr_t * theattr)
{
  int err = 0;

  if (theattr != NULL)
    {
      err = pthread_attr_destroy (theattr);
      if (err)
        logerr_err (LOGERR_FUNC,
                    "could not dispose of the attributes, error %d", err);
      else
        free (theattr);
    }
}

/**
 * Make the customized attributes needed for initializing the mutex.
 * The threads controlling each axis must be created detatched, so the default
 * thread creation attributes will not do: we create our own custom attributes.
 * @return the newly-allocated creation attributes.
 * @see disposetattr()
 */
static pthread_mutexattr_t *
newcustommattr (void)
{
  pthread_mutexattr_t *attr;
  int err = 0;

  attr = malloc (sizeof (pthread_mutexattr_t));
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (attr != NULL));
  err = pthread_mutexattr_init (attr);
  if (!err)
    {
      err = pthread_mutexattr_settype (attr, PTHREAD_MUTEX_ERRORCHECK_NP);
      if (err)
        pthread_mutexattr_destroy (attr);
    }
   if (err)
    {
      free (attr);
      attr = NULL;
      logerr_crit (LOGERR_FUNC,
                   "could not create custom mutex attributes; error %d", err);
    }
  return attr;
}

/**
 * Destroy custom mutex initialization attributes.
 * @param theattr  the attributes to destroy.
 * @see newcustommattr()
 */
static void
disposemattr (pthread_mutexattr_t * theattr)
{
  int err = 0;

  if (theattr != NULL)
    {
      err = pthread_mutexattr_destroy (theattr);
      if (err)
        logerr_err (LOGERR_FUNC,
                    "could not destroy the mutex attributes, error %d", err);
      else
        free (theattr);
    }
}

/**
 * Cause the motor driving a stage axis to speed up.
 * The motor speed is determined by the delay between steps,
 * so reduce this.
 * @param theaxis the axis that should accelerate.
 * @see brake()
 */
static void
accelerate (translationp_t theaxis)
{
  if (theaxis->delay == 0)
    theaxis->delay = theaxis->maxdelay;
  else
    {
      theaxis->delay -= theaxis->accel;
      if (theaxis->delay < theaxis->mindelay)
        theaxis->delay = theaxis->mindelay;
      else if (theaxis->delay > theaxis->maxdelay)
        theaxis->delay = theaxis->maxdelay;
    }
}

/**
 * Cause the motor driving a stage axis to slow down.
 * The motor speed is determined by the delay between steps, so
 * increase this.
 * @param theaxis the axis that should slow down.
 * @see accelerate()
 */
static void
brake (translationp_t theaxis)
{
  if (theaxis->delay == 0)
    theaxis->delay = theaxis->maxdelay;
  else
    {
      theaxis->delay += theaxis->braking;
      if (theaxis->delay < theaxis->mindelay)
        theaxis->delay = theaxis->mindelay;
      else if (theaxis->delay > (theaxis->maxdelay - theaxis->braking / 2))
        theaxis->delay = theaxis->maxdelay;
    }
}

/**
 * Generate absolute times usable for scheduling threads.
 * We represent delays by a scalar type @a wait_t, but need absolute times
 * for scheduling the time delays for sleeping threads.
 * @param  thedelay  the scalar version of a time delay.
 * @param  thewake   the corresponding absolute time (from now).
 * @return The system error code for obtaining the current time, 0 on success.
 */
static int
nextwake (const wait_t thedelay, struct timespec *thewake)
{
  struct timeval clocktime;
  long int tick;
  int err = 0;

  err = gettimeofday (&clocktime, NULL);
  if (err)
    logerr_crit (LOGERR_FUNC, "could not get the time; error %d", err);
  else
    {
      thewake->tv_sec = clocktime.tv_sec;
      tick = clocktime.tv_usec * nmilli + thedelay;
      if (tick >= nnano)
        {
          (thewake->tv_sec)++;
          tick -= nnano;
        }
      thewake->tv_nsec = tick;
    }
  return err;
}

/**
 * Find the interval between the specified absolute time and now.
 * @param  thewake  an absolute time.
 * @return a scalar representation of the delay until that time.
 */
static wait_t
waitremaining (const struct timespec *thewake)
{
  struct timeval clocktime;
  long int tick;
  wait_t w;
  int err = 0;

  w = 0;
  err = gettimeofday (&clocktime, NULL);
  if (err)
    logerr_crit (LOGERR_FUNC, "could not get the time; error %d", err);
  else
    {
      tick = (nnano * (thewake->tv_sec - clocktime.tv_sec)
              * (thewake->tv_nsec - nmilli * clocktime.tv_usec));
      w = (wait_t) ((tick > 0) ? tick : 0);
    }
  return w;
}

/**
 * Move (by one step) the motor connected to the low bits of a parallel port.
 * We're assuming we've already obtained a lock on the stage structure,
 * so are free to modify the representation of the parallel port.
 * Moving the motor (changing the bit pattern and advancing the
 * circular @a drivestate counter) may or may not change the axis
 * position, depending on the backlash compensation state, so we don't
 * deal with it here.  At the moment this is almost identical to
 * @a histep(), but need not be so.
 * @warning This will only work correctly if the stage mutex is locked!
 * @param  theaxis  the axis whose motor should move.
 * @return an error code on failure, zero on success.
 * @see histep()
 */
static int
lostep (translationp_t theaxis)
{
  devicestatep_t state;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((theaxis != NULL) && (theaxis->controller != NULL)
                  && (theaxis->controller->devstate != NULL)));
  state = theaxis->controller->devstate;
  theaxis->drivestate += theaxis->delta;
  if (theaxis->drivestate < 0)
    theaxis->drivestate += nlostates;
  else if (theaxis->drivestate >= nlostates)
    theaxis->drivestate -= nlostates;
  state->portdata = (state->portdata & 0xf0) | lostate[theaxis->drivestate];
  err = ioctl (state->pportfd, PPWDATA, &state->portdata);
  if (err)
    logerr_warning (LOGERR_FUNC,
                    "encountered an error when controlling the hardware port; axis label = %s, error = %d",
                    theaxis->label, err);
  return err;
}

/**
 * Move (by one step) the motor connected to the high bits of a parallel port.
 * We're assuming we've already obtained a lock on the stage structure,
 * so are free to modify the representation of the parallel port.
 * Moving the motor (changing the bit pattern and advancing the
 * circular @a drivestate counter) may or may not change the axis
 * position, depending on the backlash compensation state, so we don't
 * deal with it here.  At the moment this is almost identical to
 * @a lostep(), but need not be so.
 * @warning This will only work correctly if the stage mutex is locked!
 * @param  theaxis  the axis whose motor should move.
 * @return an error code on failure, zero on success.
 * @see lostep()
 */
static int
histep (translationp_t theaxis)
{
  devicestatep_t state;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((theaxis != NULL) && (theaxis->controller != NULL)
                  && (theaxis->controller->devstate != NULL)));
  state = theaxis->controller->devstate;
  theaxis->drivestate += theaxis->delta;
  if (theaxis->drivestate < 0)
    theaxis->drivestate += nhistates;
  else if (theaxis->drivestate >= nhistates)
    theaxis->drivestate -= nhistates;
  state->portdata = (state->portdata & 0x0f) | histate[theaxis->drivestate];
  err = ioctl (state->pportfd, PPWDATA, &state->portdata);
  if (err)
    logerr_warning (LOGERR_FUNC,
                    "encountered an error when controlling the hardware port; axis label = %s, error = %d",
                    theaxis->label, err);
  return  err;
}

/**
 * Track the changes needed to make actual axis states match requested states.
 * Each axis is controlled by its own thread, and each thread spends
 * its entire lifetime executing the outer loop within @a track().
 * Threads sleep until explicitly signalled via @a offtrack that they
 * must make the actual axis state match a request, communicated
 * through the settings of various members of the axis structure.  The
 * request is fulfilled when a single condition, @a resting_condition,
 * is true; this corresponds to the axis being at rest, with no
 * outstanding backlash corrections, at the position given by
 * @a goalpos.  On reaching this state, it signals anything waiting
 * via @a ontrack (although the final resting state may differ from
 * that originally requested by something waiting on @a ontrack).
 *
 * We would like to move the motor as fast as possible in the right
 * direction to make the actual axis position match the requested
 * goal.  But there are several mechanical limitations: the motor
 * torque needed to start the axis moving and continue at low speeds
 * is greater than that needed to sustain rapid movement, so motions
 * ideally start slowly, then accelerate; when the axis has moved in
 * one direction and must reverse to move in the opposite direction,
 * the motor may move to re-tension a drive belt or move back the
 * teeth of a cog without any corresponding change in the axis
 * position, so we must allow for this extra motion (known as backlash
 * compensation).  Slow motor speeds are required for backlash
 * compensation, as well as for starting and stopping the axis.
 * Because there is no direct control over how the requests are
 * signalled, there is no guarantee that the axis is at rest when a
 * request arrives (a possible advantage if we're about to collide
 * with an obstacle on the path to the requested goal position and the
 * revised request is to back off before this happens).  A revised
 * goal may complicate the sequence of motions, from moving in one
 * direction to braking to a halt, reversing, accelerating in the
 * opposite direction, and finally braking to a halt once more.
 *
 * A stepper motor only moves in discrete increments, so we control
 * its speed by adjusting the time delays between steps.  There will
 * be a maximum speed, corresponding to the maximum frequency at which
 * the motor can step with adequate torque, and if we exceed this
 * speed the axis may miss a step.  Accurate speed control is at the
 * mercy of thread scheduling: we're using a call to
 * @a pthread_cond_timedwait() to determine the delay between steps,
 * so if the thread is running at too low a priority, some of the
 * speeds may be slower than expected, and vary unpredictably.
 * Increasing the thread priority may solve this problem, at the
 * expense of overall system performance.
 *
 * @param  arg  the axis to be controlled (cast from @a translationp_t).
 * @return NULL (the thread only really returns when there is an error).
 */
static void *
track (void *arg)
{
  translationp_t theaxis;
  stagecontrolp_t thestage;
  struct timespec wakeup;
  wait_t dt;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (arg != NULL));
  theaxis = (translationp_t) arg;
  thestage = theaxis->controller;
  err = pthread_mutex_lock (&thestage->modlock);
  if (err)
    logerr_err (LOGERR_FUNC,
                "could not acquire the lock; axis label = %s, error = %d",
                theaxis->label, err);
  theaxis->responsive = (thestage->ctrlstatus & STAGE_LIVE);
  while (!err && theaxis->responsive)
    {
      while (!err && (thestage->ctrlstatus & STAGE_LIVE)
	     && resting_condition (theaxis))
        err = pthread_cond_wait (&thestage->offtrack, &thestage->modlock);
      while (!err && !(resting_condition (theaxis)))
        {
          logerr_debug (LOGERR_FUNC,
                        "the axis labelled %s is moving from %d to %d with delay offset = %d, delta = %d and backlash state = %d",
                        theaxis->label, theaxis->pos, theaxis->goalpos,
                        (int) (theaxis->maxdelay - theaxis->delay),
                        theaxis->delta, theaxis->backlash);
          err = nextwake (theaxis->delay, &wakeup);
          if (!err)
            {
              err = pthread_cond_timedwait (&thestage->offtrack,
                                            &thestage->modlock, &wakeup);
              while (!err && !(resting_condition (theaxis)))
                {               /* Not just a timeout: something signalled offtrack. */
                  dt = theaxis->delay - waitremaining (&wakeup);
                  if (dt < theaxis->mindelay)
                    {           /* Too soon after last move: wait before moving again. */
                      dt = theaxis->mindelay - dt;
                      err = nextwake (dt, &wakeup);
                      if (!err)
                        err = pthread_cond_timedwait (&thestage->offtrack,
                                                      &thestage->modlock,
                                                      &wakeup);
                    }
                }
              if (err == ETIMEDOUT)
                err = 0;
              /* The axis is ready for another move, or has already arrived (or an error). */
              if (!err && !(resting_condition (theaxis)))
                {               /* The axis needs to move. */
                  if (theaxis->delay >= theaxis->maxdelay)
                    {           /* Moving slowly enough for direction changes. */
                      if ((theaxis->delta *
                           (theaxis->goalpos - theaxis->pos)) < 0)
                        theaxis->delta = -theaxis->delta;       /* Change direction. */
                      if ((theaxis->delta > 0)
                          && (theaxis->backlash != theaxis->maxlash))
                        {       /* Forward backlash compensation. */
                          err = (*(theaxis->step)) (theaxis);
                          (theaxis->backlash)++;
                        }
                      else if ((theaxis->delta < 0)
                               && (theaxis->backlash != 0))
                        {       /* Backward backlash compensation. */
                          err = (*(theaxis->step)) (theaxis);
                          (theaxis->backlash)--;
                        }
                      else if (theaxis->pos != theaxis->goalpos)
                        {       /* No backlash compensation. */
                          err = (*(theaxis->step)) (theaxis);
                          theaxis->pos += theaxis->delta;
                          if (!err && (theaxis->pos != theaxis->goalpos))
                            accelerate (theaxis);
                        }
                    }
                  else
                    {           /* Moving too rapidly for direction changes. */
                      err = (*(theaxis->step)) (theaxis);
                      theaxis->pos += theaxis->delta;
                      if ((theaxis->maxdelay - theaxis->delay)
                          < (theaxis->braking * theaxis->delta
                             * (theaxis->goalpos - theaxis->pos)))
                        accelerate (theaxis);
                      else
                        brake (theaxis);
                    }
                }
            }
        }
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                     (err || resting_condition (theaxis)
		      || !(thestage->ctrlstatus & STAGE_LIVE)));
      theaxis->responsive = (thestage->ctrlstatus & STAGE_LIVE);
      pthread_cond_broadcast (&thestage->ontrack);
    }
  logerr_debug (LOGERR_FUNC,
                "exiting with the status or error code = %d", err);
  err = pthread_mutex_unlock (&thestage->modlock);
  if (err)
    logerr_err (LOGERR_FUNC,
                "could not release the lock; axis label = %s, error = %d",
                theaxis->label, err);
  return NULL;
}

/**
 * Initialize an axis and start its controlling thread.
 * @param theaxis  the axis to initialize.
 * @return an error code on failure, zero on success.
 */
static int
startaxis (translationp_t theaxis)
{
  stagecontrolp_t thestage;
  int err = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (theaxis != NULL));
  thestage = theaxis->controller;
  if ((err = pthread_create (&theaxis->tid, theaxis->tattr,
                                  track, (void *) theaxis)))
    logerr_err (LOGERR_FUNC,
                "could not create the thread; axis label = %s, error = %d",
                theaxis->label, err);
  return err;
}

/**
 * Wait for an axis to stop moving, then cancel its controlling thread.
 * @param theaxis  the axis to stop.
 * @return an error code on failure, zero on success.
 */
static int
stopaxis (translationp_t theaxis)
{
  int err = 0;
  int unlockfail = 0;
  stagecontrolp_t thestage;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (theaxis != NULL));
  logerr_debug (LOGERR_FUNC,
                "waiting for the axis labelled %s to stop", theaxis->label);
  thestage = theaxis->controller;
  err = pthread_mutex_lock (&thestage->modlock);
  if (err)
    {
      logerr_err (LOGERR_FUNC,
                  "could not acquire the lock; axis label = %s, error = %d",
                  theaxis->label, err);
      return err;
    }
  if (!err)
    {
      err = pthread_cond_broadcast (&thestage->offtrack);
      if (err)
        logerr_err (LOGERR_FUNC,
                    "could not cancel the thread; axis label = %s, error = %d",
                    theaxis->label, err);
    }
  while (!err && theaxis->responsive)
    err = pthread_cond_wait (&thestage->ontrack, &thestage->modlock);
  /* Try to unlock, even if there are errors. */
  unlockfail = pthread_mutex_unlock (&thestage->modlock);
  if (unlockfail)
    logerr_err (LOGERR_FUNC,
                "could not release the lock; axis label = %s, error = %d",
                theaxis->label, unlockfail);
  if (err)
    logerr_err (LOGERR_FUNC,
                "the axis has not stopped correctly; axis label = %s, error = %d",
                theaxis->label, err);
  else
    logerr_debug (LOGERR_FUNC, "the axis has now stopped");
  return (err) ? err : unlockfail;
}

/**
 * Make a new logical to physical axis mapping table.
 * The logical axes are numbered from 0 to @a axes - 1, and the
 * physical axes are initially assigned the same numbers.
 * @param axes  the number of axes in the mapping.
 * @return the newly-allocated mapping.
 * @see disposeaxismapping()
 * @see setaxismapping()
 */
axismappingp_t
newaxismapping (const unsigned int axes)
{
  axismappingp_t themap;
  int logical;

  themap = (axismappingp_t) malloc (sizeof (axismappingp_t));
  if (themap != NULL)
    {
      themap->naxes = axes;
      /* Note the intentional allocation of axes + 1 (one extra zero entry). */
      themap->mapping =
        (axismapentry_t *) calloc (axes + 1, sizeof (axismapentry_t));
      if (themap->mapping == NULL)
        {
          free (themap);
          themap = NULL;
        }
      else
        {
          for (logical = 0; logical < axes; logical++)
            themap->mapping[logical] = (axismapentry_t) logical;
        }
    }
  return themap;
}

/**
 * Destroy a logical to physical axis mapping table.
 * @param themap  the map to destroy.
 * @see newaxismapping()
 */
void
disposeaxismapping (axismappingp_t themap)
{
  if (themap != NULL)
    {
      free (themap->mapping);
      free (themap);
    }
}

/**
 * Change an entry in a logical to physical axis mapping table.
 * @param themap  the map to modify.
 * @param logical  the logical axis number.
 * @param physical  the new corresponding physical number.
 */
void
setaxismapping (axismappingp_t themap,
                const unsigned int logical, const axismapentry_t physical)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((themap != NULL) && (themap->mapping != NULL)));
  if (logical >= themap->naxes)
    logerr_err (LOGERR_FUNC,
                "trying to set axis %d when there are only %d axes",
                logical, themap->naxes);
  else
    themap->mapping[logical] = physical;
}

/**
 * Make a new generalized stage coordinate vector.
 * @param axes  the number of axes requiring positions.
 * @return the newly-allocated position information.
 * @see disposestagelocation()
 * @see setstagelocation()
 * @see getstagelocation()
 */
stagelocationp_t
newstagelocation (unsigned int axes)
{
  stagelocationp_t theloc;

  theloc = (stagelocationp_t) malloc (sizeof (stagelocationp_t));
  if (theloc != NULL)
    {
      theloc->naxes = axes;
      /* Note the intentional allocation of axes + 1 (one extra zero entry). */
      theloc->pos = (coord_t *) calloc (axes + 1, sizeof (coord_t));
      if (theloc->pos == NULL)
        {
          free (theloc);
          theloc = NULL;
        }
    }
  return theloc;
}

/**
 * Destroy a generalized stage coordinate vector.
 * @param theloc  the stage coordinates to destroy.
 * @see newstagelocation()
 */
void
disposestagelocation (stagelocationp_t theloc)
{
  if (theloc != NULL)
    {
      free (theloc->pos);
      free (theloc);
    }
}

/**
 * Change one coordinate in a generalized stage coordinate vector.
 * @param theloc  the coordinate vector to modify.
 * @param axisn  selects the logical axis number.
 * @param newval  the new position ascribed to this axis.
 * @see getstagelocation()
 */
void
setstagelocation (stagelocationp_t theloc,
                  const unsigned int axisn, const coord_t newval)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((theloc != NULL) & (theloc->pos != NULL)));
  if (axisn >= theloc->naxes)
    logerr_err (LOGERR_FUNC,
                "trying to set axis %d when there are only %d axes",
                axisn, theloc->naxes);
  else
    theloc->pos[axisn] = newval;
}

/**
 * Access one coordinate in a generalized stage coordinate vector.
 * @param theloc  the coordinate vector to examine.
 * @param axisn  selects the logical axis number.
 * @return the current position ascribed to this axis. 
 * @see setstagelocation()
 */
coord_t
getstagelocation (const stagelocationp_t theloc, const unsigned int axisn)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((theloc != NULL) & (theloc->pos != NULL)));
  if (axisn >= theloc->naxes)
    {
      logerr_err (LOGERR_FUNC,
                  "trying to get axis %d when there are only %d axes",
                  axisn, theloc->naxes);
      return 0;
    }
  else
    return theloc->pos[axisn];
}

/**
 * Initialize the data structures representing a stage (including the axes).
 * The new stage is defined in terms of a logical to physical axis mapping,
 * and we construct it by allocating not only the stage structure, but
 * the structures that represent the physical port and the individual
 * axes.  However we leave to @a openstage() the task of opening the
 * physical port and starting the threads responsible for controlling
 * each axis.
 * @param themap  the logical to physical axis mapping.
 * @return  the newly-allocated stage structure.
 * @see openstage()
 * @see disposestage()
 */
stagecontrolp_t
newstage (const axismappingp_t themap)
{
  stagecontrolp_t stage;
  devicestatep_t device;
  translationp_t axisvector;
  translationp_t theaxis;
  axismapentry_t *theentry;
  pthread_mutexattr_t *customattr;
  int ax;
  int err = 0;

  stage = NULL;
  if (themap == NULL)
    return stage;
  device = (devicestatep_t) malloc (sizeof (struct Devicestate));
  if (device == NULL)
    return stage;
  axisvector =
    (translationp_t) calloc (themap->naxes, sizeof (struct Translation));
  err = (axisvector == NULL);
  if (!err)
    {
      stage = (stagecontrolp_t) malloc (sizeof (struct Stagecontrol));
      err = (stage == NULL);
    }
  if (!err)
    {
      customattr = newcustommattr();
      err = (customattr == NULL);
    }  
  if (!err)
    {
      
      ax = themap->naxes;
      theaxis = axisvector;
      theentry = themap->mapping;
      while ((ax-- > 0) && !err)
        {
          theaxis->responsive = 0;
          theaxis->controller = stage;
          theaxis->step = (*(theentry++)) ? lostep : histep;
          theaxis->tattr = newdetachedtattr ();
          err = theaxis->tattr == NULL;
          theaxis++;
        }
      if (err)
        {
          disposemattr (customattr);
          while (++ax < themap->naxes)
            disposetattr ((--theaxis)->tattr);
        }
      else
        {
          stage->naxes = themap->naxes;
          stage->devstate = device;
          stage->axis = axisvector;
          stage->mattr = customattr;
          switch (stage->naxes)
            {
            case 1:
              stage->axis[0].label = zlabel;
              stage->axis[0].maxlash = 13;
              stage->axis[0].backlash = stage->axis[0].maxlash;
              stage->axis[0].delta = deltamag;
              stage->axis[0].mindelay = 16666667;
              stage->axis[0].maxdelay = 200000000;
              stage->axis[0].delay = stage->axis[0].maxdelay;
              stage->axis[0].accel =
                ((double) (stage->axis[0].maxdelay - stage->axis[0].mindelay)
                 / (double) (32 * deltamag));
              stage->axis[0].braking = 2 * stage->axis[0].accel;
              break;
            case 2:
              stage->axis[0].label = xlabel;
              stage->axis[0].maxlash = 10;
              stage->axis[0].backlash = stage->axis[0].maxlash;
              stage->axis[0].delta = deltamag;
              stage->axis[0].mindelay = 16666667;
              stage->axis[0].maxdelay = 200000000;
              stage->axis[0].delay = stage->axis[0].maxdelay;
              stage->axis[0].accel =
                ((double) (stage->axis[0].maxdelay - stage->axis[0].mindelay)
                 / (double) (32 * deltamag));
              stage->axis[0].braking = 2 * stage->axis[0].accel;
              stage->axis[1].label = ylabel;
              stage->axis[1].maxlash = 13;
              stage->axis[1].backlash = stage->axis[1].maxlash;
              stage->axis[1].delta = deltamag;
              stage->axis[1].mindelay = 16666667;
              stage->axis[1].maxdelay = 200000000;
              stage->axis[1].delay = stage->axis[1].maxdelay;
              stage->axis[1].accel =
                ((double) (stage->axis[1].maxdelay - stage->axis[1].mindelay)
                 / (double) (32 * deltamag));
              stage->axis[1].braking = 2 * stage->axis[1].accel;
            }
        }
    }
  if (err)
    {
      free (device);
      free (axisvector);
      free (stage);
      stage = NULL;
    }
  return stage;
}


/**
 * Destroy a stage data structure (without checking to see if it is in use).
 * @param stage  the stage to destroy.
 * @see newstage()
 * @see closestage()
 */
void
disposestage (stagecontrolp_t stage)
{
  translationp_t theaxis;
  int ax;

  if (stage != NULL)
    {
      free (stage->devstate);
      theaxis = stage->axis;
      ax = stage->naxes;
      disposemattr (stage->mattr);
      while (ax-- > 0)
        disposetattr ((theaxis++)->tattr);
      free (stage->axis);
      free (stage);
    }
}

/**
 * Open the parallel port for writing and initialize it.
 * @param thestate  the previously allocated device state structure.
 * @param path  the device port path.
 * @param flags  additional flags specifying attributes for the stage.
 * @return an error code on failure, zero on success.
 */
static int
openparportdevice (devicestatep_t thestate, const char *path, const int flags)
{
  int err = 0;

  thestate->portdata = offstate;
  thestate->pportfd = open (path, O_WRONLY);
  if (thestate->pportfd < 0)
    {
      logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
                "could not open the stage device");
      err = thestate->pportfd;
    }
  else           /* Open worked, so subsequent errors need a matching close. */
    {
      if ((err = ioctl (thestate->pportfd, PPEXCL)))
        logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
                  "no exclusive access to the stage");
      else if ((err = ioctl (thestate->pportfd, PPCLAIM)))
        logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
                  "failed to claim the stage");
      else if ((err = ioctl (thestate->pportfd, PPNEGOT, &ppmode)))
        logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
                  "negotiating transfer mode for the stage failed");
      else if ((err = ioctl (thestate->pportfd, PPWDATA,
                             &thestate->portdata)))
        logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
                  "first write failed (placing stage in the all-off state)");
      else if ((err = ioctl (thestate->pportfd, PPFCONTROL,
                            ((flags | STAGEPROP_HEAVY)
                             ? &heavy_activate : &activate))))
        logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
                  "could not set up stage control lines");
      if (err)   /* Catch all the errors that leave the device open. */
        err = close (thestate->pportfd);
    }
  return err;
}

/**
 * Connect to the physical port controlling the stage.
 * We need a stage data structure previously allocated by
 * @a newstage() and a path that defines the physical port controlling
 * the stage hardware (something like @c /dev/parportN).  We attempt
 * to open the port, then perform some initialization on it (e.g.,
 * setting up control registers and making sure it is in a quiescent
 * state), then start the detached threads controlling each axis.
 * @param stage  the previously allocated stage control structure.
 * @param path  the device port path.
 * @param flags  additional flags specifying attributes for the stage.
 * @return an error code on failure, zero on success.
 * @see newstage()
 * @see closestage()
 */
int
openstage (stagecontrolp_t stage, const char *path, const int flags)
{
  int err = 0;
  int ax;
  translationp_t theaxis;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((stage != NULL)
                  && (stage->devstate != NULL) && (stage->axis != NULL)));
  err = openparportdevice (stage->devstate, path, flags);
  if (err)
    return err;
  if ((err = pthread_mutex_init (&stage->modlock, stage->mattr)))
    logerr_err (LOGERR_FUNC,
                "could not initialize the mutex; error = %d",
                err);
  else if ((err = pthread_cond_init (&stage->offtrack, NULL)))
    {
      logerr_err (LOGERR_FUNC,
                  "could not initialize the \"offtrack\" condition variable; error = %d",
                  theaxis->label, err);
      pthread_mutex_destroy (&stage->modlock);
    }
  else if ((err = pthread_cond_init (&stage->ontrack, NULL)))
    {
      logerr_err (LOGERR_FUNC,
                  "could not initialize the \"ontrack\" condition variable; error = %d",
                  theaxis->label, err);
      pthread_cond_destroy (&stage->offtrack);
      pthread_mutex_destroy (&stage->modlock);
    }
  if (!err)
    {
      stage->ctrlstatus = STAGE_LIVE;
      ax = stage->naxes;
      theaxis = stage->axis;
      while (!err && ax--)
        err = startaxis (theaxis++);
      if (err)
        {
          stage->ctrlstatus = STAGE_DEAD;
          while (ax++ < stage->naxes)
            stopaxis (--theaxis);
          pthread_cond_destroy (&stage->ontrack);
          pthread_cond_destroy (&stage->offtrack);
          pthread_mutex_destroy (&stage->modlock);
        }          
    }
  return err;
}

/**
 * Halt control of the stage, without de-allocating all data structures.
 * This is the inverse of @a openstage(), returning to a state similar
 * to that when the stage data structures were allocated but unused.
 * We wait for stage activity to cease, cancel the controlling
 * threads for each axis, place the port in a quiescent state, then
 * close it.
 * @param stage  the open stage to be closed.
 * @return an error code on failure, zero on success.
 * @see disposestage()
 * @see openstage()
 */
int
closestage (stagecontrolp_t stage)
{
  int ax;
  translationp_t theaxis;
  devicestatep_t thestate;
  int err = 0;
  int closefail = 0;
  int unlockfail = 0;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((stage != NULL)
                  && (stage->devstate != NULL) && (stage->axis != NULL)));
  stage->ctrlstatus &= ~STAGE_LIVE;
  theaxis = stage->axis;
  ax = stage->naxes;
  while (!err && (ax-- > 0))
    err = stopaxis (theaxis++);
  if (err)
    {
      logerr_err (LOGERR_FUNC, "could not stop axis; error %d", err);
      return err;
    }
  err = pthread_mutex_lock (&stage->modlock);
  if (err)
    {
      logerr_err (LOGERR_FUNC,
                  "could not acquire the lock; error = %d",
                  err);
      return err;
    }
  thestate = stage->devstate;
  thestate->portdata = offstate;
  err = ioctl (thestate->pportfd, PPWDATA, &thestate->portdata);
  if (err)
    logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
	      "last write failed (placing stage in the all-off state)");
  closefail = close (thestate->pportfd);
  if (closefail)
    logerrno (LOGERR_WARNING, errno, LOGERR_FUNC,
	      "could not close the device");
  unlockfail = pthread_mutex_unlock (&stage->modlock);
  if (unlockfail)
    {
      logerr_err (LOGERR_FUNC,
                  "could not release the lock; error = %d",
                  unlockfail);
      return unlockfail;
    }
  if (!err && !closefail)
    {
      logerr_debug (LOGERR_FUNC,
                    "starting final cleanup.",
                    theaxis->label);
      if ((err = pthread_cond_destroy (&stage->offtrack)))
        logerr_debug (LOGERR_FUNC,
                    "could not destroy the \"offtrack\" condition variable; error = %d",
                    err);
      else if ((err = pthread_cond_destroy (&stage->ontrack)))
        logerr_debug (LOGERR_FUNC,
                    "could not destroy the \"ontrack\" condition variable; error = %d",
                    err);
      else if ((err = pthread_mutex_destroy (&stage->modlock)))
        logerr_debug (LOGERR_FUNC,
                    "could not destroy the mutex; error = %d",
                    err);
    }
  return (err) ? err : closefail;
}

/**
 * Issue a request that the stage move to a particular position.
 * @param stage  the stage to move.
 * @param motn  the requested position.
 * @return an error code on failure, zero on success.
 */
int
movestage (stagecontrolp_t stage, const stagelocationp_t motn)
{
  int err = 0;
  int ax;
  translationp_t theaxis;
  coord_t *requested;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((stage != NULL)
                  && (stage->devstate != NULL) && (stage->axis != NULL)
                  && (motn != NULL) && (motn->pos != NULL)));
  ax = stage->naxes;
  if (ax != motn->naxes)
    {
      logerr_err (LOGERR_FUNC,
                  "there are %d axes, but %d directions in the motion specification",
                  ax, motn->naxes);
      return -1;
    }
  theaxis = stage->axis;
  requested = motn->pos;
  err = pthread_mutex_lock (&stage->modlock);
  if (err)
    {
      logerr_err (LOGERR_FUNC,
                  "could not acquire the lock; error = %d",
                  err);
      return err;
    }
  else
    {
      while (!err && (ax-- > 0))
        {
          theaxis->goalpos = *requested;
          logerr_debug (LOGERR_FUNC,
                        "requested moving the axis labelled %s from %d to %d",
                        theaxis->label, theaxis->pos, theaxis->goalpos);
          theaxis++;
          requested++;
        }
    }
  err = pthread_mutex_unlock (&stage->modlock);
  if (err)
    logerr_err (LOGERR_FUNC,
                "could not release the lock; error = %d",
                err);
  else
    {
      err = pthread_cond_broadcast (&stage->offtrack);
      if (err)
        logerr_err (LOGERR_FUNC,
                    "something amiss with the \"offtrack\" condition variable; error = %d",
                    err);
    }
  return err;
}

/**
 * Copy the various axis positions into a coordinate structure.
 * If the stage is at rest, this finds its current position; however if the
 * axes are in motion the stage may have moved from the recorded
 * position before this returns.
 * @param stage  the stage whose position will be found.
 * @param posn   where to record the coordinates.
 * @return an error code on failure, zero on success.
 */
int
findstage (stagecontrolp_t stage, stagelocationp_t posn)
{
  int err = 0;
  int ax;
  translationp_t theaxis;
  coord_t *nextcoord;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((stage != NULL)
                  && (stage->devstate != NULL) && (stage->axis != NULL)
                  && (posn != NULL) && (posn->pos != NULL)));
  ax = stage->naxes;
  if (ax != posn->naxes)
    {
      logerr_err (LOGERR_FUNC,
                  "there are %d axes, but %d directions in the position specification",
                  ax, posn->naxes);
      return -1;
    }
  theaxis = stage->axis;
  nextcoord = posn->pos;
  err = pthread_mutex_lock (&stage->modlock);
  if (err)
    {
      logerr_err (LOGERR_FUNC,
                  "could not acquire the lock; error = %d",
                  err);
      return err;
    }
  else
    {
      while (!err && (ax-- > 0))
        {
          *nextcoord = theaxis->pos;
          logerr_debug (LOGERR_FUNC,
                        "the axis labelled %s is at %d",
                        theaxis->label, theaxis->pos);
          theaxis++;
          nextcoord++;
        }
    }
  err = pthread_mutex_unlock (&stage->modlock);
  if (err)
    logerr_err (LOGERR_FUNC,
                "could not release the lock; error = %d",
                err);
  return err;
}

static const wait_t poll_delay = 125000;

/**
 * Examine the stage to find if it is active.
 * We try to acquire the lock on the stage structure, and immediately
 * determine that the stage is busy if we can't acquire it, without
 * examining anything else.  As soon as we find an axis that's not
 * in its resting state we also determine the stage must be busy, but
 * test it again after waiting one polling interval (currently a hard-wired
 * constant).  However we release the lock before returning, so it's quite
 * possible that the stage is busy again by the time this returns, if
 * no other conditions are imposed.
 * @param stage  the stage to examine.
 * @return a non-zero code if the stage is active, zero if it is not.
 */
int
stageactive (stagecontrolp_t stage)
{
  int err = 0;
  int active = 0;
  int ax;
  translationp_t theaxis;
  struct timespec poll_wake;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((stage != NULL)
                  && (stage->devstate != NULL) && (stage->axis != NULL)));
  err = pthread_mutex_trylock (&stage->modlock);
  if (err)
    {
      active = (err == EBUSY);
      if (!active)
        logerr_err (LOGERR_FUNC,
                    "problem acquiring the lock; error = %d",
                    err);
      return active;
    }
  ax = stage->naxes;
  theaxis = stage->axis;
  while (!active && (ax > 0))
    {
      active = !(resting_condition (theaxis));
      theaxis++;
      ax--;
    }
  if (active)
    {
      err = nextwake (poll_delay, &poll_wake);
      if (!err)
        {
          err = pthread_cond_timedwait (&stage->ontrack,
                                        &stage->modlock, &poll_wake);
          if (err == ETIMEDOUT)
            err = 0;
          active = (err != 0);
          if (!err)
            {
              ax = stage->naxes;
              theaxis = stage->axis;
              while (!active && (ax > 0))
                {
                  active = !(resting_condition (theaxis));
                  theaxis++;
                  ax--;
                }
            }
        }
    }
  err = pthread_mutex_unlock (&stage->modlock);
  if (err)
    logerr_err (LOGERR_FUNC,
                "could not release the lock; error = %d",
                err);
  return active;
}
