/** @file trees_addremoverings.c
 * Glue code for using the C add and remove ring operations with Tcl/Tk.
 * There are two aspects to this.  We determine provisional changes
 * (provisionally inserted or deleted ring boundaries), and save the
 * revised measurements, but do not modify the main data structures
 * used by the Tcl parts of TREES immediately.  Subsequently (presumably
 * with some user interaction) we do commit the provisional changes
 * so that they become indistinguishable from the original contents
 * of the main TREES data structures.  Other parts of the program handle
 * the re-analysis (interpolating ring boundaries, computing revised
 * measurements, and so on) and storage of the provisional data:
 * see the chain_provisional.c and chain_provisional.h files for the
 * details.
 * 
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: trees_addremoverings.c,v 1.4 2004/11/12 22:50:00 mmunro Exp $
 */

/* This file is part of TREES.

   TREES is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   TREES is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with TREES; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include "logerr.h"
#include <sadie.h>
#include "chain.h"
#include "chain_provisional.h"

/** Points to a list of (non-contiguous) rings. */
typedef struct Ring_candidate *ring_candidatep_t;

/**
 * One ring in a list of rings (e.g., candidates for insertion or removal).
 */
struct Ring_candidate
{
  LIST_NODE *inner; /**< inner (earlier) boundary of the ring. */
  LIST_NODE *outer; /**< outer (later) boundary of the ring. */
  ring_candidatep_t others;  /**< the rest of the list of ring candidates. */
};

/** Bounding boxes are defined by the left, top, right, and bottom edges. */
static const int box_size = 4;

/**
 * Pull enough coordinates from a list of Tcl objects to define a bounding box.
 * @param interp  the Tcl interpreter within which the object list is defined.
 * @param objhandle  a handle pointing in to the object (coordinate) list.
 * @param thebox  points to the box to initialize with the new coordinates.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
objlist_to_bbox (Tcl_Interp * interp, Tcl_Obj ***objhandle,
                 BOUNDBOX_LIST thebox)
{
  int err = TCL_OK;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (interp && objhandle && *objhandle && thebox));
  err = Tcl_GetDoubleFromObj (interp, *((*objhandle)++), &(thebox->minx));
  if (err != TCL_OK)
    return TCL_ERROR;
  err = Tcl_GetDoubleFromObj (interp, *((*objhandle)++), &(thebox->miny));
  if (err != TCL_OK)
    return TCL_ERROR;
  err = Tcl_GetDoubleFromObj (interp, *((*objhandle)++), &(thebox->maxx));
  if (err != TCL_OK)
    return TCL_ERROR;
  err = Tcl_GetDoubleFromObj (interp, *((*objhandle)++), &(thebox->maxy));
  if (err != TCL_OK)
    return TCL_ERROR;
  return err;
}

/**
 * Check if a point falls within the bounding box enclosing two boundaries.
 * Given the coordinates of the point, and the bounding box enclosing the
 * individual boundaries, we consider the new bounding box enclosing both
 * boundaries.  If the point falls within this new box, we create a new
 * ring candidate structure recording the boundary pair that might
 * enclose the point (possibly for more rigorous point inclusion tests).
 * @param x  coordinate of the point to be checked. 
 * @param y  coordinate of the point to be checked.
 * @param inbox  points to the bounding box of the inner boundary.
 * @param outbox  points to the bounding box of the outer boundary.
 * @param in  points to the chain-coded representation of the inner boundary.
 * @param out  points to the chain-coded representation of the outer boundary.
 * @param headhandle  a handle to the head of the candidate ring list.
 */
static void
check_inside_box (double x, double y, BOUNDBOX_LIST inbox,
                  BOUNDBOX_LIST outbox, LIST_NODE * in, LIST_NODE * out,
                  ring_candidatep_t * headhandle)
{
  ring_candidatep_t newposs;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (inbox && outbox && in && out && headhandle));
  if (((fmin (inbox->minx, outbox->minx) <= x)
       && (x <= fmax (inbox->maxx, outbox->maxx)))
      && ((fmin (inbox->miny, outbox->miny) <= y)
          && (y <= fmax (inbox->maxy, outbox->maxy))))
    {
      newposs = malloc (sizeof (struct Ring_candidate));
      logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newposs != NULL));
      newposs->inner = in;
      newposs->outer = out;
      newposs->others = *headhandle;
      *headhandle = newposs;
    }
}

/**
 * Check if a point falls within the bounding box enclosing a boundary.
 * If the point falls within the bounding box, we create a new ring
 * candidate structure recording the boundary and its predecessor
 * (which may be NULL if there is no such other boundary).
 * @param x  coordinate of the point to be checked. 
 * @param y  coordinate of the point to be checked.
 * @param thebox  points to the bounding box of the boundary.
 * @param outbox  points to the bounding box of the outer boundary.
 * @param predecessor  points to the chain-coded representation of a boundary.
 * @param thebound  points to the boundary to compare with the point.
 * @param headhandle  a handle to the head of the candidate ring list.
 */
static void
match_box (double x, double y, BOUNDBOX_LIST thebox, LIST_NODE * predecessor,
           LIST_NODE * thebound, ring_candidatep_t * headhandle)
{
  ring_candidatep_t newposs;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (thebox && thebound && headhandle));
  if (((thebox->minx <= x) && (x <= thebox->maxx))
      && ((thebox->miny <= y) && (y <= thebox->maxy)))
    {
      newposs = malloc (sizeof (struct Ring_candidate));
      logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newposs != NULL));
      newposs->inner = predecessor;
      newposs->outer = thebound;
      newposs->others = *headhandle;
      *headhandle = newposs;
    }
}

/**
 * Expand bounding boxes by a fixed increment in all directions.
 * We will use this to help find points that are near to boundaries,
 * but not necessarily on them.
 * @param thebox  points to the head of a list of bounding boxes.
 * @param increment  the amount by which to expand the boxes in each direction.
 */
static inline void
expand_box (BOUNDBOX_LIST thebox, double increment)
{
  while (thebox != NULL)
    {
      thebox->minx -= increment;
      thebox->maxx += increment;
      thebox->miny -= increment;
      thebox->maxy += increment;
      thebox = thebox->next;
    }
}

/**
 * Free a dynamically allocated list of ring candidate structures.
 * We set the pointer to the old list head to NULL once it is invalid.
 * @param  a handle to the head of the list to be freed.
 */
static void
demolish_candidates (ring_candidatep_t * headhandle)
{
  ring_candidatep_t tail;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (headhandle != NULL));
  while (*headhandle != NULL)
    {
      tail = (*headhandle)->others;
      free (*headhandle);
      *headhandle = tail;
    }
}

/**
 * Search for where to create a new ring boundary.
 * Given the coordinates of a point and the list of chain-coded
 * boundaries, we consider each pair of boundaries (defining
 * successive rings).  If the point falls within the bounding box
 * enclosing a boundary pair, we save the boundaries on a list of
 * rings that might possibly enclose the point.  We then apply a more
 * rigorous test to discover if the point is enclosed by the shortlist
 * of candidate rings.  If the point falls outside all the the rings,
 * we interpolate a new boundary at the closest end of the list of
 * existing boundaries, but otherwise interpolate it between two
 * existing boundaries.
 * @param doc  defines the collection of data for this particular analysis.
 * @param provbnum  the identification number to give the new boundary.
 * @param image  points to the image mosaic.
 * @param mindex  the region-of-interest definition and other indices.
 * @param chainmap  points to the head of the list of chain-coded boundaries.
 * @param scale  the downsampling factor for low-resolution boundaries.
 * @param x  coordinate of the point defining the new boundary location.
 * @param y  coordinate of the point defining the new boundary location.
 * @param interp  the Tcl interpreter within which the box list is defined.
 * @param boxlist  the list of bounding boxes as coordiates in a Tcl array.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
search_ring (char * doc, int provbnum, IMAGE_BYTE * mimage,
             MOSAIC_INDEX * mindex, CHAIN * chainmap, double scale,
             double x, double y, Tcl_Interp *interp, Tcl_Obj * boxlist)
{
  BOUNDBOX_NODE p_box, q_box;
  BOUNDBOX_LIST inbox = NULL;
  BOUNDBOX_LIST outbox = NULL;
  BOUNDBOX_LIST temp = NULL;
  LIST_NODE *inner = NULL;
  LIST_NODE *outer = NULL;
  LIST_NODE *successor = NULL;
  LIST_NODE *initial = NULL;
  Tcl_Obj **edge = NULL;
  ring_candidatep_t possibles = NULL;
  ring_candidatep_t candidate = NULL;
  int found, i, n_edges;
  int st = 0;
  int err = TCL_OK;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (doc && mimage && mindex && chainmap && interp && boxlist));
  p_box.next = NULL;
  q_box.next = NULL;
  inbox = &p_box;
  outbox = &q_box;
  outer = chainmap->list;
  while (outer && (outer->ring_type == REMOVED))
    outer = outer->next;
  if (outer == NULL)
    return TCL_ERROR;
  initial = outer;
  err = Tcl_ListObjGetElements (interp, boxlist, &n_edges, &edge);
  if (err != TCL_OK)
    return TCL_ERROR;
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((n_edges != 0) && ((n_edges % box_size) == 0)));
  err = objlist_to_bbox (interp, &edge, outbox);
  i = n_edges - box_size;
  successor = outer->next;
  while (successor && (successor->ring_type == REMOVED))
    successor = successor->next;
  while ((err == TCL_OK) && (successor != NULL) && (i > 0))
    {
      inner = outer;
      outer = successor;
      do
        {
          successor = successor->next;
        }
      while (successor && (successor->ring_type == REMOVED));
      temp = inbox;
      inbox = outbox;
      outbox = temp;
      err = objlist_to_bbox (interp, &edge, outbox);
      i -= box_size;
      if (err == TCL_OK)
        check_inside_box (x, y, inbox, outbox, inner, outer, &possibles);
    }
  if (err == TCL_OK)
    {
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                     ((successor == NULL) && (i <= 0)));
      candidate = possibles;
      found = 0;
      while ((candidate != NULL) && !found)
        {
          found = provisional_check (x, y, candidate->inner, candidate->outer);
          if (!found)
            candidate = candidate->others;
        }
      if (found)
        st = provisional_insert (doc, provbnum, mimage, mindex, scale, x, y,
                                 candidate->inner, candidate->outer);
      else
        st = provisional_extend (doc, provbnum, mimage, mindex, scale, x, y,
                                 initial, outer);
      if (st != 0)
        err = TCL_ERROR;
    }
  demolish_candidates (&possibles);
  return err;
}

/** How close a point has to be to a box for it to count as being near it. */
const static int match_tolerance = 6;

/**
 * Search for where to remove an existing ring boundary.
 * Given the coordinates of a point and the list of chain-coded
 * boundaries, if the point falls within the expanded bounding box
 * enclosing a boundary, we save the boundaries on a list of
 * boundaries that might possibly be near the point.  We then apply a
 * more rigorous test to discover the closest match within the list of
 * of candidate boundaries.  If the point falls outside all the the
 * rings, we remove the boundary at the closest end of the list of
 * existing boundaries, but otherwise the closest boundary within the
 * rings.
 * @param doc  defines the collection of data for this particular analysis.
 * @param provbnum  the identification number to give the new boundary.
 * @param image  points to the image mosaic.
 * @param mindex  the region-of-interest definition and other indices.
 * @param chainmap  points to the head of the list of chain-coded boundaries.
 * @param smallchainmap  points to the low-resolution chain-coded boundaries.
 * @param x  coordinate of the point defining the new boundary location.
 * @param y  coordinate of the point defining the new boundary location.
 * @param interp  the Tcl interpreter within which the box list is defined.
 * @param boxlist  the list of bounding boxes as coordiates in a Tcl array.
 * @found  points to an int (really boolean) set true if we find a boundary.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
search_boundary (char * doc, int provbnum, IMAGE_BYTE * mimage,
                 MOSAIC_INDEX * mindex, CHAIN * chainmap, CHAIN *smallchainmap,
                 double x, double y, Tcl_Interp *interp, Tcl_Obj * boxlist,
                 int * found)
{
  BOUNDBOX_NODE bbox;
  LIST_NODE *predecessor = NULL;
  LIST_NODE *chainbound = NULL;
  LIST_NODE *successor = NULL;
  LIST_NODE *small_deletion = NULL;
  LIST_NODE *small_predecessor = NULL; /* unused */
  ring_candidatep_t possibles = NULL;
  ring_candidatep_t matching = NULL;
  ring_candidatep_t closest = NULL;
  Tcl_Obj **edge = NULL;
  double d, d_min;
  int i, n_edges, tmpindex;
  int st = 0;
  int err = TCL_OK;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (doc && mimage && mindex && chainmap && interp && boxlist
                  && found));
  bbox.next = NULL;
  *found = 0;
  chainbound = chainmap->list;
  while (chainbound && (chainbound->ring_type == REMOVED))
    chainbound = chainbound->next;
  if (chainbound == NULL)
    return TCL_ERROR;
  err = Tcl_ListObjGetElements (interp, boxlist, &n_edges, &edge);
  if (err != TCL_OK)
    return TCL_ERROR;
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((n_edges != 0) && ((n_edges % box_size) == 0)));
  i = n_edges;
  while ((err == TCL_OK) && (chainbound != NULL) && (i > 0))
    {
      err = objlist_to_bbox (interp, &edge, &bbox);
      if (err == TCL_OK)
        {
          i -= box_size;
          expand_box (&bbox, match_tolerance);
          match_box (x, y, &bbox, predecessor, chainbound, &possibles);
          predecessor = chainbound;
          do
            {
              chainbound = chainbound->next;
            }
          while (chainbound && (chainbound->ring_type == REMOVED));
        }
    }
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((err != TCL_OK) || ((chainbound == NULL) && (i <= 0))));
  closest = possibles;
  if ((closest != NULL) && (err == TCL_OK))
    {
      if (closest->others != NULL)
        {
          d_min = provisional_point_distance (x, y, closest->outer);
          matching = closest->others;
          while (matching != NULL)
            {
              d = provisional_point_distance (x, y, matching->outer);
              if (d < d_min)
                {
                  d_min = d;
                  closest = matching;
                }
              matching = matching->others;
            }
        }
      successor = closest->outer->next;
      while (successor && successor->ring_type == REMOVED)
        successor = successor->next;
      search_chainlist_by_id (closest->outer->ring_id, smallchainmap->list,
                              &small_predecessor, &small_deletion, &tmpindex);
      st = provisional_delete (doc, provbnum, mimage, mindex,  closest->inner,
                               closest->outer, successor, small_deletion);
      *found = (st == 0);
      err = (*found) ? TCL_OK : TCL_ERROR;
    }
  demolish_candidates (&possibles);
  return err;
}

/**
 * Add a new ring boundary between two existing chain-coded boundaries.
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, document identifies the grouping
 * of saved settings for a particular analysis, provbnum will identify
 * the new boundary and the two sets of measurements associated with
 * it, mosaic is the string encoding the storage location of the image
 * mosaic, index identifies the data structure holding the
 * region-of-interest definition and other indices into this mosaic,
 * chainmap identifies the lists of full-resolution chain-coded
 * boundaries, scale is the downsampling factor used when creating
 * low-resolution boundaries, x_coord and y_coord are the coordinates
 * within the image mosaic near where we will construct the
 * interpolated boundary, and bbox_name is the name of the Tcl list of
 * bounding box coordinates.  The changes get saved into an entry in
 * the hash table and not immediately incorporated into changes to the
 * chain map and measurements.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_add_provisionalCmd (ClientData clientData, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
  char *doc = NULL;
  char * hex_chain = NULL;
  char * hex_image = NULL;
  char * hex_index = NULL;
  CHAIN * chainmap = NULL;
  IMAGE_BYTE * mosaic_image = NULL;
  MOSAIC_INDEX * mosaic_index = NULL;
  Tcl_Obj * boxlist = NULL;
  int len, provbnum;
  size_t raw_chain, raw_image, raw_index;
  double scale, x, y;
  int could_add = 0;
  int err = TCL_OK;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 10)
    {
      Tcl_WrongNumArgs (interp, 1, objv, 
                        "document provbnum mosaic index chainmap scale x_coord y_coord bbox_listname");
      return TCL_ERROR;
    }
  doc = Tcl_GetStringFromObj (objv[1], &len);
  if (len == 0)
    return TCL_ERROR;
  err = Tcl_GetIntFromObj (interp, objv[2], &provbnum);
  if (err == TCL_ERROR)
    return err;
  hex_image = Tcl_GetStringFromObj (objv[3], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_image, "%x", &raw_image);
  mosaic_image = (IMAGE_BYTE *) raw_image;
  hex_index = Tcl_GetStringFromObj (objv[4], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_index, "%x", &raw_index);
  mosaic_index = (MOSAIC_INDEX *) raw_index;
  hex_chain = Tcl_GetStringFromObj (objv[5], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_chain, "%x", &raw_chain);
  chainmap = (CHAIN *) raw_chain;
  err = Tcl_GetDoubleFromObj (interp, objv[6], &scale);
  if (err == TCL_ERROR)
    return err;
  err = Tcl_GetDoubleFromObj (interp, objv[7], &x);
  if (err == TCL_ERROR)
    return err;
  err = Tcl_GetDoubleFromObj (interp, objv[8], &y);
  if (err == TCL_ERROR)
    return err;
  boxlist = Tcl_ObjGetVar2 (interp, objv[9], NULL, 0);
  if (Tcl_IsShared (boxlist))
    boxlist = Tcl_DuplicateObj (boxlist);
  could_add = provisional_point_in_roi (mosaic_index, x, y);
  if (could_add)
    err = search_ring (doc, provbnum, mosaic_image, mosaic_index, chainmap,
                       scale, x, y, interp, boxlist);
  if (err == TCL_OK)
    Tcl_SetObjResult (interp, Tcl_NewBooleanObj (could_add));
  return err;
}

/**
 * Remove an existing ring boundary (re-computing the affected measurements).
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, document identifies the grouping
 * of saved settings for a particular analysis, provbnum will identify
 * the modified set of measurements associated with the removed
 * boundary, mosaic is the string encoding the storage location of the
 * image mosaic, index identifies the data structure holding the
 * region-of-interest definition and other indices into this mosaic,
 * chainmap identifies the lists of full-resolution chain-coded
 * boundaries, smallchainmap identifies the lists of low-resolution
 * chain-coded boundaries, x_coord and y_coord are the coordinates
 * within the image mosaic near the existing boundary, and bbox_name
 * is the name of the Tcl list of bounding box coordinates.  The
 * changes get saved into an entry in the hash table and not
 * immediately incorporated into changes to the chain map and
 * measurements.  We set the Tcl result object to a boolean flag showing
 * whether the provisional boundary deletion was successful.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_remove_provisionalCmd (ClientData clientData, Tcl_Interp *interp,
                             int objc, Tcl_Obj * const objv[])
{
  char *doc = NULL;
  char * hex_chain = NULL;
  char * hex_smallchain = NULL;
  char * hex_image = NULL;
  char * hex_index = NULL;
  CHAIN * chainmap = NULL;
  CHAIN * smallchainmap = NULL;
  IMAGE_BYTE * mosaic_image = NULL;
  MOSAIC_INDEX * mosaic_index = NULL;
  Tcl_Obj * boxlist = NULL;
  int len, provbnum;
  size_t raw_chain, raw_smallchain, raw_image, raw_index;
  double x, y;
  int could_remove = 0;
  int err = TCL_OK;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 10)
    {
      Tcl_WrongNumArgs (interp, 1, objv, 
                        "document provbnum mosaic index chainmap smallchainmap x_coord y_coord bbox_listname");
      return TCL_ERROR;
    }
  doc = Tcl_GetStringFromObj (objv[1], &len);
  if (len == 0)
    return TCL_ERROR;
  err = Tcl_GetIntFromObj (interp, objv[2], &provbnum);
  if (err == TCL_ERROR)
    return err;
  hex_image = Tcl_GetStringFromObj (objv[3], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_image, "%x", &raw_image);
  mosaic_image = (IMAGE_BYTE *) raw_image;
  hex_index = Tcl_GetStringFromObj (objv[4], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_index, "%x", &raw_index);
  mosaic_index = (MOSAIC_INDEX *) raw_index;
  hex_chain = Tcl_GetStringFromObj (objv[5], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_chain, "%x", &raw_chain);
  chainmap = (CHAIN *) raw_chain;
  hex_smallchain = Tcl_GetStringFromObj (objv[6], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_smallchain, "%x", &raw_smallchain);
  smallchainmap = (CHAIN *) raw_smallchain;
  err = Tcl_GetDoubleFromObj (interp, objv[7], &x);
  if (err == TCL_ERROR)
    return err;
  err = Tcl_GetDoubleFromObj (interp, objv[8], &y);
  if (err == TCL_ERROR)
    return err;
  boxlist = Tcl_ObjGetVar2 (interp, objv[9], NULL, 0);
  if (Tcl_IsShared (boxlist))
    boxlist = Tcl_DuplicateObj (boxlist);
  could_remove = provisional_point_in_roi (mosaic_index, x, y);
  if (could_remove)
    err = search_boundary (doc, provbnum, mosaic_image, mosaic_index, chainmap,
                           smallchainmap, x, y, interp, boxlist, &could_remove);
  if (err == TCL_OK)
    Tcl_SetObjResult (interp, Tcl_NewBooleanObj (could_remove));
  return err;
}

/** The number of Tcl objects required to encode the drawing command. */
static const int draw_command_len = 13;

/**
 * Draw a chain-coded boundary using Tcl/Tk commands.
 * We generate a command template, giving the Tk commands needed to
 * draw a line segment in terms of Tcl objects, then traverse the
 * chain, coded boundary, generating the corrdinates of each
 * successive point, substituting them into the template, and
 * executing the resulting command.
 * @param interp  the Tcl interpreter within which to define the command.
 * @param canvas  the Tk canvas on which to draw the boundary.
 * @param x_0  offset from the boundary origin to the canvas origin.
 * @param y_0  offset from the boundary origin to the canvas origin.
 * @param headnode  points to the chain-coded boundary.
 * @param btag  the identifying tag associated with each line.
 * @param color  the line color.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
draw_provisional_chain (Tcl_Interp * interp, Tcl_Obj *canvas, int x_0, int y_0,
                        LIST_NODE * headnode, char * btag, Tcl_Obj * color)
{
  int x_prev, y_prev, x, y, n, m;
  CHAIN_NODE *cnode;
  Tcl_Obj * tkcmd[draw_command_len];
  Tcl_Obj * tagarr[2];
  int err = TCL_OK;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (interp && canvas && btag && color));
  if (headnode == NULL)
    return TCL_ERROR;
  x = headnode->startk - x_0;
  y = headnode->startj - y_0;
  tkcmd[0] = canvas;
  tkcmd[1] = Tcl_NewStringObj ("create", -1);
  tkcmd[2] = Tcl_NewStringObj ("line", -1);
  tkcmd[3] = Tcl_NewDoubleObj (x);
  tkcmd[4] = Tcl_NewDoubleObj (y);
  tkcmd[5] = Tcl_NewDoubleObj (x);
  tkcmd[6] = Tcl_NewDoubleObj (y);
  tkcmd[7] = Tcl_NewStringObj ("-width", -1);
  tkcmd[8] = Tcl_NewDoubleObj (1);
  tkcmd[9] = Tcl_NewStringObj ("-fill", -1);
  tkcmd[10] = color;
  tkcmd[11] = Tcl_NewStringObj ("-tag", -1);
  tagarr[0] = Tcl_NewStringObj (btag, -1);
  tagarr[1] = Tcl_NewStringObj ("normal", -1);
  tkcmd[12] = Tcl_NewListObj (2, tagarr);
  for (n = 0; n < draw_command_len; n++)
    Tcl_IncrRefCount (tkcmd[n]);
  cnode = headnode->chain;
  while (cnode && (err == TCL_OK))
    {
      x_prev = x;
      y_prev = y;
      CHAIN_COMPUTE_COORDINATES (x_prev, y_prev, cnode->code, &x, &y);
      cnode = cnode->next;
      Tcl_DecrRefCount (tkcmd[3]);
      tkcmd[3] = tkcmd[5];
      Tcl_DecrRefCount (tkcmd[4]);
      tkcmd[4] = tkcmd[6];
      tkcmd[5] = Tcl_NewDoubleObj (x);
      Tcl_IncrRefCount (tkcmd[5]);
      tkcmd[6] = Tcl_NewDoubleObj (y);
      Tcl_IncrRefCount (tkcmd[6]);
      err = Tcl_EvalObjv (interp, draw_command_len, tkcmd, 0);
    }
  for (m = 0; m < draw_command_len; m++)
    Tcl_DecrRefCount(tkcmd[m]);
  return err;
}

/**
 * Draw a full-resolution provisional ring boundary.
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, canvas specifies the Tk canvas
 * within which the drawing takes place (such as the zoom window),
 * x_origin and y_origin define the offsets between the origin used in
 * the image mosaic and the canvas origin, doc identifies the grouping
 * of saved settings for a particular analysis, provbnum is the number
 * identifying the provisional boundary, and color is the color for
 * the lines making up the boundary.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_display_provisionalCmd (ClientData clientData, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  int len = 0;
  int err = 0;
  int provbnum, x_0, y_0;
  LIST_NODE *headnode;
  char *doc = NULL;
  char btag[SLEN];
  char pname[SLEN];
  
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 7)
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "canvas x_origin y_origin doc provbnum color");
      return TCL_ERROR;
    }
  err = Tcl_GetIntFromObj (interp, objv[2], &x_0);
  if (err == TCL_ERROR)
    return err;
  err = Tcl_GetIntFromObj (interp, objv[3], &y_0);
  if (err == TCL_ERROR)
    return err;
  doc = Tcl_GetStringFromObj (objv[4], &len);
  if (len == 0)
    return TCL_ERROR;
  err = Tcl_GetIntFromObj (interp, objv[5], &provbnum);
  if (err == TCL_ERROR)
    return err;
  snprintf (pname, SLEN - 1, "Provisional-%d", provbnum);
  headnode = get_provisional_full_chain (doc, pname);
  if (headnode == NULL)
    return TCL_ERROR;
  else
    {
      snprintf (btag, SLEN - 1, "Boundary-%d", headnode->ring_id);
      return draw_provisional_chain (interp, objv[1], x_0, y_0,
                                     headnode, btag, objv[6]);
    }
}

/**
 * Draw a low-resolution provisional ring boundary.
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, canvas specifies the Tk canvas
 * within which the drawing takes place (such as the browse window),
 * doc identifies the grouping of saved settings for a particular
 * analysis, provbnum is the number identifying the provisional
 * boundary, and color is the color for the lines making up the
 * boundary.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_display_small_provisionalCmd (ClientData clientData, Tcl_Interp *interp,
                                    int objc, Tcl_Obj * const objv[])
{
  int len = 0;
  int err = 0;
  int provbnum;
  LIST_NODE *headnode;
  char *doc = NULL;
  char btag[SLEN];
  char pname[SLEN];
  
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 5)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "canvas doc provbnum color");
      return TCL_ERROR;
    }
  doc = Tcl_GetStringFromObj (objv[2], &len);
  if (len == 0)
    return TCL_ERROR;
  err = Tcl_GetIntFromObj (interp, objv[3], &provbnum);
  if (err == TCL_ERROR)
    return err;
  snprintf (pname, SLEN - 1, "Provisional-%d", provbnum);
  headnode = get_provisional_small_chain (doc, pname);
  if (headnode == NULL)
    return TCL_ERROR;
  else
    {
      snprintf (btag, SLEN - 1, "Boundary-%d", headnode->ring_id);
      return draw_provisional_chain (interp, objv[1], 0, 0,
                                     headnode, btag, objv[4]);
    }
}

/** The number of Tcl objects required to encode the deletion command. */
static const int delete_command_len = 3;

/**
 * Delete a provisional ring boundary from the Tk canvas.
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, canvas specifies the Tk canvas
 * within which the drawing takes place, doc identifies the grouping
 * of saved settings for a particular analysis, and provbnum is the
 * number identifying the provisional boundary.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_undisplay_provisionalCmd (ClientData clientData, Tcl_Interp *interp,
                                    int objc, Tcl_Obj * const objv[])
{
  int len = 0;
  int provbnum, n, m;
  LIST_NODE *headnode;
  char *doc = NULL;
  char btag[SLEN];
  char pname[SLEN];
  Tcl_Obj * tkcmd[delete_command_len];
  int err = TCL_OK;
  
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 4)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "canvas doc provbnum");
      return TCL_ERROR;
    }
  doc = Tcl_GetStringFromObj (objv[2], &len);
  if (len == 0)
    return TCL_ERROR;
  err = Tcl_GetIntFromObj (interp, objv[3], &provbnum);
  if (err != TCL_OK)
    return TCL_ERROR;
  snprintf (pname, SLEN - 1, "Provisional-%d", provbnum);
  headnode = get_provisional_full_chain (doc, pname);
  if (headnode == NULL)
    return TCL_ERROR;
  else
    {
      snprintf (btag, SLEN - 1, "Boundary-%d", headnode->ring_id);
      tkcmd[0] = objv[1];
      tkcmd[1] = Tcl_NewStringObj ("delete", -1);
      tkcmd[2] = Tcl_NewStringObj (btag, -1);
      for (n = 0; n < delete_command_len; n++)
        Tcl_IncrRefCount (tkcmd[n]);
      err = Tcl_EvalObjv (interp, delete_command_len, tkcmd, 0);
      for (m = 0; m < delete_command_len; m++)
        Tcl_DecrRefCount(tkcmd[m]);
    }
  return err;
}

/** The number of Tcl objects required to encode the coloring command. */
static const int color_command_len = 5;

/**
 * Color a previously drawn boundary using a Tcl/Tk command.
 * We build the Tcl/Tk command in terms of Tcl objects, one of which is the
 * tag which will select all the line segments making up the drawn boundary.
 * @param interp  the Tcl interpreter within which to define the command.
 * @param canvas  the Tk canvas on which to draw the boundary.
 * @param headnode  points to the chain-coded boundary.
 * @param btag  the identifying tag associated with each line.
 * @param color  the line color.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
color_provisional_chain (Tcl_Interp * interp, Tcl_Obj *canvas,
                          LIST_NODE * headnode, char * btag, Tcl_Obj * color)
{
  int m, n;
  Tcl_Obj * tkcmd[color_command_len];
  int err = TCL_OK;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (interp && canvas && btag && color));
  if (headnode == NULL)
    return TCL_ERROR;
  tkcmd[0] = canvas;
  tkcmd[1] = Tcl_NewStringObj ("itemconfigure", -1);
  tkcmd[2] = Tcl_NewStringObj (btag, -1);
  tkcmd[3] = Tcl_NewStringObj ("-fill", -1);
  tkcmd[4] = color;
  for (n = 0; n < color_command_len; n++)
    Tcl_IncrRefCount (tkcmd[n]);
  err = Tcl_EvalObjv (interp, color_command_len, tkcmd, 0);
  for (m = 0; m < color_command_len; m++)
    Tcl_DecrRefCount(tkcmd[m]);
  return err;
}

/**
 * Change the color of a boundary that is already on the Tk canvas.
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, canvas specifies the Tk canvas
 * within which the drawing takes place, doc identifies the grouping
 * of saved settings for a particular analysis, and provbnum is the
 * number identifying the provisional boundary, and color is the new
 * color to use.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_color_provisionalCmd (ClientData clientData, Tcl_Interp *interp,
                            int objc, Tcl_Obj * const objv[])
{
  int len = 0;
  int err = 0;
  int provbnum;
  LIST_NODE *headnode;
  char *doc = NULL;
  char btag[SLEN];
  char pname[SLEN];
  
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 5)
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "canvas doc provbnum color");
      return TCL_ERROR;
    }
  doc = Tcl_GetStringFromObj (objv[2], &len);
  if (len == 0)
    return TCL_ERROR;
  err = Tcl_GetIntFromObj (interp, objv[3], &provbnum);
  if (err != TCL_OK)
    return TCL_ERROR;
  snprintf (pname, SLEN - 1, "Provisional-%d", provbnum);
  headnode = get_provisional_full_chain (doc, pname);
  if (headnode == NULL)
    return TCL_ERROR;
  else
    {
      snprintf (btag, SLEN - 1, "Boundary-%d", headnode->ring_id);
      return color_provisional_chain (interp, objv[1], headnode, btag,
                                       objv[4]);
    }
}

/**
 * Commit a provisionally added ring boundary to the main chain map structures.
 * @param chainmap  points to the head of the full-resolution boundary lists.
 * @param smallchainmap  points to the low-resolution boundary list head.
 * @parma insert_id  id of the insertion point, invalid if prepending.
 * @parma prov_full  points to the full-resolution provisional boundary.
 * @parma prov_small  points to the low-resolution provisional boundary.
 * @return  the index position of the insertion point in the data lists.
 */
static int
insert_chain_boundaries (CHAIN * chainmap, CHAIN * smallchainmap,
                         int insert_id,
                         LIST_NODE * prov_full, LIST_NODE * prov_small)
{
  int first, tmpindex;
  LIST_NODE *predecessor_full = NULL; /* unused */
  LIST_NODE *predecessor_small = NULL; /* unused */
  LIST_NODE *insert_full = NULL;
  LIST_NODE *insert_small = NULL;
  LIST_NODE *new_full = NULL;
  LIST_NODE *new_small = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (chainmap && smallchainmap && prov_full && prov_small
                  && (prov_full->next == NULL) && (prov_small->next == NULL)));
  tmpindex = first = -1;
  copy_chainlist_head (prov_full, &new_full);
  copy_chainlist_head (prov_small, &new_small);
  if (insert_id == INVALID_ID)
    {
      new_full->next = chainmap->list;
      chainmap->list = new_full;
      new_small->next = smallchainmap->list;
      smallchainmap->list = new_small;
    }
  else
    {
      search_chainlist_by_id (insert_id, chainmap->list,
                              &predecessor_full, &insert_full, &tmpindex);
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                     (insert_full && (insert_full->ring_id == insert_id)));
      search_chainlist_by_id (insert_id, smallchainmap->list,
                              &predecessor_small, &insert_small, &first);
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                     (insert_small && (insert_small->ring_id == insert_id)));
      new_full->next = insert_full->next;
      insert_full->next = new_full;
      new_small->next = insert_small->next;
      insert_small->next = new_small;
    }
  return first;
}

/**
 * Commit a provisional ring boundary removal to the chain map structures.
 * @param chainmap  points to the head of the full-resolution boundary lists.
 * @param smallchainmap  points to the low-resolution boundary list head.
 * @parma remove_id  the ring_id of the item to remove (which must be valid).
 * @return  the index position of the removed item in the data lists.
 */
static int
remove_chain_boundaries (CHAIN * chainmap, CHAIN * smallchainmap,
                         int remove_id)
{
  int first, tmpindex;
  LIST_NODE *predecessor_full = NULL;
  LIST_NODE *predecessor_small = NULL;
  LIST_NODE *remove_full = NULL;
  LIST_NODE *remove_small = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (chainmap && smallchainmap && (remove_id != INVALID_ID)));
  search_chainlist_by_id (remove_id, chainmap->list,
                          &predecessor_full, &remove_full, &tmpindex);
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (remove_full && (remove_full->ring_id == remove_id)));
  search_chainlist_by_id (remove_id, smallchainmap->list,
                          &predecessor_small, &remove_small, &first);
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (remove_small && (remove_small->ring_id == remove_id)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (first >= 0));
  first--;
  if (predecessor_full == NULL)
    chainmap->list = remove_full->next;
  else
    predecessor_full->next = remove_full->next;
  remove_full->next = NULL;
  demolish_chainlist (&remove_full);  
  if (predecessor_small == NULL)
    smallchainmap->list = remove_small->next;
  else
    predecessor_small->next = remove_small->next;
  remove_small->next = NULL;
  demolish_chainlist (&remove_small);
  return first;
}

/** An accessor function turns various saved data into Tcl object arrays. */
typedef int (*data_accessor_t) (char * doc, char * pname, int * n,
                                Tcl_Obj * objarr[]);

#define TEMP_OBJ_ARRAY_LEN 4

/**
 * Convert previously saved numeric data into Tcl objects.
 * @param val  points to the data array to convert.
 * @param n  the number of values to convert.
 * @param objarr  the array of Tcl object pointers to receive the results.
 */
static void
double_to_objarr (double * val, int n, Tcl_Obj * * objarr)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((objarr != NULL) && (val != NULL)));
  while (n-- > 0)
    {
      *objarr = Tcl_NewDoubleObj (*val++);
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*objarr != NULL));
      objarr++;
    }
}

/**
 * Convert previously determined brightness profiles into Tcl objects.
 * We generate a new Tcl object array for each profile's data, then make
 * Tcl list objects from these to place in the results.
 * @param val  points to the data array to convert.
 * @param n  the number of profiles to convert.
 * @param objarr  the array of Tcl list object pointers to receive the results.
 */
static void
profile_to_objarr (PROFILE_LIST * prof, int n, Tcl_Obj * * objarr)
{
  Tcl_Obj **bright_obj = NULL;
  Tcl_Obj **bright_arr = NULL;
  BRIGHTNESS *bright_val = NULL; 
  int i;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((objarr != NULL) && (prof != NULL) && (*prof != NULL)));
  while (n-- > 0)
    {
      i = (*prof)->len;
      bright_arr = (i > 0) ? calloc ((*prof)->len, sizeof(Tcl_Obj *)) : NULL;
      bright_obj = bright_arr;
      bright_val = (*prof)->val;
      while (i-- > 0)
        *bright_obj++ = Tcl_NewDoubleObj ((double) *bright_val++);
      *objarr = Tcl_NewListObj ((*prof)->len, bright_arr);
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, *objarr != NULL);
      free (bright_arr);
      objarr++;
      prof++;
    }
}

/**
 * Insert previously determined measurements into various Tcl data lists.
 * This manipulates most of the measurement lists in the main Tcl part
 * of TREES, although they do not contain the same types of data.  We
 * use a family of data accessor functions to convert the disparate
 * saved data resulting from boundary insertions into Tcl objects, and
 * can then use the same code to insert these into the various
 * different lists.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl measurement list into which to insert.
 * @param first  the index position of the insertion point in the Tcl list.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the particular data to convert to Tcl objects.
 * @param fetch_data  the function that retrieves and converts the saved data.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
insert_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
              char * doc, char * pname, data_accessor_t fetch_data)
{
  int len, n;
  Tcl_Obj * objarr[TEMP_OBJ_ARRAY_LEN];
  Tcl_Obj * thelist = NULL;
  Tcl_Obj * newval = NULL;
  int err = TCL_OK;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname && fetch_data));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, 0);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  err = (*fetch_data) (doc, pname, &n, objarr);
  if (err == TCL_OK)
    {
      if (first < 0)
        {
          logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 1));
          err = Tcl_ListObjReplace (interp, thelist, 0, 0, n, objarr);
        }
      else
        {
          err = Tcl_ListObjLength (interp, thelist, &len);
          if (err == TCL_OK)
            {
              if (first >= len)
                {
                  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 1));
                  err = Tcl_ListObjReplace (interp, thelist,
                                            len, 0, n, objarr);
                }
              else
                {
                  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 2));
                  err = Tcl_ListObjReplace (interp, thelist,
                                            first, 1, n, objarr);
                }
            }
        }
    }
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, 0);
  return err;
}

/**
 * Modify the main Tcl data lists to reflect a boundary deletion.
 * This modifies most of the measurement lists in the main Tcl part of
 * TREES, using the same technique as the insert_list function.  We
 * use a family of data accessor functions to convert the disparate
 * saved data resulting from boundary deletions into Tcl objects, and
 * can then use the same code to modify the various different Tcl data
 * lists to reflect these.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl measurement list to modify.
 * @param first  the index position of the removal point in the Tcl list.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the particular data to convert to Tcl objects.
 * @param fetch_data  the function that retrieves and converts the saved data.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
remove_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
             int need_update, char * doc, char * pname,
             data_accessor_t fetch_data)
{
  int len, n;
  Tcl_Obj * objarr[TEMP_OBJ_ARRAY_LEN];
  Tcl_Obj * thelist = NULL;
  Tcl_Obj * newval = NULL;
  int err = TCL_OK;
  
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname && fetch_data));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, 0);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  if (first < 0)
    err = Tcl_ListObjReplace (interp, thelist, 0, 1, 0, NULL);
  else
    {
      err = Tcl_ListObjLength (interp, thelist, &len);
      if (err == TCL_OK)
        {
          if (first >= (len - 1))
            err = Tcl_ListObjReplace (interp, thelist, (len - 1), 1, 0, NULL);
          else
            {
              err = Tcl_ListObjReplace (interp, thelist,
                                        (first + 1), 1, 0, NULL);
              if (need_update && (err == TCL_OK))
                {
                  err = (*fetch_data) (doc, pname, &n, objarr);
                  if (err == TCL_OK)
                    {
                      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 1));
                      err = Tcl_ListObjReplace (interp, thelist,
                                                first, 1, 1, objarr);
                    }
                }
            }
        }
    }
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, 0);
  return err;
}

/**
 * Insert the bounding box for a new boundary into the relevant Tcl data list.
 * We use the same approach as the generic data insertion function,
 * insert_list, but customize it for the peculiarities of the bounding
 * box data: these come in groupings of four values, and unlike ring
 * widths or other measurements, the bounding boxes associated with
 * existing ring boundaries do not change when we insert a new
 * boundary.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl bounding box list into which to insert.
 * @param first  the index position of the insertion point in the Tcl list.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the bounding box to convert to Tcl objects.
 * @param fetch_data  the function that retrieves and converts the saved data.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
insert_box_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
                 char * doc, char * pname, data_accessor_t fetch_data)
{
  int len, n;
  Tcl_Obj * objarr[TEMP_OBJ_ARRAY_LEN];
  Tcl_Obj * thelist = NULL;
  Tcl_Obj * newval = NULL;
  int err = TCL_OK;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname && fetch_data));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, TCL_GLOBAL_ONLY);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  err = (*fetch_data) (doc, pname, &n, objarr);
  if (err == TCL_OK)
    {
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == box_size));
      first *= box_size;
      if (first < 0)
        err = Tcl_ListObjReplace (interp, thelist, 0, 0, n, objarr);
      else
        {
          first += box_size;
          err = Tcl_ListObjLength (interp, thelist, &len);
          if (err == TCL_OK)
            {
              err = (first >= len)
                ? Tcl_ListObjReplace (interp, thelist, len, 0, n, objarr)
                : Tcl_ListObjReplace (interp, thelist, first, 0, n, objarr);
            }
        }
    }
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, TCL_GLOBAL_ONLY);
  return err;
}

/**
 * Remove the bounding box for a deleted boundary from the relevant Tcl list.
 * We use the same approach as the generic data removal function,
 * remove_list, but customize it for the peculiarities of the bounding
 * box data: these come in groupings of four values, and unlike ring
 * widths or other measurements, the bounding boxes associated with
 * existing ring boundaries do not change when we remove a boundary.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl bounding box list from which to remove.
 * @param first  the index position of the removal point in the Tcl list.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the bounding box to convert to Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
remove_box_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
                 char * doc, char * pname)
{
  Tcl_Obj * thelist = NULL;
  Tcl_Obj * newval = NULL;
  int len;
  int err = TCL_OK;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, TCL_GLOBAL_ONLY);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  first *= box_size;
  if (first < 0)
    err = Tcl_ListObjReplace (interp, thelist, 0, box_size, 0, NULL);
  else
    {
      first += box_size;
      err = Tcl_ListObjLength (interp, thelist, &len);
      if (err == TCL_OK)
        {
          err = (first >= (len - box_size))
            ? Tcl_ListObjReplace (interp, thelist,
                                  (len - box_size), box_size, 0, NULL)
            : Tcl_ListObjReplace (interp, thelist,
                                  first, box_size, 0, NULL);
        }
    }
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, TCL_GLOBAL_ONLY);
  return TCL_OK;
}

/**
 * A data accessor function: retrieve and convert total ring widths.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the total ring widths to convert to Tcl objects.
 * @param n  will point to the number of widths retrieved.
 * @param objarr  the array of Tcl object pointers to receive the results.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
fetch_widths (char * doc, char * pname, int * n, Tcl_Obj * objarr[])
{
  double *widtharr = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && n && objarr));
  *n = find_provisional_widths (doc, pname, &widtharr);
  if ((*n < 1) || (*n > 2))
    {
      *n = 0;
      return TCL_ERROR;
    }
  double_to_objarr (widtharr, *n, objarr);
  return TCL_OK;
}

/**
 * A data accessor function: retrieve and convert brightness profiles.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the profiles to convert to Tcl objects.
 * @param n  will point to the number of profiles retrieved.
 * @param objarr  the array of Tcl object pointers to receive the results.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
fetch_profiles (char * doc, char * pname, int * n, Tcl_Obj * objarr[])
{
  PROFILE_LIST *profarr = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && n && objarr));
  *n = find_provisional_profiles (doc, pname, &profarr);
  if ((*n < 1) || (*n > 2))
    {
      *n = 0;
      return TCL_ERROR;
    }
  profile_to_objarr (profarr, *n, objarr);
  return TCL_OK;
}

/**
 * A data accessor function: retrieve and convert differenced profiles.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the profiles to convert to Tcl objects.
 * @param n  will point to the number of profiles retrieved.
 * @param objarr  the array of Tcl object pointers to receive the results.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
fetch_diff_profiles (char * doc, char * pname, int * n, Tcl_Obj * objarr[])
{
  PROFILE_LIST *profarr = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && n && objarr));
  *n = find_provisional_diff_profiles (doc, pname, &profarr);
  if ((*n < 1) || (*n > 2))
    {
      *n = 0;
      return TCL_ERROR;
    }
  profile_to_objarr (profarr, *n, objarr);
  return TCL_OK;
}

/**
 * A data accessor function: retrieve and convert earlywood widths.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the earlywood widths to convert to Tcl objects.
 * @param n  will point to the number of widths retrieved.
 * @param objarr  the array of Tcl object pointers to receive the results.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
fetch_earlywidths (char * doc, char * pname, int * n, Tcl_Obj * objarr[])
{
  double *widtharr = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && n && objarr));
  *n = find_provisional_earlywidths (doc, pname, &widtharr);
  if ((*n < 1) || (*n > 2))
    {
      *n = 0;
      return TCL_ERROR;
    }
  double_to_objarr (widtharr, *n, objarr);
  return TCL_OK;
}

/**
 * A data accessor function: retrieve and convert latewood widths.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the latewood widths to convert to Tcl objects.
 * @param n  will point to the number of widths retrieved.
 * @param objarr  the array of Tcl object pointers to receive the results.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
fetch_latewidths (char * doc, char * pname, int * n, Tcl_Obj * objarr[])
{
  double *widtharr = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && n && objarr));
  *n = find_provisional_latewidths (doc, pname, &widtharr);
  if ((*n < 1) || (*n > 2))
    {
      *n = 0;
      return TCL_ERROR;
    }
  double_to_objarr (widtharr, *n, objarr);
  return TCL_OK;
}

/**
 * A data accessor function: retrieve and convert bounding box coordinates.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the bounding box to convert to Tcl objects.
 * @param n  will point to the number of coordinates retrieved.
 * @param objarr  the array of Tcl object pointers to receive the results.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
fetch_boundbox  (char * doc, char * pname, int * n, Tcl_Obj * objarr[])
{
  BOUNDBOX_LIST bound_box = NULL;
  double boxval[4];
  int m;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (doc && pname && n && objarr
                                            && (box_size == 4)));
  m = find_provisional_boundbox (doc, pname, &bound_box);
  if (m != 1)
    {
      *n = 0;
      return TCL_ERROR;
    }
  boxval[0] = bound_box->minx;
  boxval[1] = bound_box->miny;
  boxval[2] = bound_box->maxx;
  boxval[3] = bound_box->maxy;
  double_to_objarr (boxval, box_size, objarr);
  *n = box_size;
  return TCL_OK;
}

/**
 * Modify the master list in the Tcl part of TREES after a boundary insertion.
 * The master list of ring boundary information in the main part of
 * the TREES Tcl code differs from the measurement lists in using
 * subscripted arrays values to specify various different kinds of
 * data.  It's best to think of it as a list of boundary names,
 * defining a linear sequence of boundary positions, loosely
 * associated with the global arrays that define the additional
 * properties for each name.  An update operation on an existing
 * boundary will not need to touch the list, unless its position
 * changes, but will need to change the array values holding the
 * properties.  Insertion will of course modify both list and arrays,
 * Inserting a boundary will cause one insertion in the list, which
 * might also be accompanied by an update of an existing boundary.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl master list into which to insert.
 * @param first  the index position of the insertion point in the Tcl list.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the total ring width data modified by an insertion.
 * @param btag  the name for the new boundary.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
insert_master_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
                    char * doc, char * pname, char * btag)
{
  int len, n;
  Tcl_Obj *thelist = NULL;
  Tcl_Obj *width_comp = NULL;
  Tcl_Obj *type_comp = NULL;
  Tcl_Obj *type_list = NULL;
  Tcl_Obj *normal = NULL;
  Tcl_Obj *bobj = NULL;
  Tcl_Obj *newval = NULL;
  Tcl_Obj *widtharr[TEMP_OBJ_ARRAY_LEN];
  Tcl_Obj **oldlist = NULL;
  int err = TCL_OK;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname && btag));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, 0);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  width_comp = Tcl_NewStringObj ("width", -1);
  Tcl_IncrRefCount (width_comp);
  type_comp = Tcl_NewStringObj ("typelist", -1);
  Tcl_IncrRefCount (type_comp);
  normal = Tcl_NewStringObj ("normal", -1);
  type_list = Tcl_NewListObj (1, &normal);
  Tcl_IncrRefCount (type_list);
  bobj = Tcl_NewStringObj (btag, -1);
  Tcl_IncrRefCount (bobj);
  err = fetch_widths (doc, pname, &n, widtharr);
  newval = Tcl_ObjSetVar2 (interp, bobj, type_comp, type_list,
                           TCL_GLOBAL_ONLY);
  if (err == TCL_OK)
    {
      if (first < 0)
        {
          logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 1));
          newval = Tcl_ObjSetVar2 (interp, bobj, width_comp,
                                   widtharr[0], TCL_GLOBAL_ONLY);
          err = Tcl_ListObjReplace (interp, thelist, 0, 0, 1, &bobj);
        }
      else
        {
          err = Tcl_ListObjGetElements (interp, thelist, &len, &oldlist);
          if (err == TCL_OK)
            {
              if (first >= (len - 1))
                {
                  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 1));
                  newval = Tcl_ObjSetVar2 (interp, oldlist[len-1], width_comp,
                                           widtharr[0], TCL_GLOBAL_ONLY);
                  newval = Tcl_ObjSetVar2 (interp, oldlist[len-1], type_comp,
                                           type_list, TCL_GLOBAL_ONLY);
                  err = Tcl_ListObjReplace (interp, thelist, len, 0, 1, &bobj);
                }
              else
                {
                  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 2));
                  newval = Tcl_ObjSetVar2 (interp, oldlist[first], width_comp,
                                           widtharr[0], TCL_GLOBAL_ONLY);
                  newval = Tcl_ObjSetVar2 (interp, bobj, width_comp,
                                           widtharr[1], TCL_GLOBAL_ONLY);
                  newval = Tcl_ObjSetVar2 (interp, oldlist[first], type_comp,
                                           type_list, TCL_GLOBAL_ONLY);
                  err = Tcl_ListObjReplace (interp, thelist,
                                            (first + 1), 0, 1, &bobj);
                }
            }
        }
    }
  Tcl_DecrRefCount (bobj);
  Tcl_DecrRefCount (type_list);
  Tcl_DecrRefCount (type_comp);
  Tcl_DecrRefCount (width_comp);
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, 0);
  return err;
}

/**
 * Modify the master list in the Tcl part of TREES after a boundary removal.
 * We follow the same approach as the insert_master_list function, but for
 * boundary removal.  We will always remove an entry from the list
 * of boundary names, and delete its associated properties from the other
 * global array values, but may also update the properties of an existing
 * boundary (its predecessor in the list) without changing the list of names.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl master list into which to insert.
 * @param first  the index position of the removal point in the Tcl list.
 * @param need_update  true if the remaining data need updated after a removal.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the total ring width data modified by a removal.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
remove_master_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
                    int need_update, char * doc, char * pname)
{
  int len, n;
  Tcl_Obj *thelist = NULL;
  Tcl_Obj *width_comp = NULL;
  Tcl_Obj *type_comp = NULL;
  Tcl_Obj *type_list = NULL;
  Tcl_Obj *normal = NULL;
  Tcl_Obj *newval = NULL;
  Tcl_Obj *widtharr[TEMP_OBJ_ARRAY_LEN];
  Tcl_Obj **oldlist = NULL;
  int err = TCL_OK;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, 0);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  if (first < 0)
    err = Tcl_ListObjReplace (interp, thelist, 0, 1, 0, NULL);
  else
    {
      err = Tcl_ListObjGetElements (interp, thelist, &len, &oldlist);
      if (err == TCL_OK)
        {
          if (first >= (len - 1))
            err = Tcl_ListObjReplace (interp, thelist, (len - 1), 1, 0, NULL);
          else
            {
              err = Tcl_ListObjReplace (interp, thelist,
                                        (first + 1), 1, 0, NULL);
              if (need_update && (err == TCL_OK))

                {
                  width_comp = Tcl_NewStringObj ("width", -1);
                  Tcl_IncrRefCount (width_comp);
                  type_comp = Tcl_NewStringObj ("typelist", -1);
                  Tcl_IncrRefCount (type_comp);
                  normal = Tcl_NewStringObj ("normal", -1);
                  type_list = Tcl_NewListObj (1, &normal);
                  Tcl_IncrRefCount (type_list);
                  err = fetch_widths (doc, pname, &n, widtharr);
                  if (err == TCL_OK)
                    {
                      newval = Tcl_ObjSetVar2 (interp, oldlist[first], width_comp,
                                               widtharr[0], TCL_GLOBAL_ONLY);
                      newval = Tcl_ObjSetVar2 (interp, oldlist[first], type_comp,
                                               type_list, TCL_GLOBAL_ONLY);
                    }
                  Tcl_DecrRefCount (type_list);
                  Tcl_DecrRefCount (type_comp);
                  Tcl_DecrRefCount (width_comp);
                }
            }
        }
    }
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, 0);
  return err;
}

/**
 * Insert the analyst note for a new ring boundary into the relevant Tcl list.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl analyst note list into which to insert.
 * @param first  the index position of the insertion point in the Tcl list.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the total ring width data modified by a removal.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
insert_note_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
                  char * doc, char * pname)
{
  int len, n;
  Tcl_Obj *thelist = NULL;
  Tcl_Obj *empty_note = NULL;
  Tcl_Obj *newval = NULL;
  Tcl_Obj *widtharr[TEMP_OBJ_ARRAY_LEN];
  int err = TCL_OK;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, 0);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  empty_note = Tcl_NewStringObj ("", -1);
  Tcl_IncrRefCount (empty_note);
  err = fetch_widths (doc, pname, &n, widtharr); /* Values not used. */
  if (err == TCL_OK)
    {
      if (first <= 0)
        {
          logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 1));
          err = Tcl_ListObjReplace (interp, thelist, 0, 0, 1, &empty_note);
        }
      else
        {
          err = Tcl_ListObjLength (interp, thelist, &len);
          if (err == TCL_OK)
            {
              if (first >= len)
                {
                  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 1));
                  err = Tcl_ListObjReplace (interp, thelist, len, 0, 1, &empty_note);
                }
              else
                {
                  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (n == 2));
                  err = Tcl_ListObjReplace (interp, thelist, first, 0, 1, &empty_note);
                }
            }
        }
    }
  Tcl_DecrRefCount (empty_note);
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, 0);
  return err;
}

/**
 * Remove the analyst note for a deleted ring boundary from the Tcl list.
 * @param interp  the Tcl interpreter running the main part of TREES.
 * @param listname  the name of the Tcl analyst note list from which to remove.
 * @param first  the index position of the removal point in the Tcl list.
 * @param doc  identifies the whole collection of previously saved data.
 * @param pname  identifies the total ring width data modified by a removal.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
static int
remove_note_list (Tcl_Interp * interp, Tcl_Obj * listname, int first,
                  char * doc, char * pname)
{
  Tcl_Obj *thelist = NULL;
  Tcl_Obj *newval = NULL;
  int len;
  int err = TCL_OK;

  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (interp && listname && doc && pname));
  thelist = Tcl_ObjGetVar2 (interp, listname, NULL, 0);
  if (Tcl_IsShared (thelist))
    thelist = Tcl_DuplicateObj (thelist);
  if (first < 0)
    err = Tcl_ListObjReplace (interp, thelist, 0, 1, 0, NULL);
  else
    {
      err = Tcl_ListObjLength (interp, thelist, &len);
      if (err == TCL_OK)
        {
          err = (first >= (len - 1))
            ? Tcl_ListObjReplace (interp, thelist, (len - 1), 1, 0, NULL)
            : Tcl_ListObjReplace (interp, thelist, (first + 1), 1, 0, NULL);
        }
    }
  if (err == TCL_OK)
    newval = Tcl_ObjSetVar2 (interp, listname, NULL, thelist, 0);
  return err;
}

/**
 * Commit a provisional boundary addition to the main Tcl structures of TREES.
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, document identifies the grouping
 * of saved settings for a particular analysis, provbnum identifies
 * the provisional boundary and the sets of measurements associated
 * with it, chain identifies the lists of full-resolution chain-coded
 * boundaries, smallchain identifies the lists of low-resolution
 * chain-coded boundaries, widthlist names the list of total ring
 * widths, profilelist names the list of brightness profile data,
 * diff_profilelist names the list of first-differenced brightness
 * profiles, earlywidthlist names the list of earlywood widths,
 * latewidthlist names the list of latewood widths, boundboxlist names
 * the list of bounding box coordinates, masterlist names the list of
 * ring boundaries (in a form suitable for use in the skeleton plot
 * display with their associated data), and notes is the list of short
 * textual notes by the tree-ring analyst.  We return the (integer)
 * index position of the insertion point within the Tcl list in the
 * Tcl result object.  The provisional data must already exist, stored
 * in the location identified by document and provbnum.  When we
 * commit them, we modify all the primary data structures used by the
 * Tcl parts of TREES to store measurements and housekeeping
 * information, and the next save to a data file should make the
 * changes permanent.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_commit_provisionalCmd (ClientData clientData, Tcl_Interp *interp,
                             int objc, Tcl_Obj * const objv[])
{
  int len = 0;
  int provbnum, first;
  char btag[SLEN];
  char pname[SLEN];
  char *doc = NULL;
  char *hex_chain = NULL;
  char *hex_smallchain = NULL;
  CHAIN *chainmap = NULL;
  CHAIN *smallchainmap = NULL;
  LIST_NODE *prov_full = NULL;
  LIST_NODE *prov_small = NULL;
  size_t raw_chain, raw_smallchain;
  
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 13)
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "doc provbnum chain smallchain widthlist profilelist diff_profilelist earlywidthlist latewidthlist boundboxlist masterlist notes");
      return TCL_ERROR;
    }
  doc = Tcl_GetStringFromObj (objv[1], &len);
  if (len == 0)
    return TCL_ERROR;
  if (Tcl_GetIntFromObj (interp, objv[2], &provbnum) != TCL_OK)
    return TCL_ERROR;
  snprintf (pname, SLEN - 1, "Provisional-%d", provbnum);
  snprintf (btag, SLEN - 1, "Boundary-%d", provbnum);
  hex_chain = Tcl_GetStringFromObj (objv[3], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_chain, "%x", &raw_chain);
  chainmap = (CHAIN *) raw_chain;
  hex_smallchain = Tcl_GetStringFromObj (objv[4], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_smallchain, "%x", &raw_smallchain);
  smallchainmap = (CHAIN *) raw_smallchain;
  prov_full = get_provisional_full_chain (doc, pname);
  if (prov_full == NULL)
    return TCL_ERROR;
  prov_small = get_provisional_small_chain (doc, pname);
  if (prov_small == NULL)
    return TCL_ERROR;
  first = insert_chain_boundaries (chainmap, smallchainmap,
                                   get_provisional_insert_point (doc, pname),
                                   prov_full, prov_small);
  if (insert_list (interp, objv[5], first, doc, pname,
                   fetch_widths) != TCL_OK)
    return TCL_ERROR;
  if (insert_list (interp, objv[6], first, doc, pname,
                   fetch_profiles) != TCL_OK)
    return TCL_ERROR;
  if (insert_list (interp, objv[7], first, doc, pname,
                   fetch_diff_profiles) != TCL_OK)
    return TCL_ERROR;
  if (insert_list (interp, objv[8], first, doc, pname,
                   fetch_earlywidths) != TCL_OK)
    return TCL_ERROR;
  if (insert_list (interp, objv[9], first, doc, pname,
                   fetch_latewidths) != TCL_OK)
    return TCL_ERROR;
  if (insert_box_list (interp, objv[10], first, doc, pname,
                       fetch_boundbox) != TCL_OK)
    return TCL_ERROR;
  if (insert_master_list (interp, objv[11], first, doc, pname,
                          btag) != TCL_OK)
    return TCL_ERROR;
  if (insert_note_list (interp, objv[12], first, doc, pname) != TCL_OK)
    return TCL_ERROR;
  Tcl_SetIntObj (Tcl_GetObjResult (interp), first);
  return TCL_OK;
}

/**
 * Commit a provisional boundary removal to the main Tcl structures of TREES.
 * Follow the standard Tcl_ObjCmdProc prototype for C glue code in
 * Tcl.  On the Tcl side of things, document identifies the grouping
 * of saved settings for a particular analysis, provbnum identifies
 * the removed boundary and the sets of measurements associated with
 * it, chain identifies the lists of full-resolution chain-coded
 * boundaries, smallchain identifies the lists of low-resolution
 * chain-coded boundaries, widthlist names the list of total ring
 * widths, profilelist names the list of brightness profile data,
 * diff_profilelist names the list of first-differenced brightness
 * profiles, earlywidthlist names the list of earlywood widths,
 * latewidthlist names the list of latewood widths, boundboxlist names
 * the list of bounding box coordinates, masterlist names the list of
 * ring boundaries, and notes is the list of short textual notes by
 * the tree-ring analyst.  We return the (integer) index position of
 * the removal point within the Tcl list in the Tcl result object.
 * The provisional data must already exist, stored in the location
 * identified by document and provbnum.  When we commit them, we
 * modify all the primary data structures used by the Tcl parts of
 * TREES to store measurements and housekeeping information, and the
 * next save to a data file should make the changes permanent.
 * @param clientData  currently unused (but required for a Tcl_ObjCmdProc).
 * @param interp  the Tcl interpreter within which to define the command.
 * @param objc  the number of Tcl objects passed to the command.
 * @param objv  the array of Tcl objects.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_commit_removalCmd (ClientData clientData, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  int len = 0;
  int need_update = 0;
  int provbnum, first;
  char btag[SLEN];
  char pname[SLEN];
  char *doc = NULL;
  char *hex_chain = NULL;
  char *hex_smallchain = NULL;
  CHAIN *chainmap = NULL;
  CHAIN *smallchainmap = NULL;
  LIST_NODE *prov_full = NULL;
  size_t raw_chain, raw_smallchain;
  
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (interp && objv));
  if (objc != 13)
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "doc provbnum chain smallchain widthlist profilelist diff_profilelist earlywidthlist latewidthlist boundboxlist masterlist notes");
      return TCL_ERROR;
    }
  doc = Tcl_GetStringFromObj (objv[1], &len);
  if (len == 0)
    return TCL_ERROR;
  if (Tcl_GetIntFromObj (interp, objv[2], &provbnum) != TCL_OK)
    return TCL_ERROR;
  snprintf (pname, SLEN - 1, "Provisional-%d", provbnum);
  snprintf (btag, SLEN - 1, "Boundary-%d", provbnum);
  hex_chain = Tcl_GetStringFromObj (objv[3], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_chain, "%x", &raw_chain);
  chainmap = (CHAIN *) raw_chain;
  hex_smallchain = Tcl_GetStringFromObj (objv[4], &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_smallchain, "%x", &raw_smallchain);
  smallchainmap = (CHAIN *) raw_smallchain;
  prov_full = get_provisional_full_chain (doc, pname);
  if (prov_full == NULL)
    return TCL_ERROR;
  first = remove_chain_boundaries (chainmap, smallchainmap,
                                   prov_full->ring_id);
  need_update = have_provisional_changes (doc, pname);
  if (remove_list (interp, objv[5], first, need_update, doc, pname,
                   fetch_widths) != TCL_OK)
    return TCL_ERROR;
  if (remove_list (interp, objv[6], first, need_update, doc, pname,
                   fetch_profiles) != TCL_OK)
    return TCL_ERROR;
  if (remove_list (interp, objv[7], first, need_update, doc, pname,
                   fetch_diff_profiles) != TCL_OK)
    return TCL_ERROR;
  if (remove_list (interp, objv[8], first, need_update, doc, pname,
                   fetch_earlywidths) != TCL_OK)
    return TCL_ERROR;
  if (remove_list (interp, objv[9], first, need_update, doc, pname,
                   fetch_latewidths) != TCL_OK)
    return TCL_ERROR;
  if (remove_box_list (interp, objv[10], first, doc, pname) != TCL_OK)
    return TCL_ERROR;
  if (remove_master_list (interp, objv[11], first, need_update, doc, pname)
      != TCL_OK)
    return TCL_ERROR;
  if (remove_note_list (interp, objv[12], first, doc, pname) != TCL_OK)
    return TCL_ERROR;
  Tcl_SetIntObj (Tcl_GetObjResult (interp), first);
  return TCL_OK;
}

/**
 * Create the new Tcl commands associated with adding and removing boundaries.
 * @param interp  the Tcl interpreter within which to define the commands.
 * @return  the TCL_OK code on success, the TCL_ERROR code on failure.
 */
int
TREES_addremoverings_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "TREES_add_provisional",
                        TREES_add_provisionalCmd,
                        (ClientData) NULL, NULL);
  Tcl_CreateObjCommand (interp, "TREES_remove_provisional",
                        TREES_remove_provisionalCmd,
                        (ClientData) NULL, NULL);
  Tcl_CreateObjCommand (interp, "TREES_display_provisional",
                        TREES_display_provisionalCmd,
                        (ClientData) NULL, NULL);
  Tcl_CreateObjCommand (interp, "TREES_display_small_provisional",
                        TREES_display_small_provisionalCmd,
                        (ClientData) NULL, NULL);
  Tcl_CreateObjCommand (interp, "TREES_undisplay_provisional",
                        TREES_undisplay_provisionalCmd,
                        (ClientData) NULL, NULL);
  Tcl_CreateObjCommand (interp, "TREES_color_provisional",
                        TREES_color_provisionalCmd,
                        (ClientData) NULL, NULL);
  Tcl_CreateObjCommand (interp, "TREES_commit_provisional",
                        TREES_commit_provisionalCmd,
                        (ClientData) NULL, NULL);
  Tcl_CreateObjCommand (interp, "TREES_commit_removal",
                        TREES_commit_removalCmd,
                        (ClientData) NULL, NULL);
  return TCL_OK;
}
