/*
##########################################
# Sadie_Proto.c -
#   Set of routines for linking SADIE prototype routines to tcl/tk.
#
# RCS: $Id: Sadie_Proto.c,v 1.4 2004/09/07 23:35:08 mmunro Exp $ 
# 
# Digital Image Analysis Lab
# Dept. of Electrical and Computer Engineering
# University of Arizona
##########################################
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "proto.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/* RCS Indentification information */
static char rcsid[] = "$Id: Sadie_Proto.c,v 1.4 2004/09/07 23:35:08 mmunro Exp $";


/*-------------------------------------------------------------*/
/* Global Sadie variables that must be set from within Tcl/Tk  */
/*-------------------------------------------------------------*/
extern 	short   nlev;
extern	short   csize;
extern	double  weight;
extern	double  *count;
extern	PIXEL   gain;
extern	PIXEL   bias;
extern	PIXEL   gmin;
extern	PIXEL   gmax;
extern	PIXEL   thresh;
extern	PIXEL   gbrk[2][4];
extern	PIXEL   *table;


/*
 * Compute the delay between t1 and t2 in milliseconds
 */
long delay(struct timeval t1, struct timeval t2)
{
        long d;
        
        d = (t2.tv_sec - t1.tv_sec) * 1000;
        d += ((t2.tv_usec - t1.tv_usec + 500) / 1000);
        return(d);
}

 
/* -------------------------------------------------------------------------- */


/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1999 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This procedure provides an interface to the SADIE function               */
/*   FINDBESTFOCUS from Tcl/Tk.  It expects a tcl global array                */
/*   with these indices to exist:                                             */
/*      array(inimg,addr[1-nimg]) --  SADIE image address                     */
/*      array(nimg)               --  int                                   */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int Sadie_Proto_FindBestFocusCmd(ClientData client_data, Tcl_Interp* interp, int argc, char *argv[])
{  
    Tcl_Obj* tclobj=NULL;
    Tcl_Obj* tclarrayname = NULL;
    Tcl_Obj* tclindexname = NULL;
    char msg[SLEN];
    char* array=NULL;
    char* tempstr=NULL;
    int strlen;
    int inimgaddr;
    IMAGE **inimg=NULL;
    int i, nimg, best;

    if( argc != 2 ) {
	Tcl_AppendResult(interp,"wrong # args: should be \"",
	argv[0], " arrayname\"", (char *) NULL);
	return TCL_ERROR;
    }
    array=argv[1];
		

    /* Read the input integer array(nimg) */
    tclarrayname = Tcl_NewStringObj(array,-1);
    tclindexname = Tcl_NewStringObj("nimg",-1);
    if (tclobj = Tcl_ObjGetVar2(interp,  tclarrayname , tclindexname , TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)) {
	if (Tcl_GetIntFromObj(interp,tclobj,&nimg) == TCL_ERROR) return TCL_ERROR;
    } else {
	return TCL_ERROR;
    }
    Tcl_DecrRefCount(tclarrayname);
    Tcl_DecrRefCount(tclindexname);


    inimg = (IMAGE **) malloc(nimg * sizeof(IMAGE *));
    
    /* Read input image array(inimg,addr[1-nimg]) */
    for (i=0; i<nimg; i++) {
	tclarrayname = Tcl_NewStringObj(array,-1);
	sprintf(msg,"inimg,addr%d",i+1);
	tclindexname = Tcl_NewStringObj(msg,-1);
	if (tclobj = Tcl_ObjGetVar2(interp,  tclarrayname , tclindexname , TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)) {
	    tempstr = Tcl_GetStringFromObj(tclobj,&strlen);
	    if (strlen <= 0) return TCL_ERROR;
	    sscanf(tempstr,"%x",&inimgaddr);
	    inimg[i] = (IMAGE *) inimgaddr;
	} else {
	    return TCL_ERROR;
	}
	Tcl_DecrRefCount(tclarrayname);
	Tcl_DecrRefCount(tclindexname);
    }


    FINDBESTFOCUS(inimg,nimg,&best);
    printf("best focus = %d\n",best);

    sprintf(msg, "%d", best);
    Tcl_SetResult(interp, msg, TCL_VOLATILE);
    
    if (inimg) free(inimg);

    return TCL_OK;
}  




/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1999 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This procedure provides an interface to the SADIE function               */
/*   FIND_REGISTER from Tcl/Tk.  It expects a tcl global array                */
/*   with these indices to exist:                                             */
/*      array(inimg1,addr)        --  SADIE image address                     */
/*      array(inimg2,addr)        --  SADIE image address                     */
/*      array(est_x_offset)       --  int                                     */
/*      array(est_y_offset)       --  int                                     */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int Sadie_Proto_FindRegisterCmd(ClientData client_data, Tcl_Interp* interp, int argc, char *argv[])
{
    Tcl_Obj* tclobj=NULL;
    Tcl_Obj* tclarrayname = NULL;
    Tcl_Obj* tclindexname = NULL;
    char msg[SLEN];
    char* array=NULL;
    char* tempstr=NULL;
    int strlen;
    int inimg1addr, inimg2addr;
    IMAGE *inimg1=NULL, *inimg2=NULL;
    int est_x_offset, est_y_offset;
    int x_offset, y_offset;
    PIXEL bias_adj, gain_adj;

    if( argc != 2 ) {
	Tcl_AppendResult(interp,"wrong # args: should be \"",
	argv[0], " arrayname\"", (char *) NULL);
	return TCL_ERROR;
    }
    array=argv[1];
            

    /* Read input image array(inimg1,addr) */
    tclarrayname = Tcl_NewStringObj(array,-1);
    tclindexname = Tcl_NewStringObj("inimg1,addr",-1);
    if (tclobj = Tcl_ObjGetVar2(interp,  tclarrayname , tclindexname , TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)) {
	tempstr = Tcl_GetStringFromObj(tclobj,&strlen);
	if (strlen <= 0) return TCL_ERROR;
	sscanf(tempstr,"%x",&inimg1addr);
	inimg1 = (IMAGE *) inimg1addr;
    } else {
	return TCL_ERROR;
    }
    Tcl_DecrRefCount(tclarrayname);
    Tcl_DecrRefCount(tclindexname);
		

    /* Read input image array(inimg2,addr) */
    tclarrayname = Tcl_NewStringObj(array,-1);
    tclindexname = Tcl_NewStringObj("inimg2,addr",-1);
    if (tclobj = Tcl_ObjGetVar2(interp,  tclarrayname , tclindexname , TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)) {
	tempstr = Tcl_GetStringFromObj(tclobj,&strlen);
	if (strlen <= 0) return TCL_ERROR;
	sscanf(tempstr,"%x",&inimg2addr);
	inimg2 = (IMAGE *) inimg2addr;
    } else {
	return TCL_ERROR;
    }
    Tcl_DecrRefCount(tclarrayname);
    Tcl_DecrRefCount(tclindexname);
		

    /* Read the input integer array(est_x_offset) */
    tclarrayname = Tcl_NewStringObj(array,-1);
    tclindexname = Tcl_NewStringObj("est_x_offset",-1);
    if (tclobj = Tcl_ObjGetVar2(interp,  tclarrayname , tclindexname , TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)) {
	if (Tcl_GetIntFromObj(interp,tclobj,&est_x_offset) == TCL_ERROR) return TCL_ERROR;
    } else {
	return TCL_ERROR;
    }
    Tcl_DecrRefCount(tclarrayname);
    Tcl_DecrRefCount(tclindexname);


    /* Read the input integer array(est_y_offset) */
    tclarrayname = Tcl_NewStringObj(array,-1);
    tclindexname = Tcl_NewStringObj("est_y_offset",-1);
    if (tclobj = Tcl_ObjGetVar2(interp,  tclarrayname , tclindexname , TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)) {
	if (Tcl_GetIntFromObj(interp,tclobj,&est_y_offset) == TCL_ERROR) return TCL_ERROR;
    } else {
	return TCL_ERROR;
    }
    Tcl_DecrRefCount(tclarrayname);
    Tcl_DecrRefCount(tclindexname);
		

    FIND_REGISTER(inimg1, inimg2, est_x_offset, est_y_offset, &x_offset, &y_offset, &bias_adj, &gain_adj);

    sprintf(msg, "%d %d %f %f", x_offset, y_offset, bias_adj, gain_adj);
    Tcl_SetResult(interp, msg, TCL_VOLATILE);
	    
    return TCL_OK;
}



/*************************************************************************
 *
 * This procedure initializes the the Sadie_Proto procedures when
 * this package is loaded.
 *
 *************************************************************************/
int Sadie_Proto_Init(Tcl_Interp *interp) {

  Tcl_CreateCommand(interp, "Sadie_Proto_FindRegister", Sadie_Proto_FindRegisterCmd, 
        (ClientData) NULL, NULL);
  Tcl_CreateCommand(interp, "Sadie_Proto_FindBestFocus", Sadie_Proto_FindBestFocusCmd,
        (ClientData) NULL, NULL);

  return TCL_OK;
}

